Package: debrief
Title: Text-Based Summaries for 'profvis' Profiling Data
Version: 0.1.0
Authors@R: c(
    person("Emil", "Hvitfeldt", , "emil.hvitfeldt@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"), comment = c(ROR = "03wc8by49"))
  )
Description: Provides text-based summaries and analysis tools for
    'profvis' profiling output. Designed for terminal workflows and
    artificial intelligence (AI) agent consumption, offering views
    including hotspot analysis, call
    trees, source context, caller/callee relationships, and memory
    allocation breakdowns.
License: MIT + file LICENSE
URL: https://github.com/r-lib/debrief, https://r-lib.github.io/debrief/
BugReports: https://github.com/r-lib/debrief/issues
Depends: R (>= 4.1)
Suggests: knitr, profvis, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-21 19:06:03 UTC; emilhvitfeldt
Author: Emil Hvitfeldt [aut, cre],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Emil Hvitfeldt <emil.hvitfeldt@posit.co>
Repository: CRAN
Date/Publication: 2026-04-21 21:22:26 UTC
Built: R 4.5.2; ; 2026-04-22 00:05:50 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,debrief_call_depth)
S3method(print,debrief_call_stats)
S3method(print,debrief_callers_callees)
S3method(print,debrief_compare)
S3method(print,debrief_compare_many)
S3method(print,debrief_debrief)
S3method(print,debrief_file_summary)
S3method(print,debrief_flame)
S3method(print,debrief_flame_condense)
S3method(print,debrief_focus)
S3method(print,debrief_gc_pressure)
S3method(print,debrief_help)
S3method(print,debrief_hot_lines)
S3method(print,debrief_hot_paths)
S3method(print,debrief_memory)
S3method(print,debrief_recursive)
S3method(print,debrief_source_context)
S3method(print,debrief_suggestions)
export(pv_call_depth)
export(pv_call_stats)
export(pv_callees)
export(pv_callers)
export(pv_compare)
export(pv_compare_many)
export(pv_debrief)
export(pv_example)
export(pv_file_summary)
export(pv_flame)
export(pv_flame_condense)
export(pv_focus)
export(pv_gc_pressure)
export(pv_help)
export(pv_hot_lines)
export(pv_hot_paths)
export(pv_memory)
export(pv_memory_lines)
export(pv_print_call_depth)
export(pv_print_call_stats)
export(pv_print_callers_callees)
export(pv_print_compare)
export(pv_print_compare_many)
export(pv_print_debrief)
export(pv_print_file_summary)
export(pv_print_gc_pressure)
export(pv_print_hot_lines)
export(pv_print_hot_paths)
export(pv_print_memory)
export(pv_print_recursive)
export(pv_print_suggestions)
export(pv_recursive)
export(pv_self_time)
export(pv_source_context)
export(pv_suggestions)
export(pv_to_json)
export(pv_to_list)
export(pv_total_time)
export(pv_worst_line)
importFrom(utils,head)

# debrief: Project Summarizer

`debrief` generates a comprehensive `BRIEF.md` file designed to provide new contributors or coding agents with a high-density overview of a Python project.

## Installation

```bash
pip install debrief
```

## Usage

```bash
debrief [run|lint] [path] [--output BRIEF.md]
```

### Arguments

| Argument               | Description                                        | Default             |
| :--------------------- | :------------------------------------------------- | :------------------ |
| `mode`                 | `run` generates BRIEF.md, `lint` only runs checks. | `run`               |
| `path`                 | Project root path.                                 | `.`                 |
| `-o`, `--output`       | Output filename.                                   | `BRIEF.md`          |
| `--tree-budget`        | Max lines for Directory Tree (auto-depth).         | `60`                |
| `--max-tree-siblings`  | Max items at same level in tree.                   | `tree_budget/3`     |
| `--max-readme`         | Max lines to include from README.                  | `20`                |
| `--max-deps`           | Max lines for dependencies list.                   | `15`                |
| `--max-imports`        | Max lines for Import Tree.                         | `50`                |
| `--max-definitions`    | Max lines for Module Definitions.                  | `200`               |
| `--max-class-methods`  | Max public methods shown per class.                | `max_definitions/3` |
| `--max-module-defs`    | Max top-level defs shown per module.               | `max_definitions/3` |
| `--include-docstrings` | Include docstrings in the output.                  | `False`             |
| `--exclude`            | Additional patterns to exclude.                    | None                |

### Lint Mode

```bash
debrief lint [path]
```

Runs all quality checks without generating `BRIEF.md`:

| Check                  | Threshold         | Condition                        |
| :--------------------- | :---------------- | :------------------------------- |
| README non-empty lines | ≥ 2               | Always                           |
| Description length     | ≥ 16 non-ws chars | Always                           |
| Docstring length       | ≥ 16 non-ws chars | Only with `--include-docstrings` |

## Features

- **Project Metadata**: Extracts description and dependencies from `pyproject.toml` (with `requirements.txt` fallback).
- **Directory Tree**: Adaptive depth tree that fits within a line budget, respecting `.gitignore`.
- **Import Analysis**: Generates an import dependency tree to visualize project structure.
- **Code Definitions**: Extracts class and function signatures with docstrings for all Python files.
- **Optimized Output**:
  - **Truncation**: Automatically truncates long lines (>300 chars) and large sections.
  - **"Read more"**: Links to local files for truncated content.
  - **Markdown**: Formatted for optimal readability.

## Example Output (`BRIEF.md`)

The [BRIEF.md](BRIEF.md) file in this repository serves as a real example of the output of `debrief`.

# Helper functions for testing debrief

# Create a mock profvis object for testing
# This allows testing without running actual profiling
mock_profvis <- function(
  prof = NULL,
  interval = 10,
  files = NULL
) {
  if (is.null(prof)) {
    # Default simple profile data
    prof <- data.frame(
      time = c(1L, 1L, 2L, 2L, 2L, 3L, 3L, 4L, 4L, 4L, 5L),
      depth = c(1L, 2L, 1L, 2L, 3L, 1L, 2L, 1L, 2L, 3L, 1L),
      label = c(
        "outer",
        "inner",
        "outer",
        "inner",
        "deep",
        "outer",
        "helper",
        "outer",
        "inner",
        "deep",
        "outer"
      ),
      filename = c(
        "R/main.R",
        "R/main.R",
        "R/main.R",
        "R/main.R",
        "R/utils.R",
        "R/main.R",
        "R/helper.R",
        "R/main.R",
        "R/main.R",
        "R/utils.R",
        "R/main.R"
      ),
      linenum = as.double(c(
        10,
        15,
        10,
        15,
        5,
        10,
        20,
        10,
        15,
        5,
        10
      )),
      filenum = as.double(c(1, 1, 1, 1, 2, 1, 3, 1, 1, 2, 1)),
      memalloc = c(100, 150, 150, 200, 250, 250, 300, 300, 350, 400, 400),
      meminc = c(0, 50, 0, 50, 50, 0, 50, 0, 50, 50, 0),
      stringsAsFactors = FALSE
    )
  }

  if (is.null(files)) {
    files <- list(
      list(
        filename = "R/main.R",
        content = paste(
          "# Main file",
          "outer <- function() {",
          "  x <- 1",
          "  y <- 2",

          "  inner()",
          "}",
          "",
          "inner <- function() {",
          "  result <- deep()",
          "  result",
          "}",
          "",
          "",
          "",
          "  z <- heavy_computation()",
          sep = "\n"
        ),
        normpath = "R/main.R"
      ),
      list(
        filename = "R/utils.R",
        content = paste(
          "# Utils",
          "deep <- function() {",
          "  Sys.sleep(0.01)",
          "  42",
          "  x <- rnorm(1000)",
          "}",
          sep = "\n"
        ),
        normpath = "R/utils.R"
      ),
      list(
        filename = "R/helper.R",
        content = paste(
          "# Helper functions",
          "helper <- function() {",
          "  paste('hello', 'world')",
          "}",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "",
          "  do_work()",
          sep = "\n"
        ),
        normpath = "R/helper.R"
      )
    )
  }

  structure(
    list(
      x = list(
        message = list(
          prof = prof,
          interval = interval,
          files = files
        )
      )
    ),
    class = "profvis"
  )
}

# Create a mock profvis with no source references
mock_profvis_no_source <- function() {
  prof <- data.frame(
    time = c(1L, 1L, 2L, 2L, 3L),
    depth = c(1L, 2L, 1L, 2L, 1L),
    label = c("foo", "bar", "foo", "baz", "foo"),
    filename = rep(NA_character_, 5),
    linenum = rep(NA_real_, 5),
    filenum = rep(NA_real_, 5),
    memalloc = c(100, 150, 150, 200, 200),
    meminc = c(0, 50, 0, 50, 0),
    stringsAsFactors = FALSE
  )

  mock_profvis(prof = prof, files = list())
}

# Create a mock profvis with recursive function
mock_profvis_recursive <- function() {
  prof <- data.frame(
    time = c(1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L, 3L, 3L, 3L),
    depth = c(1L, 2L, 3L, 1L, 2L, 3L, 4L, 1L, 2L, 3L, 4L, 5L),
    label = c(
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse",
      "recurse"
    ),
    filename = rep("R/recursive.R", 12),
    linenum = rep(5, 12),
    filenum = rep(1, 12),
    memalloc = seq(100, 1200, by = 100),
    meminc = rep(100, 12),
    stringsAsFactors = FALSE
  )

  files <- list(
    list(
      filename = "R/recursive.R",
      content = paste(
        "recurse <- function(n) {",
        "  if (n <= 0) return(1)",
        "  recurse(n - 1)",
        "}",
        "  recurse(n - 1)",
        sep = "\n"
      ),
      normpath = "R/recursive.R"
    )
  )

  mock_profvis(prof = prof, files = files)
}

# Create a mock profvis with GC and memory pressure
mock_profvis_gc <- function() {
  prof <- data.frame(
    time = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L),
    depth = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L),
    label = c(
      "work",
      "work",
      "<GC>",
      "work",
      "<GC>",
      "<GC>",
      "work",
      "<GC>",
      "work",
      "work"
    ),
    filename = c(
      rep("R/work.R", 2),
      NA,
      "R/work.R",
      NA,
      NA,
      "R/work.R",
      NA,
      rep("R/work.R", 2)
    ),
    linenum = as.double(c(5, 5, NA, 5, NA, NA, 5, NA, 5, 5)),
    filenum = as.double(c(1, 1, NA, 1, NA, NA, 1, NA, 1, 1)),
    memalloc = c(100, 200, 200, 300, 300, 300, 400, 400, 500, 600),
    meminc = c(0, 100, 0, 100, 0, 0, 100, 0, 100, 100),
    stringsAsFactors = FALSE
  )

  files <- list(
    list(
      filename = "R/work.R",
      content = "work <- function() {\n  x <- rnorm(1e6)\n  y <- cumsum(x)\n  z <- paste(x)\n  z\n}",
      normpath = "R/work.R"
    )
  )

  mock_profvis(prof = prof, files = files)
}

# Create a mock profvis with string operations
mock_profvis_strings <- function() {
  prof <- data.frame(
    time = c(1L, 1L, 2L, 2L, 3L, 3L, 4L, 4L, 5L),
    depth = c(1L, 2L, 1L, 2L, 1L, 2L, 1L, 2L, 1L),
    label = c(
      "process",
      "paste",
      "process",
      "paste0",
      "process",
      "sprintf",
      "process",
      "gsub",
      "process"
    ),
    filename = rep("R/strings.R", 9),
    linenum = as.double(c(1, 2, 1, 3, 1, 4, 1, 5, 1)),
    filenum = rep(1, 9),
    memalloc = seq(100, 900, by = 100),
    meminc = rep(100, 9),
    stringsAsFactors = FALSE
  )

  files <- list(
    list(
      filename = "R/strings.R",
      content = paste(
        "process <- function() {",
        "  paste('a', 'b')",
        "  paste0('a', 'b')",
        "  sprintf('%s', 'a')",
        "  gsub('a', 'b', 'aaa')",
        "}",
        sep = "\n"
      ),
      normpath = "R/strings.R"
    )
  )

  mock_profvis(prof = prof, files = files)
}

# Create a mock profvis with data frame operations
mock_profvis_df_ops <- function() {
  prof <- data.frame(
    time = rep(1:10, each = 3),
    depth = rep(1:3, 10),
    label = rep(c("process", "loop_body", "[.data.frame"), 10),
    filename = rep(c("R/df.R", "R/df.R", NA), 10),
    linenum = as.double(rep(c(1, 5, NA), 10)),
    filenum = as.double(rep(c(1, 1, NA), 10)),
    memalloc = seq(100, 3000, by = 100),
    meminc = rep(c(0, 50, 50), 10),
    stringsAsFactors = FALSE
  )

  files <- list(
    list(
      filename = "R/df.R",
      content = paste(
        "process <- function(df) {",
        "  for (i in 1:nrow(df)) {",
        "    loop_body(df, i)",
        "  }",
        "  df[i, ]$col",
        "}",
        sep = "\n"
      ),
      normpath = "R/df.R"
    )
  )

  mock_profvis(prof = prof, files = files)
}

# Tests to verify mock profvis objects match real profvis structure

test_that("mock profvis has correct class", {
  p <- pv_example()
  expect_s3_class(p, "profvis")
})

test_that("mock profvis has correct top-level structure", {
  p <- pv_example()

  expect_true("x" %in% names(p))
  expect_true("message" %in% names(p$x))
  expect_true("prof" %in% names(p$x$message))
  expect_true("interval" %in% names(p$x$message))
  expect_true("files" %in% names(p$x$message))
})

test_that("mock profvis prof has correct columns", {
  p <- pv_example()
  prof <- p$x$message$prof

  expected_cols <- c(
    "time",
    "depth",
    "label",
    "filename",
    "linenum",
    "filenum",
    "memalloc",
    "meminc"
  )

  expect_s3_class(prof, "data.frame")
  expect_true(all(expected_cols %in% names(prof)))
})

test_that("mock profvis prof has correct column types", {
  p <- pv_example()
  prof <- p$x$message$prof

  expect_type(prof$time, "integer")
  expect_type(prof$depth, "integer")
  expect_type(prof$label, "character")
  expect_type(prof$filename, "character")
  expect_type(prof$linenum, "double")
  expect_type(prof$filenum, "double")
  expect_type(prof$memalloc, "double")
  expect_type(prof$meminc, "double")
})

test_that("mock profvis interval is numeric", {
  p <- pv_example()

  expect_type(p$x$message$interval, "double")
  expect_length(p$x$message$interval, 1)
})

test_that("mock profvis files is a list", {
  p <- pv_example()

  expect_type(p$x$message$files, "list")
})

test_that("mock profvis files have correct structure", {
  p <- pv_example()
  files <- p$x$message$files

  skip_if(length(files) == 0)

  for (f in files) {
    expect_true("filename" %in% names(f))
    expect_true("content" %in% names(f))
    expect_type(f$filename, "character")
    expect_type(f$content, "character")
  }
})

test_that("all pv_example types produce valid structure", {
  types <- c("default", "no_source", "recursive", "gc")

  for (type in types) {
    p <- pv_example(type)

    expect_s3_class(p, "profvis")
    expect_s3_class(p$x$message$prof, "data.frame")
    expect_type(p$x$message$interval, "double")
    expect_type(p$x$message$files, "list")
  }
})

# Validate all test helper mock objects
test_that("mock_profvis produces valid structure", {
  p <- mock_profvis()

  expect_s3_class(p, "profvis")
  expect_s3_class(p$x$message$prof, "data.frame")
  expect_type(p$x$message$interval, "double")
  expect_type(p$x$message$files, "list")

  prof <- p$x$message$prof
  expect_true(all(
    c(
      "time",
      "depth",
      "label",
      "filename",
      "linenum",
      "memalloc",
      "meminc"
    ) %in%
      names(prof)
  ))
})

test_that("mock_profvis_no_source produces valid structure", {
  p <- mock_profvis_no_source()

  expect_s3_class(p, "profvis")
  expect_s3_class(p$x$message$prof, "data.frame")
  expect_type(p$x$message$interval, "double")
  expect_type(p$x$message$files, "list")

  prof <- p$x$message$prof
  expect_true(all(
    c(
      "time",
      "depth",
      "label",
      "filename",
      "linenum",
      "memalloc",
      "meminc"
    ) %in%
      names(prof)
  ))

  # Verify no source refs
  expect_true(all(is.na(prof$filename)))
})

test_that("mock_profvis_recursive produces valid structure", {
  p <- mock_profvis_recursive()

  expect_s3_class(p, "profvis")
  expect_s3_class(p$x$message$prof, "data.frame")
  expect_type(p$x$message$interval, "double")
  expect_type(p$x$message$files, "list")

  prof <- p$x$message$prof
  expect_true(all(
    c(
      "time",
      "depth",
      "label",
      "filename",
      "linenum",
      "memalloc",
      "meminc"
    ) %in%
      names(prof)
  ))

  # Verify it has recursive structure (same function at multiple depths)
  expect_true("recurse" %in% prof$label)
})

test_that("mock_profvis_gc produces valid structure", {
  p <- mock_profvis_gc()

  expect_s3_class(p, "profvis")
  expect_s3_class(p$x$message$prof, "data.frame")
  expect_type(p$x$message$interval, "double")
  expect_type(p$x$message$files, "list")

  prof <- p$x$message$prof
  expect_true(all(
    c(
      "time",
      "depth",
      "label",
      "filename",
      "linenum",
      "memalloc",
      "meminc"
    ) %in%
      names(prof)
  ))

  # Verify it has GC entries
  expect_true("<GC>" %in% prof$label)
})

test_that("mock_profvis_strings produces valid structure", {
  p <- mock_profvis_strings()

  expect_s3_class(p, "profvis")
  expect_s3_class(p$x$message$prof, "data.frame")
  expect_type(p$x$message$interval, "double")
  expect_type(p$x$message$files, "list")

  prof <- p$x$message$prof
  expect_true(all(
    c(
      "time",
      "depth",
      "label",
      "filename",
      "linenum",
      "memalloc",
      "meminc"
    ) %in%
      names(prof)
  ))

  # Verify it has string operations
  string_funcs <- c("paste", "paste0", "sprintf", "gsub")
  expect_true(any(string_funcs %in% prof$label))
})

test_that("mock_profvis_df_ops produces valid structure", {
  p <- mock_profvis_df_ops()

  expect_s3_class(p, "profvis")
  expect_s3_class(p$x$message$prof, "data.frame")
  expect_type(p$x$message$interval, "double")
  expect_type(p$x$message$files, "list")

  prof <- p$x$message$prof
  expect_true(all(
    c(
      "time",
      "depth",
      "label",
      "filename",
      "linenum",
      "memalloc",
      "meminc"
    ) %in%
      names(prof)
  ))

  # Verify it has data frame operations
  expect_true("[.data.frame" %in% prof$label)
})

# Test against real profvis if available
test_that("mock has same essential structure as real profvis", {
  skip_if_not_installed("profvis")
  skip_on_cran()
  skip_on_covr()

  # Create a real profvis object using pause() to ensure samples are collected
  real_p <- profvis::profvis(
    {
      profvis::pause(0.1)
    },
    simplify = FALSE
  )

  mock_p <- pv_example()

  # Both should have "profvis" as a class (real may have additional classes)
  expect_true("profvis" %in% class(real_p))
  expect_true("profvis" %in% class(mock_p))

  # Both should have x$message structure
  expect_true("x" %in% names(real_p))
  expect_true("x" %in% names(mock_p))
  expect_true("message" %in% names(real_p$x))
  expect_true("message" %in% names(mock_p$x))

  # Compare message structure - mock should have required fields
  real_msg_names <- names(real_p$x$message)
  mock_msg_names <- names(mock_p$x$message)
  required_names <- c("prof", "interval", "files")
  expect_true(all(required_names %in% real_msg_names))
  expect_true(all(required_names %in% mock_msg_names))

  # Compare prof data frame
  real_prof <- real_p$x$message$prof
  mock_prof <- mock_p$x$message$prof

  expect_s3_class(real_prof, "data.frame")
  expect_s3_class(mock_prof, "data.frame")

  # Core columns our package uses should be in both
  core_cols <- c("time", "depth", "label")
  expect_true(all(core_cols %in% names(real_prof)))
  expect_true(all(core_cols %in% names(mock_prof)))

  # Column types should match
  expect_identical(typeof(real_prof$time), typeof(mock_prof$time))
  expect_identical(typeof(real_prof$depth), typeof(mock_prof$depth))
  expect_identical(typeof(real_prof$label), typeof(mock_prof$label))
})

# Helper function to validate a mock against real profvis structure
validate_mock_against_real <- function(mock_p, real_prof, mock_name) {
  mock_prof <- mock_p$x$message$prof

  # Class check
  expect_true(
    "profvis" %in% class(mock_p),
    label = paste(mock_name, "has profvis class")
  )

  # Structure check
  expect_true("x" %in% names(mock_p), label = paste(mock_name, "has x"))
  expect_true(
    "message" %in% names(mock_p$x),
    label = paste(mock_name, "has message")
  )
  expect_true(
    "prof" %in% names(mock_p$x$message),
    label = paste(mock_name, "has prof")
  )
  expect_true(
    "interval" %in% names(mock_p$x$message),
    label = paste(mock_name, "has interval")
  )
  expect_true(
    "files" %in% names(mock_p$x$message),
    label = paste(mock_name, "has files")
  )

  # Prof is data frame
  expect_s3_class(mock_prof, "data.frame")

  # These are the columns our package actually uses
  used_cols <- c(
    "time",
    "depth",
    "label",
    "filename",
    "linenum",
    "memalloc",
    "meminc"
  )

  # All used columns should be present
  for (col in used_cols) {
    expect_true(
      col %in% names(mock_prof),
      label = paste(mock_name, "has column", col)
    )
  }

  # Column types should match real profvis
  for (col in used_cols) {
    expect_identical(
      typeof(real_prof[[col]]),
      typeof(mock_prof[[col]]),
      label = paste(mock_name, "type match for", col)
    )
  }
}

test_that("all pv_example types match real profvis structure", {
  skip_if_not_installed("profvis")
  skip_on_cran()
  skip_on_covr()

  # Create a real profvis object
  real_p <- profvis::profvis(
    {
      profvis::pause(0.1)
    },
    simplify = FALSE
  )
  real_prof <- real_p$x$message$prof

  # Test all pv_example types
  types <- c("default", "no_source", "recursive", "gc")
  for (type in types) {
    mock_p <- pv_example(type)
    validate_mock_against_real(
      mock_p,
      real_prof,
      paste("pv_example(", type, ")")
    )
  }
})

test_that("all test helper mocks match real profvis structure", {
  skip_if_not_installed("profvis")
  skip_on_cran()
  skip_on_covr()

  # Create a real profvis object
  real_p <- profvis::profvis(
    {
      profvis::pause(0.1)
    },
    simplify = FALSE
  )
  real_prof <- real_p$x$message$prof

  # Test all helper mock functions
  validate_mock_against_real(mock_profvis(), real_prof, "mock_profvis()")
  validate_mock_against_real(
    mock_profvis_no_source(),
    real_prof,
    "mock_profvis_no_source()"
  )
  validate_mock_against_real(
    mock_profvis_recursive(),
    real_prof,
    "mock_profvis_recursive()"
  )
  validate_mock_against_real(mock_profvis_gc(), real_prof, "mock_profvis_gc()")
  validate_mock_against_real(
    mock_profvis_strings(),
    real_prof,
    "mock_profvis_strings()"
  )
  validate_mock_against_real(
    mock_profvis_df_ops(),
    real_prof,
    "mock_profvis_df_ops()"
  )
})

test_that("pv_call_depth returns correct structure", {
  p <- mock_profvis()
  result <- pv_call_depth(p)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("depth", "samples", "time_ms", "pct", "top_funcs"))
})

test_that("pv_call_depth has sequential depths", {
  p <- mock_profvis()
  result <- pv_call_depth(p)

  # Depths should be sequential starting from 1
  expect_equal(min(result$depth), 1)
  expect_true(all(diff(result$depth) == 1))
})

test_that("pv_print_call_depth snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_call_depth(p))
})

test_that("pv_callers returns correct structure", {
  p <- mock_profvis()
  result <- pv_callers(p, "inner")

  expect_s3_class(result, "data.frame")
  expect_named(result, c("label", "samples", "pct"))
})

test_that("pv_callers finds correct callers", {
  p <- mock_profvis()
  result <- pv_callers(p, "inner")

  # inner is called by outer
  expect_true("outer" %in% result$label)
})

test_that("pv_callers handles non-existent function", {
  p <- mock_profvis()

  expect_message(result <- pv_callers(p, "nonexistent"), "not found")
  expect_equal(nrow(result), 0)
})

test_that("pv_callees returns correct structure", {
  p <- mock_profvis()
  result <- pv_callees(p, "outer")

  expect_s3_class(result, "data.frame")
  expect_named(result, c("label", "samples", "pct"))
})

test_that("pv_callees finds correct callees", {
  p <- mock_profvis()
  result <- pv_callees(p, "outer")

  # outer calls inner and helper
  expect_true("inner" %in% result$label || "helper" %in% result$label)
})

test_that("pv_callees handles non-existent function", {
  p <- mock_profvis()

  expect_message(result <- pv_callees(p, "nonexistent"), "not found")
  expect_equal(nrow(result), 0)
})

test_that("pv_print_callers_callees snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_callers_callees(p, "inner"))
})

test_that("pv_call_stats returns correct structure", {
  p <- mock_profvis()
  result <- pv_call_stats(p)

  expect_s3_class(result, "data.frame")
  expect_named(
    result,
    c("label", "calls", "total_ms", "self_ms", "child_ms", "ms_per_call", "pct")
  )
})

test_that("pv_call_stats child_ms = total_ms - self_ms", {
  p <- mock_profvis()
  result <- pv_call_stats(p)

  expect_equal(result$child_ms, result$total_ms - result$self_ms)
})

test_that("pv_call_stats respects n parameter", {
  p <- mock_profvis()

  result_all <- pv_call_stats(p)
  result_2 <- pv_call_stats(p, n = 2)

  expect_lte(nrow(result_2), 2)
})

test_that("pv_print_call_stats snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_call_stats(p))
})

test_that("call analysis functions reject non-profvis input", {
  expect_error(pv_call_depth(list()), "must be a profvis object")
  expect_error(pv_callers("bad", "func"), "must be a profvis object")
  expect_error(pv_callees(42, "func"), "must be a profvis object")
  expect_error(pv_call_stats(NULL), "must be a profvis object")
})

test_that("pv_call_stats handles recursive functions", {
  p <- mock_profvis_recursive()
  result <- pv_call_stats(p)

  expect_s3_class(result, "data.frame")
  expect_in("recurse", result$label)
  recurse_row <- result[result$label == "recurse", ]
  expect_gt(recurse_row$calls, 1)
})

test_that("pv_callees returns empty for leaf function", {
  p <- mock_profvis_no_source()
  result <- pv_callees(p, "bar")

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_print_callers_callees shows none when callees empty", {
  p <- mock_profvis_no_source()
  expect_snapshot(pv_print_callers_callees(p, "bar"))
})

test_that("pv_focus snapshot for existing function", {
  p <- mock_profvis()
  expect_snapshot(pv_focus(p, "inner"))
})

test_that("pv_focus handles non-existent function", {
  p <- mock_profvis()
  expect_snapshot(pv_focus(p, "nonexistent"))
})

test_that("pv_focus returns invisibly", {
  p <- mock_profvis()

  result <- expect_invisible(pv_focus(p, "inner"))
  expect_s3_class(result, "debrief_focus")
})

test_that("pv_focus returns NULL for non-existent function", {
  p <- mock_profvis()

  result <- expect_invisible(pv_focus(p, "nonexistent"))
  expect_s3_class(result, "debrief_focus")
})

test_that("pv_focus handles no source refs", {
  p <- mock_profvis_no_source()
  expect_snapshot(pv_focus(p, "foo"))
})

test_that("pv_focus context parameter works", {
  p <- mock_profvis()

  # Should not error with different context values
  expect_no_error(capture.output(pv_focus(p, "inner", context = 2)))
  expect_no_error(capture.output(pv_focus(p, "inner", context = 10)))
})

test_that("pv_focus rejects non-profvis input", {
  expect_error(pv_focus(list(), "func"), "must be a profvis object")
})

test_that("pv_focus handles more than 5 callers", {
  n_callers <- 6
  times <- seq_len(n_callers)
  prof <- data.frame(
    time = rep(times, each = 2),
    depth = rep(c(1L, 2L), n_callers),
    label = c(rbind(
      paste0("caller", times),
      rep("target", n_callers)
    )),
    filename = rep(NA_character_, n_callers * 2),
    linenum = rep(NA_real_, n_callers * 2),
    filenum = rep(NA_real_, n_callers * 2),
    memalloc = seq(100, by = 100, length.out = n_callers * 2),
    meminc = rep(0, n_callers * 2),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_focus(p, "target"))
})

test_that("pv_focus shows callees none for leaf function", {
  p <- mock_profvis_no_source()
  expect_snapshot(pv_focus(p, "bar"))
})

test_that("pv_focus handles more than 5 callees", {
  n_callees <- 6
  times <- seq_len(n_callees)
  prof <- data.frame(
    time = rep(times, each = 2),
    depth = rep(c(1L, 2L), n_callees),
    label = c(rbind(
      rep("target", n_callees),
      paste0("callee", times)
    )),
    filename = rep(NA_character_, n_callees * 2),
    linenum = rep(NA_real_, n_callees * 2),
    filenum = rep(NA_real_, n_callees * 2),
    memalloc = seq(100, by = 100, length.out = n_callees * 2),
    meminc = rep(0, n_callees * 2),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_focus(p, "target"))
})

test_that("pv_focus handles function with source refs but no self-time with source", {
  prof <- data.frame(
    time = c(1L, 1L),
    depth = c(1L, 2L),
    label = c("target", "deep"),
    filename = c("R/main.R", NA_character_),
    linenum = c(5.0, NA_real_),
    filenum = c(1.0, NA_real_),
    memalloc = c(100, 200),
    meminc = c(0, 100),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_focus(p, "target"))
})

test_that("pv_source_context snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_source_context(p, "R/main.R"))
})

test_that("pv_source_context finds file by partial match", {
  p <- mock_profvis()

  # Should match partial filename
  expect_no_error(capture.output(pv_source_context(p, "main.R")))
})

test_that("pv_source_context handles non-existent file", {
  p <- mock_profvis()
  expect_snapshot(pv_source_context(p, "nonexistent.R"))
})

test_that("pv_source_context auto-selects hottest line when linenum is NULL", {
  p <- mock_profvis()
  expect_snapshot(pv_source_context(p, "R/main.R", linenum = NULL))
})

test_that("pv_source_context respects linenum parameter", {
  p <- mock_profvis()
  expect_snapshot(pv_source_context(p, "R/main.R", linenum = 5))
})

test_that("pv_source_context respects context parameter", {
  p <- mock_profvis()

  # Different context values should work
  expect_no_error(capture.output(pv_source_context(p, "R/main.R", context = 3)))
  expect_no_error(capture.output(pv_source_context(
    p,
    "R/main.R",
    context = 15
  )))
})

test_that("pv_source_context returns line data invisibly", {
  p <- mock_profvis()

  result <- expect_invisible(pv_source_context(p, "R/main.R"))
  expect_s3_class(result, "debrief_source_context")
})

test_that("pv_file_summary returns correct structure", {
  p <- mock_profvis()
  result <- pv_file_summary(p)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("filename", "samples", "time_ms", "pct"))
})

test_that("pv_file_summary returns empty when no source refs", {
  p <- mock_profvis_no_source()
  result <- pv_file_summary(p)

  expect_equal(nrow(result), 0)
})

test_that("pv_file_summary includes all source files", {
  p <- mock_profvis()
  result <- pv_file_summary(p)

  expect_true("R/main.R" %in% result$filename)
})

test_that("pv_print_file_summary snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_file_summary(p))
})

test_that("pv_print_file_summary handles no source refs", {
  p <- mock_profvis_no_source()
  expect_snapshot(pv_print_file_summary(p))
})

test_that("source context functions reject non-profvis input", {
  expect_error(pv_source_context(list(), "file.R"), "must be a profvis object")
  expect_error(pv_file_summary("bad"), "must be a profvis object")
  expect_error(pv_print_file_summary(42), "must be a profvis object")
})

test_that("pv_source_context warns when multiple files match", {
  p <- mock_profvis()
  expect_snapshot(pv_source_context(p, "R/"))
})

test_that("pv_source_context uses min linenum when no top_of_stack for file", {
  prof <- data.frame(
    time = c(1L, 1L),
    depth = c(1L, 2L),
    label = c("outer", "deep"),
    filename = c("R/main.R", NA_character_),
    linenum = c(10.0, NA_real_),
    filenum = c(1.0, NA_real_),
    memalloc = c(100, 200),
    meminc = c(0, 100),
    stringsAsFactors = FALSE
  )
  files <- list(
    list(
      filename = "R/main.R",
      content = paste(paste0("line", 1:15), collapse = "\n"),
      normpath = "R/main.R"
    )
  )
  p <- mock_profvis(prof = prof, files = files)
  expect_snapshot(pv_source_context(p, "R/main.R", linenum = NULL))
})

test_that("pv_source_context shows source not available when files empty", {
  prof <- data.frame(
    time = c(1L, 2L),
    depth = rep(1L, 2),
    label = rep("func", 2),
    filename = rep("R/main.R", 2),
    linenum = rep(5.0, 2),
    filenum = rep(1.0, 2),
    memalloc = c(100, 200),
    meminc = rep(0, 2),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_source_context(p, "R/main.R", linenum = 5))
})

# Tests for GC pressure detection and suggestions

test_that("pv_gc_pressure returns correct structure", {
  p <- mock_profvis_gc()
  result <- pv_gc_pressure(p)

  expect_s3_class(result, "data.frame")
  expect_named(
    result,
    c("severity", "pct", "time_ms", "issue", "cause", "actions")
  )
})

test_that("pv_gc_pressure detects high GC", {
  p <- mock_profvis_gc()
  result <- pv_gc_pressure(p)

  # mock_profvis_gc has 40% GC time, should be detected
  expect_equal(nrow(result), 1)
  expect_equal(result$severity, "high")
  expect_gt(result$pct, 25)
})

test_that("pv_gc_pressure returns empty for low GC", {
  # Profile with no GC
  p <- mock_profvis()
  result <- pv_gc_pressure(p)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_gc_pressure returns empty for profile without GC entries", {
  prof <- data.frame(
    time = 1:3,
    depth = rep(1L, 3),
    label = c("good_func", "good_func", "good_func"),
    filename = rep(NA_character_, 3),
    linenum = rep(NA_real_, 3),
    filenum = rep(NA_real_, 3),
    memalloc = c(100, 100, 100),
    meminc = c(0, 0, 0),
    stringsAsFactors = FALSE
  )

  p <- mock_profvis(prof = prof, files = list())
  result <- pv_gc_pressure(p)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_print_gc_pressure snapshot with high GC", {
  p <- mock_profvis_gc()
  expect_snapshot(pv_print_gc_pressure(p))
})

test_that("pv_print_gc_pressure snapshot with no GC", {
  p <- mock_profvis()
  expect_snapshot(pv_print_gc_pressure(p))
})

test_that("pv_suggestions returns correct structure", {
  p <- mock_profvis()
  result <- pv_suggestions(p)

  expect_s3_class(result, "data.frame")
  expect_named(
    result,
    c(
      "priority",
      "category",
      "location",
      "action",
      "pattern",
      "replacement",
      "potential_impact"
    )
  )
})

test_that("pv_suggestions priority is ordered", {
  p <- mock_profvis_gc()
  result <- pv_suggestions(p)

  if (nrow(result) > 1) {
    # Priorities should be in increasing order
    expect_true(all(diff(result$priority) >= 0))
  }
})

test_that("pv_print_suggestions snapshot with GC pressure", {
  p <- mock_profvis_gc()
  expect_snapshot(pv_print_suggestions(p))
})

test_that("pv_print_suggestions handles profile with no suggestions", {
  prof <- data.frame(
    time = 1L,
    depth = 1L,
    label = "x",
    filename = NA_character_,
    linenum = NA_real_,
    filenum = NA_real_,
    memalloc = 100,
    meminc = 0,
    stringsAsFactors = FALSE
  )

  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_print_suggestions(p))
})

test_that("diagnostics functions reject non-profvis input", {
  expect_error(pv_gc_pressure(list()), "must be a profvis object")
  expect_error(pv_suggestions("bad"), "must be a profvis object")
  expect_error(pv_print_gc_pressure(42), "must be a profvis object")
  expect_error(pv_print_suggestions(NULL), "must be a profvis object")
})

test_that("pv_gc_pressure returns empty when GC exists but below threshold", {
  prof <- data.frame(
    time = 1:10,
    depth = rep(1L, 10),
    label = c(rep("work", 9), "<GC>"),
    filename = c(rep("R/work.R", 9), NA_character_),
    linenum = as.double(c(rep(5, 9), NA)),
    filenum = as.double(c(rep(1, 9), NA)),
    memalloc = seq(100, 1000, by = 100),
    meminc = rep(0, 10),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  result <- pv_gc_pressure(p)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_gc_pressure detects medium severity", {
  prof <- data.frame(
    time = 1:20,
    depth = rep(1L, 20),
    label = c(rep("work", 16), rep("<GC>", 4)),
    filename = c(rep("R/work.R", 16), rep(NA_character_, 4)),
    linenum = as.double(c(rep(5, 16), rep(NA, 4))),
    filenum = as.double(c(rep(1, 16), rep(NA, 4))),
    memalloc = seq(100, 2000, by = 100),
    meminc = rep(0, 20),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  result <- pv_gc_pressure(p)

  expect_equal(nrow(result), 1)
  expect_equal(result$severity, "medium")
})

test_that("pv_gc_pressure detects low severity", {
  prof <- data.frame(
    time = 1:20,
    depth = rep(1L, 20),
    label = c(rep("work", 17), rep("<GC>", 3)),
    filename = c(rep("R/work.R", 17), rep(NA_character_, 3)),
    linenum = as.double(c(rep(5, 17), rep(NA, 3))),
    filenum = as.double(c(rep(1, 17), rep(NA, 3))),
    memalloc = seq(100, 2000, by = 100),
    meminc = rep(0, 20),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  result <- pv_gc_pressure(p)

  expect_equal(nrow(result), 1)
  expect_equal(result$severity, "low")
})

test_that("pv_suggestions returns empty data frame for profile with only internal functions", {
  prof <- data.frame(
    time = 1L,
    depth = 1L,
    label = "(top-level)",
    filename = NA_character_,
    linenum = NA_real_,
    filenum = NA_real_,
    memalloc = 100,
    meminc = 0,
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  result <- pv_suggestions(p)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_print_suggestions handles profile with truly no suggestions", {
  prof <- data.frame(
    time = 1L,
    depth = 1L,
    label = "(top-level)",
    filename = NA_character_,
    linenum = NA_real_,
    filenum = NA_real_,
    memalloc = 100,
    meminc = 0,
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_print_suggestions(p))
})

test_that("pv_suggestions detects df vectorization antipattern", {
  p <- mock_profvis_df_ops()
  result <- pv_suggestions(p)

  expect_s3_class(result, "data.frame")
  expect_in("data structure", result$category)
})

test_that("pv_suggestions detects recursion optimization", {
  p <- mock_profvis_recursive()
  result <- pv_suggestions(p)

  expect_s3_class(result, "data.frame")
  expect_in("algorithm", result$category)
})

test_that("pv_suggestions detects string optimization", {
  p <- mock_profvis_strings()
  result <- pv_suggestions(p)

  expect_s3_class(result, "data.frame")
  expect_in("string operations", result$category)
})

test_that("pv_suggestions handles has_source=TRUE but no hot lines", {
  prof <- data.frame(
    time = c(1L, 1L),
    depth = c(1L, 2L),
    label = c("outer", "inner"),
    filename = c("R/main.R", NA_character_),
    linenum = c(5.0, NA_real_),
    filenum = c(1.0, NA_real_),
    memalloc = c(100, 200),
    meminc = c(0, 100),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())

  expect_no_error(pv_suggestions(p))
})

test_that("pv_suggestions handles top function that is internal", {
  prof <- data.frame(
    time = 1:5,
    depth = rep(1L, 5),
    label = rep("(top-level)", 5),
    filename = rep(NA_character_, 5),
    linenum = rep(NA_real_, 5),
    filenum = rep(NA_real_, 5),
    memalloc = seq(100, 500, by = 100),
    meminc = rep(0, 5),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  result <- pv_suggestions(p)

  expect_s3_class(result, "data.frame")
})

test_that("pv_self_time returns correct structure", {
  p <- mock_profvis()
  result <- pv_self_time(p)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("label", "samples", "time_ms", "pct"))
  expect_type(result$label, "character")
  expect_type(result$samples, "integer")
  expect_type(result$time_ms, "double")
  expect_type(result$pct, "double")
})

test_that("pv_self_time calculates self-time correctly", {
  p <- mock_profvis()
  result <- pv_self_time(p)

  # Self-time should only count when function is at top of stack
  # In mock data: inner is at top at time 1, deep at times 2,4, helper at time 3, outer at time 5
  expect_true("deep" %in% result$label)
  expect_true("inner" %in% result$label)
  expect_true("helper" %in% result$label)
  expect_true("outer" %in% result$label)

  # deep appears at top of stack at times 2 and 4 = 2 samples
  deep_row <- result[result$label == "deep", ]
  expect_equal(deep_row$samples, 2L)
})

test_that("pv_self_time percentages sum to 100", {
  p <- mock_profvis()
  result <- pv_self_time(p)

  # Percentages should sum to approximately 100
  expect_equal(sum(result$pct), 100, tolerance = 0.1)
})

test_that("pv_total_time returns correct structure", {
  p <- mock_profvis()
  result <- pv_total_time(p)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("label", "samples", "time_ms", "pct"))
})

test_that("pv_total_time calculates total time correctly", {
  p <- mock_profvis()
  result <- pv_total_time(p)

  # outer appears in all 5 time samples
  outer_row <- result[result$label == "outer", ]
  expect_equal(outer_row$samples, 5L)
  expect_equal(outer_row$pct, 100)
})

test_that("pv_total_time >= pv_self_time for all functions", {
  p <- mock_profvis()
  self <- pv_self_time(p)
  total <- pv_total_time(p)

  merged <- merge(self, total, by = "label", suffixes = c("_self", "_total"))
  expect_true(all(merged$samples_total >= merged$samples_self))
})

test_that("time functions reject non-profvis input", {
  expect_error(pv_self_time(list()), "must be a profvis object")
  expect_error(pv_total_time(data.frame()), "must be a profvis object")
})

test_that("time functions work with no source references", {
  p <- mock_profvis_no_source()

  result_self <- pv_self_time(p)
  result_total <- pv_total_time(p)

  expect_s3_class(result_self, "data.frame")
  expect_s3_class(result_total, "data.frame")
  expect_gt(nrow(result_self), 0)
  expect_gt(nrow(result_total), 0)
})

# Tests for export functions

test_that("pv_to_json returns valid JSON string", {
  p <- pv_example()
  json <- pv_to_json(p)

  expect_type(json, "character")
  expect_length(json, 1)
  expect_true(grepl("^\\{", json))
  expect_true(grepl("\\}$", json))
})

test_that("pv_to_json includes metadata", {
  p <- pv_example()
  json <- pv_to_json(p)

  expect_true(grepl("metadata", json))
  expect_true(grepl("total_time_ms", json))
  expect_true(grepl("total_samples", json))
  expect_true(grepl("interval_ms", json))
})

test_that("pv_to_json respects include parameter", {
  p <- pv_example()

  json_all <- pv_to_json(p)
  json_limited <- pv_to_json(p, include = c("self_time"))

  expect_true(grepl("self_time", json_all))
  expect_true(grepl("\"total_time\":", json_all))
  expect_true(grepl("self_time", json_limited))
  # Note: "total_time_ms" appears in metadata, so check for the section key

  expect_false(grepl("\"total_time\":", json_limited))
})

test_that("pv_to_json pretty parameter works", {
  p <- pv_example()

  json_pretty <- pv_to_json(p, pretty = TRUE)
  json_compact <- pv_to_json(p, pretty = FALSE)

  expect_true(grepl("\n", json_pretty))
  expect_false(grepl("\n", json_compact))
})

test_that("pv_to_json writes to file", {
  p <- pv_example()
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))

  result <- pv_to_json(p, file = tmp)

  expect_equal(result, tmp)
  expect_true(file.exists(tmp))

  content <- readLines(tmp)
  expect_true(length(content) > 0)
})

test_that("pv_to_list returns list with correct structure", {
  p <- pv_example()
  result <- pv_to_list(p)

  expect_type(result, "list")
  expect_true("metadata" %in% names(result))
  expect_true("self_time" %in% names(result))
  expect_true("total_time" %in% names(result))
})

test_that("pv_to_list metadata contains expected fields", {
  p <- pv_example()
  result <- pv_to_list(p)

  expect_true("total_time_ms" %in% names(result$metadata))
  expect_true("total_samples" %in% names(result$metadata))
  expect_true("interval_ms" %in% names(result$metadata))
  expect_true("has_source_refs" %in% names(result$metadata))
})

test_that("pv_to_list self_time is a data frame", {
  p <- pv_example()
  result <- pv_to_list(p)

  expect_s3_class(result$self_time, "data.frame")
  expect_true("label" %in% names(result$self_time))
  expect_true("time_ms" %in% names(result$self_time))
})

test_that("pv_to_list respects include parameter", {
  p <- pv_example()

  result_all <- pv_to_list(p)
  result_limited <- pv_to_list(p, include = "self_time")

  expect_true("self_time" %in% names(result_all))
  expect_true("total_time" %in% names(result_all))
  expect_true("self_time" %in% names(result_limited))
  expect_false("total_time" %in% names(result_limited))
})

test_that("pv_to_json handles empty results", {
  p <- pv_example("no_source")
  json <- pv_to_json(p)

  expect_type(json, "character")
  expect_true(grepl("hot_lines", json))
})

test_that("JSON serializer handles special characters", {
  # Test internal to_json function
  result <- debrief:::to_json(list(text = "line1\nline2\ttab"))
  expect_true(grepl("\\\\n", result))
  expect_true(grepl("\\\\t", result))
})

test_that("JSON serializer handles NA values", {
  result <- debrief:::to_json(list(val = NA))
  expect_true(grepl("null", result))
})

test_that("JSON serializer handles empty lists", {
  # Empty unnamed list is an empty array
  result <- debrief:::to_json(list())
  expect_equal(result, "[]")

  # Empty named list is an empty object
  result <- debrief:::to_json(structure(list(), names = character()))
  expect_equal(result, "{}")

  # Named list with empty value
  result <- debrief:::to_json(list(arr = list()))
  expect_true(grepl("\\[\\]", result))
})

test_that("pv_to_json includes system info when requested", {
  p <- pv_example()
  json <- pv_to_json(p, system_info = TRUE)

  expect_match(json, "r_version")
  expect_match(json, "platform")
})

test_that("pv_to_list includes system info when requested", {
  p <- pv_example()
  result <- pv_to_list(p, system_info = TRUE)

  expect_in("system", names(result$metadata))
  expect_in("r_version", names(result$metadata$system))
})

test_that("pv_to_json handles profile with no gc pressure", {
  p <- mock_profvis()
  json <- pv_to_json(p, include = "gc_pressure")

  expect_type(json, "character")
  expect_match(json, "gc_pressure")
})

test_that("pv_to_json handles profile with gc pressure", {
  p <- mock_profvis_gc()
  json <- pv_to_json(p, include = "gc_pressure")

  expect_type(json, "character")
  expect_match(json, "severity")
})

test_that("pv_to_json handles profile with no suggestions", {
  prof <- data.frame(
    time = 1L,
    depth = 1L,
    label = "(top-level)",
    filename = NA_character_,
    linenum = NA_real_,
    filenum = NA_real_,
    memalloc = 100,
    meminc = 0,
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  json <- pv_to_json(p, include = "suggestions")

  expect_type(json, "character")
  expect_match(json, "suggestions")
})

test_that("pv_to_json handles profile with no recursive functions", {
  p <- mock_profvis()
  json <- pv_to_json(p, include = "recursive")

  expect_type(json, "character")
  expect_match(json, "recursive")
})

test_that("pv_to_json handles profile with recursive functions", {
  p <- mock_profvis_recursive()
  json <- pv_to_json(p, include = "recursive")

  expect_type(json, "character")
  expect_match(json, "recurse")
})

test_that("JSON serializer handles NULL input", {
  result <- debrief:::to_json(NULL)
  expect_equal(result, "null")
})

test_that("JSON serializer handles atomic vector of length > 1", {
  result <- debrief:::to_json(c(1, 2, 3))
  expect_match(result, "\\[")
  expect_match(result, "1")
  expect_match(result, "3")
})

test_that("JSON serializer handles data frame input", {
  result <- debrief:::to_json(data.frame(x = 1, y = "a"))
  expect_match(result, "\\[")
  expect_match(result, '"x"')
})

test_that("pv_to_json covers gc_pressure empty path with default include", {
  prof <- data.frame(
    time = 1:3,
    depth = rep(1L, 3),
    label = rep("work", 3),
    filename = rep(NA_character_, 3),
    linenum = rep(NA_real_, 3),
    filenum = rep(NA_real_, 3),
    memalloc = c(100, 200, 300),
    meminc = rep(0, 3),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  json <- pv_to_json(p)

  expect_type(json, "character")
  expect_match(json, "gc_pressure")
  expect_match(json, "recursive")
})

test_that("JSON serializer returns null for non-serializable objects", {
  result <- debrief:::to_json(new.env())
  expect_equal(result, "null")
})

test_that("pv_recursive returns correct structure", {
  p <- mock_profvis_recursive()
  result <- pv_recursive(p)

  expect_s3_class(result, "data.frame")
  expect_named(
    result,
    c(
      "label",
      "max_depth",
      "avg_depth",
      "recursive_samples",
      "total_samples",
      "pct_recursive",
      "total_ms",
      "pct_time"
    )
  )
})

test_that("pv_recursive detects recursive functions", {
  p <- mock_profvis_recursive()
  result <- pv_recursive(p)

  expect_gt(nrow(result), 0)
  expect_true("recurse" %in% result$label)
})

test_that("pv_recursive returns empty for non-recursive profile", {
  # Create a profile with no recursion
  prof <- data.frame(
    time = c(1L, 1L, 2L, 2L, 3L),
    depth = c(1L, 2L, 1L, 2L, 1L),
    label = c("a", "b", "a", "c", "a"),
    filename = rep(NA_character_, 5),
    linenum = rep(NA_integer_, 5),
    filenum = rep(NA_integer_, 5),
    memalloc = c(100, 150, 150, 200, 200),
    meminc = c(0, 50, 0, 50, 0),
    stringsAsFactors = FALSE
  )

  p <- mock_profvis(prof = prof, files = list())
  result <- pv_recursive(p)

  expect_equal(nrow(result), 0)
})

test_that("pv_recursive max_depth is correct", {
  p <- mock_profvis_recursive()
  result <- pv_recursive(p)

  recurse_row <- result[result$label == "recurse", ]
  # In mock_profvis_recursive, max depth of recursion is 5 (time 3)
  expect_equal(recurse_row$max_depth, 5)
})

test_that("pv_print_recursive snapshot for recursive profile", {
  p <- mock_profvis_recursive()
  expect_snapshot(pv_print_recursive(p))
})

test_that("pv_print_recursive handles non-recursive profile", {
  # Create a profile with no recursion
  prof <- data.frame(
    time = c(1L, 1L, 2L, 2L, 3L),
    depth = c(1L, 2L, 1L, 2L, 1L),
    label = c("a", "b", "a", "c", "a"),
    filename = rep(NA_character_, 5),
    linenum = rep(NA_integer_, 5),
    filenum = rep(NA_integer_, 5),
    memalloc = c(100, 150, 150, 200, 200),
    meminc = c(0, 50, 0, 50, 0),
    stringsAsFactors = FALSE
  )

  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_print_recursive(p))
})

test_that("pv_recursive rejects non-profvis input", {
  expect_error(pv_recursive(list()), "must be a profvis object")
  expect_error(pv_print_recursive("bad"), "must be a profvis object")
})

# Tests for pv_debrief and pv_print_debrief

test_that("pv_debrief returns correct structure", {
  p <- mock_profvis()
  result <- pv_debrief(p)

  expect_type(result, "list")
  expect_named(
    result,
    c(
      "total_time_ms",
      "total_samples",
      "interval_ms",
      "has_source",
      "self_time",
      "total_time",
      "hot_lines",
      "hot_paths",
      "suggestions",
      "gc_pressure",
      "memory",
      "memory_lines"
    )
  )

  expect_type(result$total_time_ms, "double")
  expect_type(result$total_samples, "integer")
  expect_type(result$has_source, "logical")
  expect_s3_class(result$self_time, "data.frame")
  expect_s3_class(result$total_time, "data.frame")
  expect_s3_class(result$hot_paths, "data.frame")
})

test_that("pv_debrief respects n parameter", {
  p <- mock_profvis()
  result <- pv_debrief(p, n = 2)

  expect_true(nrow(result$self_time) <= 2)
  expect_true(nrow(result$total_time) <= 2)
})

test_that("pv_debrief rejects non-profvis input", {
  expect_error(pv_debrief(list()), "must be a profvis object")
  expect_error(pv_debrief("not profvis"), "must be a profvis object")
})

test_that("pv_print_debrief snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_debrief(p))
})

test_that("pv_print_debrief handles no source refs", {
  p <- mock_profvis_no_source()
  expect_snapshot(pv_print_debrief(p))
})

test_that("pv_print_debrief respects n parameters", {
  p <- mock_profvis()

  expect_no_error(capture.output(pv_print_debrief(
    p,
    n_functions = 2,
    n_lines = 2,
    n_paths = 2,
    n_memory = 2
  )))
})

test_that("pv_print_debrief returns pv_debrief invisibly", {
  p <- mock_profvis()

  result <- expect_invisible(pv_print_debrief(p))
  expect_s3_class(result, "debrief_debrief")
  expect_true("total_time_ms" %in% names(result))
  expect_true("self_time" %in% names(result))
})

test_that("pv_print_debrief rejects non-profvis input", {
  expect_error(pv_print_debrief(list()), "must be a profvis object")
  expect_error(pv_print_debrief("not profvis"), "must be a profvis object")
})

# Tests for pv_worst_line

test_that("pv_worst_line returns correct structure", {
  p <- mock_profvis()
  result <- pv_worst_line(p)

  expect_type(result, "list")
  expect_named(
    result,
    c(
      "location",
      "label",
      "filename",
      "linenum",
      "time_ms",
      "pct",
      "code",
      "context",
      "callers"
    )
  )

  expect_type(result$location, "character")
  expect_type(result$time_ms, "double")
  expect_type(result$pct, "double")
  expect_type(result$context, "character")
  expect_s3_class(result$callers, "data.frame")
})

test_that("pv_worst_line returns NULL when no source refs", {
  p <- mock_profvis_no_source()
  result <- pv_worst_line(p)

  expect_null(result)
})

test_that("pv_worst_line respects context parameter", {
  p <- mock_profvis()
  result <- pv_worst_line(p, context = 2)

  # Context should be at most 2*context + 1 lines
  expect_true(length(result$context) <= 5)
})

test_that("pv_worst_line rejects non-profvis input", {
  expect_error(pv_worst_line(list()), "must be a profvis object")
})

#' Example profvis data
#'
#' Creates an example profvis object for use in examples and testing.
#' This avoids the need to run actual profiling code in examples.
#'
#' @param type Type of example data to create:
#'   - `"default"`: A real profile captured from example code with source refs
#'   - `"no_source"`: A synthetic profile without source references
#'   - `"recursive"`: A synthetic profile with recursive function calls
#'   - `"gc"`: A synthetic profile with garbage collection pressure
#'
#' @return A profvis object that can be used with all debrief functions.
#'
#' @examples
#' # Get default example data
#' p <- pv_example()
#' pv_self_time(p)
#'
#' # Get example with recursive calls
#' p_recursive <- pv_example("recursive")
#' pv_recursive(p_recursive)
#'
#' @export
pv_example <- function(type = c("default", "no_source", "recursive", "gc")) {
  type <- match.arg(type)

  switch(
    type,
    default = example_default(),
    no_source = example_no_source(),
    recursive = example_recursive(),
    gc = example_gc()
  )
}

example_default <- function() {
  readRDS(system.file("extdata/example_profile.rds", package = "debrief"))
}

example_no_source <- function() {
  prof <- data.frame(
    time = c(1L, 1L, 2L, 2L, 3L),
    depth = c(1L, 2L, 1L, 2L, 1L),
    label = c("foo", "bar", "foo", "baz", "foo"),
    filename = rep(NA_character_, 5),
    linenum = rep(NA_real_, 5),
    filenum = rep(NA_real_, 5),
    memalloc = c(100, 150, 150, 200, 200),
    meminc = c(0, 50, 0, 50, 0)
  )

  structure(
    list(
      x = list(
        message = list(
          prof = prof,
          interval = 10,
          files = list()
        )
      )
    ),
    class = "profvis"
  )
}

example_recursive <- function() {
  prof <- data.frame(
    time = c(1L, 1L, 1L, 2L, 2L, 2L, 2L, 3L, 3L, 3L, 3L, 3L),
    depth = c(1L, 2L, 3L, 1L, 2L, 3L, 4L, 1L, 2L, 3L, 4L, 5L),
    label = rep("recurse", 12),
    filename = rep("R/recursive.R", 12),
    linenum = rep(5, 12),
    filenum = rep(1, 12),
    memalloc = seq(100, 1200, by = 100),
    meminc = rep(100, 12)
  )

  files <- list(
    list(
      filename = "R/recursive.R",
      content = paste(
        "recurse <- function(n) {",
        "  if (n <= 0) return(1)",
        "  recurse(n - 1)",
        "}",
        "  recurse(n - 1)",
        sep = "\n"
      ),
      normpath = "R/recursive.R"
    )
  )

  structure(
    list(
      x = list(
        message = list(
          prof = prof,
          interval = 10,
          files = files
        )
      )
    ),
    class = "profvis"
  )
}

example_gc <- function() {
  prof <- data.frame(
    time = c(1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L, 9L, 10L),
    depth = rep(1L, 10),
    label = c(
      "work",
      "work",
      "<GC>",
      "work",
      "<GC>",
      "<GC>",
      "work",
      "<GC>",
      "work",
      "work"
    ),
    filename = c(
      rep("R/work.R", 2),
      NA,
      "R/work.R",
      NA,
      NA,
      "R/work.R",
      NA,
      rep("R/work.R", 2)
    ),
    linenum = as.double(c(5, 5, NA, 5, NA, NA, 5, NA, 5, 5)),
    filenum = as.double(c(1, 1, NA, 1, NA, NA, 1, NA, 1, 1)),
    memalloc = c(100, 200, 200, 300, 300, 300, 400, 400, 500, 600),
    meminc = c(0, 100, 0, 100, 0, 0, 100, 0, 100, 100)
  )

  files <- list(
    list(
      filename = "R/work.R",
      content = "work <- function() {\n  x <- rnorm(1e6)\n  y <- cumsum(x)\n  z <- paste(x)\n  z\n}",
      normpath = "R/work.R"
    )
  )

  structure(
    list(
      x = list(
        message = list(
          prof = prof,
          interval = 10,
          files = files
        )
      )
    ),
    class = "profvis"
  )
}

#' File-level time summary
#'
#' Aggregates profiling time by source file. Requires source references
#' (use `devtools::load_all()` for best results).
#'
#' @param x A profvis object.
#'
#' @return A data frame with columns:
#'   - `filename`: Source file path
#'   - `samples`: Number of profiling samples
#'   - `time_ms`: Time in milliseconds
#'   - `pct`: Percentage of total time
#'
#' @examples
#' p <- pv_example()
#' pv_file_summary(p)
#' @export
pv_file_summary <- function(x) {
  pd <- extract_profile_data(x)

  # Filter to rows with source info
  with_source <- pd$prof[!is.na(pd$prof$filename), ]
  if (nrow(with_source) == 0) {
    return(data.frame(
      filename = character(),
      samples = integer(),
      time_ms = numeric(),
      pct = numeric()
    ))
  }

  # Count unique time-filename combinations (total time per file)
  unique_pairs <- unique(with_source[, c("time", "filename")])
  counts <- table(unique_pairs$filename)

  result <- data.frame(
    filename = names(counts),
    samples = as.integer(counts)
  )
  result$time_ms <- result$samples * pd$interval_ms
  result$pct <- round(100 * result$samples / pd$total_samples, 1)
  result <- result[order(-result$samples), ]
  rownames(result) <- NULL
  result
}

#' Print file summary
#'
#' @param x A profvis object.
#'
#' @return Invisibly returns a `debrief_file_summary` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_file_summary(p)
#'
#' @export
pv_print_file_summary <- function(x) {
  check_profvis(x)
  check_empty_profile(x)

  summary_df <- pv_file_summary(x)
  obj <- structure(
    list(summary_df = summary_df),
    class = "debrief_file_summary"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_file_summary <- function(x, ...) {
  summary_df <- x$summary_df

  if (nrow(summary_df) == 0) {
    cat("No source location data available.\n")
    cat("Use devtools::load_all() to enable source references.\n")
    cat_help_hint()
    return(invisible(x))
  }

  cat_header("FILE SUMMARY")
  cat("\n")

  for (i in seq_len(nrow(summary_df))) {
    row <- summary_df[i, ]
    cat(sprintf(
      "%6.0f ms (%5.1f%%)  %s\n",
      row$time_ms,
      row$pct,
      row$filename
    ))
  }

  # Next steps
  if (nrow(summary_df) > 0) {
    top_file <- summary_df$filename[1]
    cat_next_steps(c(
      sprintf("pv_source_context(p, \"%s\")", top_file),
      "pv_hot_lines(p)"
    ))
  }

  invisible(x)
}

#' Export profiling results as JSON
#'
#' Exports profiling analysis results in JSON format for consumption by
#' AI agents, automated tools, or external applications.
#'
#' @param x A profvis object.
#' @param file Optional file path to write JSON to. If `NULL`, returns the
#'   JSON string.
#' @param pretty If `TRUE`, formats JSON with indentation for readability.
#' @param include Character vector specifying which analyses to include.
#'   Options: "summary", "self_time", "total_time", "hot_lines", "memory",
#'   "callers", "gc_pressure", "suggestions", "recursive". Default includes all.
#' @param system_info If `TRUE`, includes R version and platform info in
#'   metadata. Useful for reproducibility.
#'
#' @return If `file` is `NULL`, returns a JSON string. Otherwise writes to file
#'   and returns the file path invisibly.
#'
#' @examples
#' p <- pv_example()
#' json <- pv_to_json(p)
#' cat(json)
#'
#' # Include only specific analyses
#' json <- pv_to_json(p, include = c("self_time", "hot_lines"))
#'
#' # Include system info for reproducibility
#' json <- pv_to_json(p, system_info = TRUE)
#'
#' @export
pv_to_json <- function(
  x,
  file = NULL,
  pretty = TRUE,
  include = c(
    "summary",
    "self_time",
    "total_time",
    "hot_lines",
    "memory",
    "gc_pressure",
    "suggestions",
    "recursive"
  ),
  system_info = FALSE
) {
  check_profvis(x)
  check_empty_profile(x)

  include <- match.arg(
    include,
    c(
      "summary",
      "self_time",
      "total_time",
      "hot_lines",
      "memory",
      "gc_pressure",
      "suggestions",
      "recursive"
    ),
    several.ok = TRUE
  )

  result <- list()

  # Basic metadata
  prof <- extract_prof(x)
  interval_ms <- extract_interval(x)
  total_samples <- extract_total_samples(x)

  result$metadata <- list(
    total_time_ms = total_samples * interval_ms,
    total_samples = total_samples,
    interval_ms = interval_ms,
    has_source_refs = has_source_refs(x),
    exported_at = format(Sys.time(), "%Y-%m-%dT%H:%M:%S%z")
  )

  # Add system info if requested
  if (system_info) {
    r_info <- R.Version()
    result$metadata$system <- list(
      r_version = paste(r_info$major, r_info$minor, sep = "."),
      platform = r_info$platform,
      os = r_info$os,
      arch = r_info$arch
    )
  }

  # Include requested analyses
  if ("summary" %in% include) {
    result$summary <- list(
      total_time_ms = total_samples * interval_ms,
      unique_functions = length(unique(prof$label)),
      max_depth = max(prof$depth)
    )
  }

  if ("self_time" %in% include) {
    result$self_time <- df_to_list(pv_self_time(x))
  }

  if ("total_time" %in% include) {
    result$total_time <- df_to_list(pv_total_time(x))
  }

  if ("hot_lines" %in% include) {
    hot_lines <- pv_hot_lines(x)
    if (nrow(hot_lines) > 0) {
      result$hot_lines <- df_to_list(hot_lines)
    } else {
      result$hot_lines <- list()
    }
  }

  if ("memory" %in% include) {
    result$memory <- list(
      by_function = df_to_list(pv_memory(x)),
      by_line = df_to_list(pv_memory_lines(x))
    )
  }

  if ("gc_pressure" %in% include) {
    patterns <- pv_gc_pressure(x)
    if (nrow(patterns) > 0) {
      result$gc_pressure <- df_to_list(patterns)
    } else {
      result$gc_pressure <- list()
    }
  }

  if ("suggestions" %in% include) {
    suggestions <- pv_suggestions(x)
    if (nrow(suggestions) > 0) {
      result$suggestions <- df_to_list(suggestions)
    } else {
      result$suggestions <- list()
    }
  }

  if ("recursive" %in% include) {
    recursive <- pv_recursive(x)
    if (nrow(recursive) > 0) {
      result$recursive <- df_to_list(recursive)
    } else {
      result$recursive <- list()
    }
  }

  # Convert to JSON
  json <- to_json(result, pretty = pretty)

  if (!is.null(file)) {
    writeLines(json, file)
    invisible(file)
  } else {
    json
  }
}

#' Export profiling results as a list
#'
#' Returns all profiling analysis results as a nested R list, useful for
#' programmatic access to results without JSON serialization.
#'
#' @param x A profvis object.
#' @param include Character vector specifying which analyses to include.
#'   Same options as [pv_to_json()].
#' @param system_info If `TRUE`, includes R version and platform info in
#'   metadata.
#'
#' @return A named list containing the requested analyses.
#'
#' @examples
#' p <- pv_example()
#' results <- pv_to_list(p)
#' names(results)
#' results$self_time
#'
#' @export
pv_to_list <- function(
  x,
  include = c(
    "summary",
    "self_time",
    "total_time",
    "hot_lines",
    "memory",
    "gc_pressure",
    "suggestions",
    "recursive"
  ),
  system_info = FALSE
) {
  check_profvis(x)
  check_empty_profile(x)

  include <- match.arg(
    include,
    c(
      "summary",
      "self_time",
      "total_time",
      "hot_lines",
      "memory",
      "gc_pressure",
      "suggestions",
      "recursive"
    ),
    several.ok = TRUE
  )

  result <- list()

  prof <- extract_prof(x)
  interval_ms <- extract_interval(x)
  total_samples <- extract_total_samples(x)

  result$metadata <- list(
    total_time_ms = total_samples * interval_ms,
    total_samples = total_samples,
    interval_ms = interval_ms,
    has_source_refs = has_source_refs(x),
    exported_at = Sys.time()
  )

  if (system_info) {
    r_info <- R.Version()
    result$metadata$system <- list(
      r_version = paste(r_info$major, r_info$minor, sep = "."),
      platform = r_info$platform,
      os = r_info$os,
      arch = r_info$arch
    )
  }

  if ("summary" %in% include) {
    result$summary <- list(
      total_time_ms = total_samples * interval_ms,
      unique_functions = length(unique(prof$label)),
      max_depth = max(prof$depth)
    )
  }

  if ("self_time" %in% include) {
    result$self_time <- pv_self_time(x)
  }

  if ("total_time" %in% include) {
    result$total_time <- pv_total_time(x)
  }

  if ("hot_lines" %in% include) {
    result$hot_lines <- pv_hot_lines(x)
  }

  if ("memory" %in% include) {
    result$memory <- list(
      by_function = pv_memory(x),
      by_line = pv_memory_lines(x)
    )
  }

  if ("gc_pressure" %in% include) {
    result$gc_pressure <- pv_gc_pressure(x)
  }

  if ("suggestions" %in% include) {
    result$suggestions <- pv_suggestions(x)
  }

  if ("recursive" %in% include) {
    result$recursive <- pv_recursive(x)
  }

  result
}

# Convert data frame to list of rows for JSON serialization
df_to_list <- function(df) {
  if (is.null(df) || nrow(df) == 0) {
    return(list())
  }
  lapply(seq_len(nrow(df)), function(i) as.list(df[i, , drop = FALSE]))
}

# Simple JSON serialization without external dependencies
to_json <- function(x, pretty = TRUE, indent = 0) {
  indent_str <- if (pretty) strrep("  ", indent) else ""
  newline <- if (pretty) "\n" else ""

  if (is.null(x)) {
    return("null")
  }

  if (is.logical(x) && length(x) == 1) {
    return(
      if (is.na(x)) {
        "null"
      } else if (x) {
        "true"
      } else {
        "false"
      }
    )
  }

  if (is.atomic(x) && length(x) == 1) {
    if (is.na(x)) {
      return("null")
    }
    if (is.character(x)) {
      # Escape special characters
      x <- gsub("\\\\", "\\\\\\\\", x)
      x <- gsub("\"", "\\\\\"", x)
      x <- gsub("\n", "\\\\n", x)
      x <- gsub("\r", "\\\\r", x)
      x <- gsub("\t", "\\\\t", x)
      return(sprintf("\"%s\"", x))
    }
    return(as.character(x))
  }

  if (is.atomic(x) && length(x) > 1) {
    # Array of primitives
    elements <- vapply(
      x,
      function(el) to_json(el, pretty = FALSE),
      character(1)
    )
    return(sprintf("[%s]", paste(elements, collapse = ", ")))
  }

  if (is.data.frame(x)) {
    return(to_json(df_to_list(x), pretty = pretty, indent = indent))
  }

  if (is.list(x)) {
    if (length(x) == 0) {
      if (!is.null(names(x))) {
        return("{}")
      }
      return("[]")
    }

    if (is.null(names(x))) {
      # Array
      elements <- vapply(
        x,
        function(el) {
          paste0(
            if (pretty) strrep("  ", indent + 1) else "",
            to_json(el, pretty = pretty, indent = indent + 1)
          )
        },
        character(1)
      )
      return(sprintf(
        "[%s%s%s%s]",
        newline,
        paste(elements, collapse = paste0(",", newline)),
        newline,
        indent_str
      ))
    } else {
      # Object
      keys <- names(x)
      pairs <- vapply(
        seq_along(x),
        function(i) {
          key <- keys[i]
          val <- to_json(x[[i]], pretty = pretty, indent = indent + 1)
          sprintf(
            "%s\"%s\": %s",
            if (pretty) strrep("  ", indent + 1) else "",
            key,
            val
          )
        },
        character(1)
      )
      return(sprintf(
        "{%s%s%s%s}",
        newline,
        paste(pairs, collapse = paste0(",", newline)),
        newline,
        indent_str
      ))
    }
  }

  # Fallback
  "null"
}

empty_prof <- structure(
  list(
    x = list(
      message = list(
        prof = data.frame(
          time = integer(),
          depth = integer(),
          label = character(),
          filename = character(),
          linenum = numeric(),
          filenum = numeric(),
          memalloc = numeric(),
          meminc = numeric(),
          stringsAsFactors = FALSE
        ),
        interval = 10,
        files = list()
      )
    )
  ),
  class = "profvis"
)

test_that("empty profile errors with helpful message", {
  expect_snapshot(pv_self_time(empty_prof), error = TRUE)
  expect_snapshot(pv_total_time(empty_prof), error = TRUE)
  expect_snapshot(pv_debrief(empty_prof), error = TRUE)
  expect_snapshot(pv_print_debrief(empty_prof), error = TRUE)
  expect_snapshot(pv_worst_line(empty_prof), error = TRUE)
  expect_snapshot(pv_hot_lines(empty_prof), error = TRUE)
  expect_snapshot(pv_hot_paths(empty_prof), error = TRUE)
  expect_snapshot(pv_flame(empty_prof), error = TRUE)
  expect_snapshot(pv_memory(empty_prof), error = TRUE)
  expect_snapshot(pv_callers(empty_prof, "foo"), error = TRUE)
  expect_snapshot(pv_callees(empty_prof, "foo"), error = TRUE)
  expect_snapshot(pv_focus(empty_prof, "foo"), error = TRUE)
  expect_snapshot(pv_suggestions(empty_prof), error = TRUE)
  expect_snapshot(pv_gc_pressure(empty_prof), error = TRUE)
  expect_snapshot(pv_recursive(empty_prof), error = TRUE)
  expect_snapshot(pv_call_stats(empty_prof), error = TRUE)
  expect_snapshot(pv_call_depth(empty_prof), error = TRUE)
  expect_snapshot(pv_file_summary(empty_prof), error = TRUE)
  expect_snapshot(pv_source_context(empty_prof, "foo.R"), error = TRUE)
  expect_snapshot(pv_to_json(empty_prof), error = TRUE)
  expect_snapshot(pv_to_list(empty_prof), error = TRUE)
})

test_that("compare functions error on empty profile", {
  valid_prof <- pv_example()

  expect_snapshot(pv_compare(empty_prof, valid_prof), error = TRUE)
  expect_snapshot(pv_compare(valid_prof, empty_prof), error = TRUE)
  expect_snapshot(pv_compare_many(a = empty_prof, b = valid_prof), error = TRUE)
})

#' Text-based flame graph
#'
#' Generates an ASCII representation of a flame graph showing the hierarchical
#' breakdown of time spent in the call tree.
#'
#' @param x A profvis object.
#' @param width Width of the flame graph in characters.
#' @param min_pct Minimum percentage to show (filters small slices).
#' @param max_depth Maximum depth to display.
#'
#' @return Invisibly returns a `debrief_flame` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_flame(p)
#' @export
pv_flame <- function(x, width = 70, min_pct = 2, max_depth = 15) {
  pd <- extract_profile_data(x)
  total_time <- pd$total_samples * pd$interval_ms
  tree <- build_flame_tree_v2(pd$prof, pd$total_samples)

  obj <- structure(
    list(
      tree = tree,
      total_time = total_time,
      width = width,
      min_pct = min_pct,
      max_depth = max_depth,
      total_samples = pd$total_samples
    ),
    class = "debrief_flame"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_flame <- function(x, ...) {
  cat_header("FLAME GRAPH (text)")
  cat("\n")
  cat(sprintf(
    "Total time: %.0f ms | Width: %d chars | Min: %.0f%%\n\n",
    x$total_time,
    x$width,
    x$min_pct
  ))

  print_flame_tree(x$tree, x$width, x$min_pct, x$total_samples, x$max_depth)

  cat("\nLegend: [====] = time spent, width proportional to time\n")

  # Next steps - find the top function from level 1
  tree <- x$tree
  if (length(tree) >= 1 && length(tree[[1]]) > 0) {
    items <- tree[[1]][order(-sapply(tree[[1]], function(item) item$samples))]
    top_func <- items[[1]]$name
    suggestions <- "pv_hot_paths(p)"
    if (is_user_function(top_func)) {
      suggestions <- c(sprintf("pv_focus(p, \"%s\")", top_func), suggestions)
    }
    cat_next_steps(suggestions)
  }

  invisible(x)
}

build_flame_tree_v2 <- function(prof, total_samples) {
  # Build tree by aggregating paths
  times <- unique(prof$time)

  # Get stack for each time point
  stacks <- lapply(times, function(t) {
    stack <- prof[prof$time == t, ]
    stack <- stack[order(stack$depth), ]
    stack$label
  })

  # Find max depth and pre-allocate
  max_depth <- max(sapply(stacks, length))
  tree <- vector("list", max_depth)
  for (i in seq_len(max_depth)) {
    tree[[i]] <- list()
  }

  # Count path prefixes at each level
  for (stack in stacks) {
    for (depth in seq_along(stack)) {
      path <- paste(stack[1:depth], collapse = "|")
      if (is.null(tree[[depth]][[path]])) {
        tree[[depth]][[path]] <- list(
          name = stack[depth],
          samples = 0,
          depth = depth,
          parent_path = if (depth > 1) {
            paste(stack[1:(depth - 1)], collapse = "|")
          } else {
            ""
          }
        )
      }
      tree[[depth]][[path]]$samples <- tree[[depth]][[path]]$samples + 1
    }
  }

  tree
}

print_flame_tree <- function(tree, width, min_pct, total_samples, max_depth) {
  # Print root bar
  bar <- strrep("=", width)
  cat(sprintf("[%s] (root) 100%%\n", bar))

  if (length(tree) == 0) {
    return()
  }

  # Print each level
  for (depth in seq_len(min(length(tree), max_depth))) {
    level <- tree[[depth]]
    if (is.null(level)) {
      next
    }

    # Sort by samples descending
    items <- level[order(-sapply(level, function(x) x$samples))]

    for (item in items) {
      pct <- 100 * item$samples / total_samples
      if (pct < min_pct) {
        next
      }

      bar_width <- max(1, round(width * item$samples / total_samples))
      bar <- strrep("=", bar_width)
      padding <- strrep(" ", width - bar_width)
      indent <- strrep("  ", item$depth)

      name <- truncate_string(item$name, 40)
      cat(sprintf("[%s%s]%s %s (%.1f%%)\n", bar, padding, indent, name, pct))
    }
  }
}

#' Condensed flame graph
#'
#' Shows a simplified, condensed view of the flame graph focusing on the
#' hottest paths.
#'
#' @param x A profvis object.
#' @param n Number of hot paths to show.
#' @param width Width of bars.
#'
#' @return Invisibly returns a `debrief_flame_condense` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_flame_condense(p)
#' @export
pv_flame_condense <- function(x, n = 10, width = 50) {
  pd <- extract_profile_data(x)

  # Get unique paths and their frequencies
  times <- unique(pd$prof$time)

  path_counts <- list()
  for (t in times) {
    stack <- pd$prof[pd$prof$time == t, ]
    stack <- stack[order(stack$depth), ]
    path <- paste(stack$label, collapse = " > ")

    if (is.null(path_counts[[path]])) {
      path_counts[[path]] <- 0
    }
    path_counts[[path]] <- path_counts[[path]] + 1
  }

  # Convert to data frame and sort
  paths_df <- data.frame(
    path = names(path_counts),
    samples = unlist(path_counts)
  )
  paths_df <- paths_df[order(-paths_df$samples), ]
  paths_df$pct <- round(100 * paths_df$samples / pd$total_samples, 1)

  obj <- structure(
    list(
      paths_df = paths_df,
      n = n,
      width = width,
      total_samples = pd$total_samples
    ),
    class = "debrief_flame_condense"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_flame_condense <- function(x, ...) {
  paths_df <- x$paths_df
  n <- x$n
  width <- x$width
  total_samples <- x$total_samples

  cat_header("CONDENSED FLAME VIEW")
  cat("\n")

  # Show top n paths
  for (i in seq_len(min(n, nrow(paths_df)))) {
    row <- paths_df[i, ]
    bar_width <- max(1, round(width * row$samples / total_samples))
    bar <- strrep("#", bar_width)

    cat(sprintf("\n%s %.1f%% (%d samples)\n", bar, row$pct, row$samples))

    # Show path vertically
    parts <- strsplit(row$path, " > ")[[1]]
    for (j in seq_along(parts)) {
      indent <- strrep("  ", j - 1)
      cat(sprintf("%s-> %s\n", indent, parts[j]))
    }
  }

  # Next steps - suggest focusing on the leaf of the hottest path
  if (nrow(paths_df) > 0) {
    parts <- strsplit(paths_df$path[1], " > ")[[1]]
    leaf_func <- parts[length(parts)]
    suggestions <- "pv_hot_lines(p)"
    if (is_user_function(leaf_func)) {
      suggestions <- c(sprintf("pv_focus(p, \"%s\")", leaf_func), suggestions)
    }
    cat_next_steps(suggestions)
  }

  invisible(x)
}

test_that("pv_flame snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_flame(p))
})

test_that("pv_flame respects width parameter", {
  p <- mock_profvis()

  # Different widths should work
  expect_snapshot(pv_flame(p, width = 40))
})

test_that("pv_flame respects min_pct parameter", {
  p <- mock_profvis()
  expect_snapshot(pv_flame(p, min_pct = 30))
})

test_that("pv_flame respects max_depth parameter", {
  p <- mock_profvis()

  # Should not error with different max_depth values
  expect_no_error(capture.output(pv_flame(p, max_depth = 2)))
  expect_no_error(capture.output(pv_flame(p, max_depth = 20)))
})

test_that("pv_flame returns invisibly", {
  p <- mock_profvis()

  result <- expect_invisible(pv_flame(p))
  expect_type(result, "list")
})

test_that("pv_flame_condense snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_flame_condense(p, n = 3))
})

test_that("pv_flame_condense respects n parameter", {
  p <- mock_profvis()

  # Should not error with different n values
  expect_no_error(capture.output(pv_flame_condense(p, n = 1)))
  expect_no_error(capture.output(pv_flame_condense(p, n = 10)))
})

test_that("pv_flame_condense respects width parameter", {
  p <- mock_profvis()
  expect_snapshot(pv_flame_condense(p, n = 2, width = 30))
})

test_that("pv_flame_condense returns data frame", {
  p <- mock_profvis()

  result <- expect_invisible(pv_flame_condense(p))
  expect_s3_class(result, "debrief_flame_condense")
  expect_true("path" %in% names(result$paths_df))
  expect_true("samples" %in% names(result$paths_df))
  expect_true("pct" %in% names(result$paths_df))
})

test_that("flame functions reject non-profvis input", {
  expect_error(pv_flame(list()), "must be a profvis object")
  expect_error(pv_flame_condense("bad"), "must be a profvis object")
})

#' Generate optimization suggestions
#'
#' Analyzes the profile and generates specific, actionable optimization
#' suggestions based on detected patterns and hotspots.
#'
#' @param x A profvis object.
#'
#' @return A data frame with columns:
#'   - `priority`: 1 (highest) to 5 (lowest)
#'   - `category`: Type of optimization (e.g., "data structure", "algorithm")
#'   - `location`: Where to apply the optimization
#'   - `action`: What to do
#'   - `pattern`: Code pattern to look for (or NA)
#'   - `replacement`: Suggested replacement (or NA)
#'   - `potential_impact`: Estimated time that could be saved
#'
#' @examples
#' p <- pv_example("gc")
#' pv_suggestions(p)
#' @export
pv_suggestions <- function(x) {
  pd <- extract_profile_data(x)
  total_time <- pd$total_samples * pd$interval_ms
  has_source <- has_source_refs(x)

  suggestions <- list()

  # Check for data frame subsetting in recursion
  suggestions <- c(
    suggestions,
    suggest_df_vectorization(
      pd$prof,
      pd$interval_ms,
      pd$total_samples,
      total_time
    )
  )

  # Check for recursive functions
  suggestions <- c(
    suggestions,
    suggest_recursion_optimization(
      pd$prof,
      pd$interval_ms,
      pd$total_samples,
      total_time
    )
  )

  # Check for GC pressure
  suggestions <- c(
    suggestions,
    suggest_gc_reduction(pd$prof, pd$interval_ms, pd$total_samples, total_time)
  )

  # Check for string operations
  suggestions <- c(
    suggestions,
    suggest_string_optimization(
      pd$prof,
      pd$interval_ms,
      pd$total_samples,
      total_time
    )
  )

  # Check for hot lines (if source available)
  if (has_source) {
    suggestions <- c(
      suggestions,
      suggest_hotline_optimization(
        x,
        pd$interval_ms,
        pd$total_samples,
        total_time
      )
    )
  }

  # Generic suggestions based on top functions
  suggestions <- c(
    suggestions,
    suggest_top_function_optimization(
      pd$prof,
      pd$interval_ms,
      pd$total_samples,
      total_time
    )
  )

  if (length(suggestions) == 0) {
    return(data.frame(
      priority = integer(),
      category = character(),
      location = character(),
      action = character(),
      pattern = character(),
      replacement = character(),
      potential_impact = character()
    ))
  }

  result <- do.call(rbind, suggestions)
  result <- result[order(result$priority), ]
  rownames(result) <- NULL

  # Remove duplicates
  result <- result[!duplicated(paste(result$action, result$location)), ]

  result
}

suggest_df_vectorization <- function(
  prof,
  interval_ms,
  total_samples,
  total_time
) {
  df_subset_times <- unique(prof$time[
    prof$label %in% c("[.data.frame", "[[.data.frame")
  ])
  df_time <- length(df_subset_times) * interval_ms
  df_pct <- 100 * df_time / total_time

  if (df_pct > 15) {
    list(data.frame(
      priority = 1L,
      category = "data structure",
      location = "[.data.frame / [[.data.frame",
      action = "Replace row subsetting with vector indexing",
      pattern = "df[i, ]$col",
      replacement = "df$col[i]",
      potential_impact = sprintf(
        "Up to %.0f ms (%.0f%%)",
        df_time * 0.8,
        df_pct * 0.8
      )
    ))
  } else {
    list()
  }
}

suggest_recursion_optimization <- function(
  prof,
  interval_ms,
  total_samples,
  total_time
) {
  # Find recursive functions
  times <- unique(prof$time)
  recursive_stats <- list()

  for (t in times) {
    stack <- prof[prof$time == t, ]
    func_counts <- table(stack$label)
    for (func in names(func_counts[func_counts > 1])) {
      if (is.null(recursive_stats[[func]])) {
        recursive_stats[[func]] <- 0
      }
      recursive_stats[[func]] <- recursive_stats[[func]] + 1
    }
  }

  suggestions <- list()
  for (func in names(recursive_stats)) {
    func_time <- recursive_stats[[func]] * interval_ms
    func_pct <- 100 * func_time / total_time

    if (func_pct > 20 && is_user_function(func)) {
      # Skip internal functions
      suggestions[[length(suggestions) + 1]] <- data.frame(
        priority = 2L,
        category = "algorithm",
        location = func,
        action = "Convert recursive to iterative",
        pattern = sprintf("recursive %s()", func),
        replacement = "stack/queue data structure",
        potential_impact = sprintf(
          "Potentially %.0f ms (%.0f%%)",
          func_time * 0.3,
          func_pct * 0.3
        )
      )
    }
  }

  suggestions
}

suggest_gc_reduction <- function(prof, interval_ms, total_samples, total_time) {
  gc_times <- unique(prof$time[prof$label == "<GC>"])
  gc_time <- length(gc_times) * interval_ms
  gc_pct <- 100 * gc_time / total_time

  if (gc_pct > 10) {
    list(data.frame(
      priority = 2L,
      category = "memory",
      location = "memory allocation hotspots",
      action = "Reduce memory allocation",
      pattern = "c(x, new), rbind(), growing vectors",
      replacement = "pre-allocate to final size",
      potential_impact = sprintf(
        "Up to %.0f ms (%.0f%%)",
        gc_time * 0.5,
        gc_pct * 0.5
      )
    ))
  } else {
    list()
  }
}

suggest_string_optimization <- function(
  prof,
  interval_ms,
  total_samples,
  total_time
) {
  string_funcs <- c(
    "paste",
    "paste0",
    "sprintf",
    "gsub",
    "sub",
    "grep",
    "grepl",
    "strsplit",
    "substring",
    "substr"
  )

  total_string_time <- 0
  hot_string_func <- NULL
  max_time <- 0

  for (func in string_funcs) {
    func_times <- unique(prof$time[prof$label == func])
    func_time <- length(func_times) * interval_ms
    total_string_time <- total_string_time + func_time
    if (func_time > max_time) {
      max_time <- func_time
      hot_string_func <- func
    }
  }

  string_pct <- 100 * total_string_time / total_time

  if (string_pct > 5 && !is.null(hot_string_func)) {
    list(data.frame(
      priority = 3L,
      category = "string operations",
      location = hot_string_func,
      action = sprintf("Optimize string operations (%.1f%%)", string_pct),
      pattern = "string ops in loops, regex without fixed=TRUE",
      replacement = "pre-compute, fixed=TRUE, stringi package",
      potential_impact = sprintf(
        "Up to %.0f ms (%.0f%%)",
        total_string_time * 0.5,
        string_pct * 0.5
      )
    ))
  } else {
    list()
  }
}

suggest_hotline_optimization <- function(
  x,
  interval_ms,
  total_samples,
  total_time
) {
  hot_lines <- pv_hot_lines(x, n = 3)

  if (nrow(hot_lines) == 0) {
    return(list())
  }

  suggestions <- list()
  for (i in seq_len(nrow(hot_lines))) {
    row <- hot_lines[i, ]
    if (row$pct > 5) {
      suggestions[[length(suggestions) + 1]] <- data.frame(
        priority = 1L,
        category = "hot line",
        location = row$location,
        action = sprintf("Optimize hot line (%.1f%%)", row$pct),
        pattern = truncate_string(row$label, 50),
        replacement = NA_character_,
        potential_impact = sprintf("%.0f ms (%.1f%%)", row$time_ms, row$pct)
      )
    }
  }

  suggestions
}

suggest_top_function_optimization <- function(
  prof,
  interval_ms,
  total_samples,
  total_time
) {
  # Get self-time (deepest frame for each time point)
  top_of_stack <- extract_top_of_stack(prof)

  counts <- table(top_of_stack$label)
  top_func <- names(sort(counts, decreasing = TRUE))[1]
  top_time <- as.integer(counts[top_func]) * interval_ms
  top_pct <- 100 * top_time / total_time

  # Only suggest if it's not an internal R function
  if (top_pct > 10 && is_user_function(top_func)) {
    list(data.frame(
      priority = 2L,
      category = "hot function",
      location = top_func,
      action = sprintf("Profile in isolation (%.1f%% self-time)", top_pct),
      pattern = top_func,
      replacement = NA_character_,
      potential_impact = sprintf("%.0f ms (%.1f%%)", top_time, top_pct)
    ))
  } else {
    list()
  }
}

#' Print optimization suggestions
#'
#' @param x A profvis object.
#'
#' @return Invisibly returns a `debrief_suggestions` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example("gc")
#' pv_print_suggestions(p)
#'
#' @export
pv_print_suggestions <- function(x) {
  check_profvis(x)
  check_empty_profile(x)

  suggestions <- pv_suggestions(x)
  obj <- structure(
    list(suggestions = suggestions),
    class = "debrief_suggestions"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_suggestions <- function(x, ...) {
  suggestions <- x$suggestions

  cat_header("OPTIMIZATION SUGGESTIONS")
  cat("\n")

  if (nrow(suggestions) == 0) {
    cat("No suggestions.\n")
    cat_help_hint()
    return(invisible(x))
  }

  current_priority <- 0
  for (i in seq_len(nrow(suggestions))) {
    row <- suggestions[i, ]

    if (row$priority != current_priority) {
      current_priority <- row$priority
      cat(sprintf("### Priority %d\n\n", current_priority))
    }

    cat(sprintf("category: %s\n", row$category))
    cat(sprintf("location: %s\n", row$location))
    cat(sprintf("action: %s\n", row$action))
    if (!is.na(row$pattern)) {
      cat(sprintf("pattern: %s\n", row$pattern))
    }
    if (!is.na(row$replacement)) {
      cat(sprintf("replacement: %s\n", row$replacement))
    }
    cat(sprintf("potential_impact: %s\n\n", row$potential_impact))
  }

  # Next steps suggestions based on top suggestion
  if (nrow(suggestions) > 0) {
    top_loc <- suggestions$location[1]
    # Check if location looks like a function name (not a file:line or internal)
    if (
      !grepl(":", top_loc) && !grepl(" ", top_loc) && is_user_function(top_loc)
    ) {
      cat_next_steps(c(
        sprintf("pv_focus(p, \"%s\")", top_loc),
        "pv_gc_pressure(p)"
      ))
    } else {
      cat_next_steps(c("pv_hot_lines(p)", "pv_gc_pressure(p)"))
    }
  }

  invisible(x)
}

#' Detect recursive functions
#'
#' Identifies functions that call themselves (directly recursive) or appear
#' multiple times in the same call stack. Recursive functions in hot paths
#' are often optimization targets.
#'
#' @param x A profvis object.
#'
#' @return A data frame with columns:
#'   - `label`: Function name
#'   - `max_depth`: Maximum recursion depth observed
#'   - `avg_depth`: Average recursion depth when recursive
#'
#'   - `recursive_samples`: Number of samples where function appears multiple times
#'   - `total_samples`: Total samples where function appears
#'
#'   - `pct_recursive`: Percentage of appearances that are recursive
#'   - `total_ms`: Total time on call stack
#'   - `pct_time`: Percentage of total profile time
#'
#' @examples
#' p <- pv_example("recursive")
#' pv_recursive(p)
#' @export
pv_recursive <- function(x) {
  pd <- extract_profile_data(x)

  # For each time point, find functions that appear multiple times
  times <- unique(pd$prof$time)

  recursive_info <- lapply(times, function(t) {
    stack <- pd$prof[pd$prof$time == t, ]
    func_counts <- table(stack$label)
    recursive_funcs <- names(func_counts[func_counts > 1])

    if (length(recursive_funcs) == 0) {
      return(NULL)
    }

    lapply(recursive_funcs, function(func) {
      depths <- stack$depth[stack$label == func]
      data.frame(
        label = func,
        time = t,
        count = length(depths),
        min_depth = min(depths),
        max_depth = max(depths)
      )
    })
  })

  recursive_info <- unlist(recursive_info, recursive = FALSE)
  if (length(recursive_info) == 0) {
    return(data.frame(
      label = character(),
      max_depth = integer(),
      avg_depth = numeric(),
      recursive_samples = integer(),
      total_samples = integer(),
      pct_recursive = numeric(),
      total_ms = numeric(),
      pct_time = numeric()
    ))
  }

  recursive_df <- do.call(rbind, recursive_info)

  # Aggregate by function
  funcs <- unique(recursive_df$label)
  result <- lapply(funcs, function(func) {
    func_recursive <- recursive_df[recursive_df$label == func, ]
    func_all <- pd$prof[pd$prof$label == func, ]

    recursive_samples <- nrow(func_recursive)
    total_func_samples <- length(unique(func_all$time))

    data.frame(
      label = func,
      max_depth = max(func_recursive$count),
      avg_depth = round(mean(func_recursive$count), 1),
      recursive_samples = recursive_samples,
      total_samples = total_func_samples,
      pct_recursive = round(100 * recursive_samples / total_func_samples, 1),
      total_ms = total_func_samples * pd$interval_ms,
      pct_time = round(100 * total_func_samples / pd$total_samples, 1)
    )
  })

  result <- do.call(rbind, result)
  result <- result[order(-result$total_ms), ]
  rownames(result) <- NULL
  result
}

#' Print recursive functions analysis
#'
#' @param x A profvis object.
#'
#' @return Invisibly returns a `debrief_recursive` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example("recursive")
#' pv_print_recursive(p)
#'
#' @export
pv_print_recursive <- function(x) {
  check_profvis(x)
  check_empty_profile(x)

  recursive <- pv_recursive(x)
  obj <- structure(list(recursive = recursive), class = "debrief_recursive")
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_recursive <- function(x, ...) {
  recursive <- x$recursive

  if (nrow(recursive) == 0) {
    cat("No recursive functions detected in the profile.\n")
    cat_help_hint()
    return(invisible(x))
  }

  cat_header("RECURSIVE FUNCTIONS")
  cat("\n")
  cat(sprintf(
    "%-30s %8s %8s %10s %8s\n",
    "Function",
    "MaxDepth",
    "AvgDepth",
    "Total ms",
    "Pct"
  ))

  for (i in seq_len(nrow(recursive))) {
    row <- recursive[i, ]
    label <- truncate_string(row$label, 30)
    cat(sprintf(
      "%-30s %8d %8.1f %10.0f %7.1f%%\n",
      label,
      row$max_depth,
      row$avg_depth,
      row$total_ms,
      row$pct_time
    ))
  }

  # Next steps
  if (nrow(recursive) > 0) {
    top_func <- recursive$label[1]
    suggestions <- "pv_suggestions(p)"
    if (is_user_function(top_func)) {
      suggestions <- c(sprintf("pv_focus(p, \"%s\")", top_func), suggestions)
    }
    cat_next_steps(suggestions)
  }

  invisible(x)
}

#' Self-time summary by function
#'
#' Returns the time spent directly in each function (at the top of the call
#' stack). This shows where CPU cycles are actually being consumed.
#'
#' @param x A profvis object.
#' @param n Maximum number of functions to return. If `NULL`, returns all that
#'   pass the filters.
#' @param min_pct Minimum percentage of total time to include (default 0).
#' @param min_time_ms Minimum time in milliseconds to include (default 0).
#'
#' @return A data frame with columns:
#'   - `label`: Function name
#'   - `samples`: Number of profiling samples
#'   - `time_ms`: Time in milliseconds
#'   - `pct`: Percentage of total time
#'
#' @examples
#' p <- pv_example()
#' pv_self_time(p)
#'
#' # Only functions with >= 5% self-time
#' pv_self_time(p, min_pct = 5)
#' @export
pv_self_time <- function(x, n = NULL, min_pct = 0, min_time_ms = 0) {
  pd <- extract_profile_data(x)
  top_of_stack <- extract_top_of_stack(pd$prof)
  counts <- table(top_of_stack$label)
  result <- build_count_result(counts, pd$interval_ms, pd$total_samples)
  apply_result_filters(result, n, min_pct, min_time_ms)
}

#' Total time summary by function
#'
#' Returns the time spent in each function including all its callees. This
#' shows which functions are on the call stack when time is being spent.
#'
#' @param x A profvis object.
#' @param n Maximum number of functions to return. If `NULL`, returns all that
#'   pass the filters.
#' @param min_pct Minimum percentage of total time to include (default 0).
#' @param min_time_ms Minimum time in milliseconds to include (default 0).
#'
#' @return A data frame with columns:
#'   - `label`: Function name
#'   - `samples`: Number of profiling samples where function appeared
#'   - `time_ms`: Time in milliseconds
#'   - `pct`: Percentage of total time
#'
#' @examples
#' p <- pv_example()
#' pv_total_time(p)
#'
#' # Only functions with >= 50% total time
#' pv_total_time(p, min_pct = 50)
#' @export
pv_total_time <- function(x, n = NULL, min_pct = 0, min_time_ms = 0) {
  pd <- extract_profile_data(x)
  unique_pairs <- unique(pd$prof[, c("time", "label")])
  counts <- table(unique_pairs$label)
  result <- build_count_result(counts, pd$interval_ms, pd$total_samples)
  apply_result_filters(result, n, min_pct, min_time_ms)
}

#' Hot source lines by self-time
#'
#' Returns the source lines where the most CPU time is spent. Requires source
#' references (use `devtools::load_all()` for best results).
#'
#' @param x A profvis object.
#' @param n Maximum number of lines to return. If `NULL`, returns all that pass
#'   the filters.
#' @param min_pct Minimum percentage of total time to include (default 0).
#' @param min_time_ms Minimum time in milliseconds to include (default 0).
#'
#' @return A data frame with columns:
#'   - `location`: File path and line number (e.g., "R/foo.R:42")
#'   - `label`: Function name at this location
#'   - `filename`: Source file path
#'   - `linenum`: Line number
#'   - `samples`: Number of profiling samples
#'   - `time_ms`: Time in milliseconds
#'   - `pct`: Percentage of total time
#'
#' @examples
#' p <- pv_example()
#' pv_hot_lines(p)
#'
#' # Only lines with >= 10% of time
#' pv_hot_lines(p, min_pct = 10)
#' @export
pv_hot_lines <- function(x, n = NULL, min_pct = 0, min_time_ms = 0) {
  pd <- extract_profile_data(x)
  top_of_stack <- extract_top_of_stack(pd$prof)

  # Filter to rows with source info
  with_source <- top_of_stack[!is.na(top_of_stack$filename), ]
  if (nrow(with_source) == 0) {
    return(empty_location_result())
  }

  # Create location key
  with_source$location <- paste0(with_source$filename, ":", with_source$linenum)

  # Aggregate by location
  counts <- table(with_source$location)
  result <- data.frame(
    location = names(counts),
    samples = as.integer(counts)
  )

  # Get label, filename, linenum for each location
  first_occurrence <- with_source[!duplicated(with_source$location), ]
  result <- merge(
    result,
    first_occurrence[, c("location", "label", "filename", "linenum")],
    by = "location"
  )

  result$time_ms <- result$samples * pd$interval_ms
  result$pct <- round(100 * result$samples / pd$total_samples, 1)
  result <- result[order(-result$samples), ]
  rownames(result) <- NULL

  apply_result_filters(result, n, min_pct, min_time_ms)
}

#' Print hot lines with source context
#'
#' Prints the hot source lines along with surrounding code context.
#'
#' @param x A profvis object.
#' @param n Number of hot lines to show.
#' @param context Number of lines to show before and after each hotspot.
#'
#' @return Invisibly returns a `debrief_hot_lines` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_hot_lines(p, n = 5, context = 3)
#' @export
pv_print_hot_lines <- function(x, n = 5, context = 3) {
  check_profvis(x)
  check_empty_profile(x)

  hot_lines <- pv_hot_lines(x, n = n)

  files <- extract_files(x)
  file_contents <- build_file_contents(files)

  obj <- structure(
    list(
      hot_lines = hot_lines,
      file_contents = file_contents,
      context = context
    ),
    class = "debrief_hot_lines"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_hot_lines <- function(x, ...) {
  hot_lines <- x$hot_lines
  file_contents <- x$file_contents
  context <- x$context

  if (nrow(hot_lines) == 0) {
    cat("No source location data available.\n")
    cat("Use devtools::load_all() to enable source references.\n")
    cat_help_hint()
    return(invisible(x))
  }

  cat_header("HOT SOURCE LINES")
  cat("\n")

  for (i in seq_len(nrow(hot_lines))) {
    row <- hot_lines[i, ]
    cat(sprintf(
      "Rank %d: %s (%.0f ms, %.1f%%)\n",
      i,
      row$location,
      row$time_ms,
      row$pct
    ))
    cat(sprintf("Function: %s\n\n", row$label))

    # Show source context
    start <- row$linenum - context
    end <- row$linenum + context
    lines <- get_source_lines(row$filename, start, end, file_contents)

    if (!is.null(lines)) {
      actual_start <- max(1L, start)
      for (j in seq_along(lines)) {
        ln <- actual_start + j - 1
        marker <- if (ln == row$linenum) ">    " else "     "
        cat(sprintf("%s%4d: %s\n", marker, ln, lines[j]))
      }
    } else {
      cat("  (source not available)\n")
    }
    cat("\n")
  }

  # Next steps suggestions
  if (nrow(hot_lines) > 0) {
    top_func <- hot_lines$label[1]
    top_file <- hot_lines$filename[1]
    suggestions <- character()
    if (is_user_function(top_func)) {
      suggestions <- c(suggestions, sprintf("pv_focus(p, \"%s\")", top_func))
    }
    suggestions <- c(
      suggestions,
      sprintf("pv_source_context(p, \"%s\")", top_file)
    )
    cat_next_steps(suggestions)
  }

  invisible(x)
}

#' Get the single hottest line
#'
#' Returns the hottest source line with full context. Useful for quickly
#' identifying the #1 optimization target.
#'
#' @param x A profvis object.
#' @param context Number of source lines to include before and after.
#'
#' @return A list with:
#'   - `location`: File path and line number (e.g., "R/foo.R:42")
#'   - `label`: Function name
#'   - `filename`: Source file path
#'   - `linenum`: Line number
#'   - `time_ms`: Time in milliseconds
#'   - `pct`: Percentage of total time
#'   - `code`: The source line
#'   - `context`: Vector of surrounding source lines
#'   - `callers`: Data frame of functions that call this location
#'
#'   Returns `NULL` if no source references are available.
#'
#' @examples
#' p <- pv_example()
#' pv_worst_line(p)
#' @export
pv_worst_line <- function(x, context = 5) {
  check_profvis(x)
  check_empty_profile(x)

  hot_lines <- pv_hot_lines(x, n = 1)

  if (nrow(hot_lines) == 0) {
    return(NULL)
  }

  row <- hot_lines[1, ]
  files <- extract_files(x)
  file_contents <- build_file_contents(files)

  # Get source line
  code <- get_source_line(row$filename, row$linenum, file_contents)
  if (is.null(code)) {
    code <- NA_character_
  }

  # Get context lines
  start <- max(1L, row$linenum - context)
  end <- row$linenum + context
  context_lines <- get_source_lines(row$filename, start, end, file_contents)
  if (is.null(context_lines)) {
    context_lines <- character()
  }

  # Get callers for this function
  callers <- pv_callers(x, row$label)

  list(
    location = row$location,
    label = row$label,
    filename = row$filename,
    linenum = row$linenum,
    time_ms = row$time_ms,
    pct = row$pct,
    code = code,
    context = context_lines,
    callers = callers
  )
}

test_that("pv_hot_lines returns correct structure", {
  p <- mock_profvis()
  result <- pv_hot_lines(p)

  expect_s3_class(result, "data.frame")
  expect_named(
    result,
    c("location", "samples", "label", "filename", "linenum", "time_ms", "pct")
  )
})

test_that("pv_hot_lines respects n parameter", {
  p <- mock_profvis()

  result_all <- pv_hot_lines(p)
  result_2 <- pv_hot_lines(p, n = 2)

  expect_lte(nrow(result_2), 2)
  expect_lte(nrow(result_2), nrow(result_all))
})

test_that("pv_hot_lines returns empty data frame when no source refs", {
  p <- mock_profvis_no_source()
  result <- pv_hot_lines(p)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_hot_lines location format is correct", {
  p <- mock_profvis()
  result <- pv_hot_lines(p)

  # Location should be filename:linenum format
  expect_true(all(grepl("^.+:\\d+$", result$location)))
})

test_that("pv_hot_paths returns correct structure", {
  p <- mock_profvis()
  result <- pv_hot_paths(p)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("stack", "samples", "time_ms", "pct"))
})

test_that("pv_hot_paths respects n parameter", {
  p <- mock_profvis()

  result_all <- pv_hot_paths(p)
  result_2 <- pv_hot_paths(p, n = 2)

  expect_lte(nrow(result_2), 2)
})

test_that("pv_hot_paths stacks contain arrows", {
  p <- mock_profvis()
  result <- pv_hot_paths(p)

  # Multi-level stacks should contain arrow separator
  multi_level <- result[grepl("->", result$stack), ]
  expect_gt(nrow(multi_level), 0)
})

test_that("pv_hot_paths include_source parameter works", {
  p <- mock_profvis()

  with_source <- pv_hot_paths(p, include_source = TRUE)
  without_source <- pv_hot_paths(p, include_source = FALSE)

  # With source should have file:line references
  expect_true(any(grepl("R/", with_source$stack)))

  # Without source should not have file references in labels
  expect_false(any(grepl(":\\d+\\)", without_source$stack)))
})

test_that("pv_print_hot_lines snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_hot_lines(p, n = 3))
})

test_that("pv_print_hot_lines handles no source refs", {
  p <- mock_profvis_no_source()
  expect_snapshot(pv_print_hot_lines(p))
})

test_that("pv_print_hot_paths snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_hot_paths(p, n = 3))
})

test_that("hot spot functions reject non-profvis input", {
  expect_error(pv_hot_lines(list()), "must be a profvis object")
  expect_error(pv_hot_paths("not profvis"), "must be a profvis object")
})

test_that("pv_print_hot_lines shows source not available when files empty", {
  prof <- data.frame(
    time = c(1L, 2L, 3L),
    depth = rep(1L, 3),
    label = rep("func", 3),
    filename = rep("R/main.R", 3),
    linenum = rep(5.0, 3),
    filenum = rep(1.0, 3),
    memalloc = c(100, 200, 300),
    meminc = rep(0, 3),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_print_hot_lines(p))
})

test_that("pv_worst_line returns NA code when source unavailable", {
  prof <- data.frame(
    time = c(1L, 2L),
    depth = rep(1L, 2),
    label = rep("func", 2),
    filename = rep("R/main.R", 2),
    linenum = rep(5.0, 2),
    filenum = rep(1.0, 2),
    memalloc = c(100, 200),
    meminc = rep(0, 2),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  result <- pv_worst_line(p)

  expect_equal(result$code, NA_character_)
  expect_equal(length(result$context), 0)
})

#' List available debrief functions
#'
#' Prints a summary of all available functions in the debrief package,
#' organized by category. Useful for discovering the API.
#'
#' @param category Optional category to filter by. If `NULL`, shows all
#'   categories. Options: "overview", "time", "hotspots", "memory", "calls",
#'   "source", "visualization", "comparison", "diagnostics", "export".
#'
#' @return Invisibly returns a `debrief_help` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' pv_help()
#' pv_help("time")
#' pv_help("hotspots")
#'
#' @export
pv_help <- function(category = NULL) {
  categories <- list(
    overview = list(
      title = "Overview",
      functions = c(
        "pv_print_debrief(p)" = "Print comprehensive profiling summary",
        "pv_debrief(p)" = "Get all profiling data as a list",
        "pv_help()" = "Show this help",
        "pv_example()" = "Load example profvis data for testing"
      )
    ),
    time = list(
      title = "Time Analysis",
      functions = c(
        "pv_self_time(p)" = "Functions by self-time (time at top of stack)",
        "pv_total_time(p)" = "Functions by total time (including callees)"
      )
    ),
    hotspots = list(
      title = "Hot Spots",
      functions = c(
        "pv_hot_lines(p)" = "Hottest source lines by self-time",
        "pv_print_hot_lines(p)" = "Print hot lines with source context",
        "pv_hot_paths(p)" = "Most common call stack paths",
        "pv_print_hot_paths(p)" = "Print hot paths in readable format",
        "pv_worst_line(p)" = "Get the single hottest line with context"
      )
    ),
    memory = list(
      title = "Memory Analysis",
      functions = c(
        "pv_memory(p)" = "Memory allocation by function",
        "pv_memory_lines(p)" = "Memory allocation by source line",
        "pv_print_memory(p)" = "Print memory allocation summary"
      )
    ),
    calls = list(
      title = "Call Analysis",
      functions = c(
        "pv_focus(p, \"func\")" = "Deep dive into a specific function",
        "pv_callers(p, \"func\")" = "Who calls this function?",
        "pv_callees(p, \"func\")" = "What does this function call?",
        "pv_print_callers_callees(p, \"func\")" = "Print caller/callee analysis",
        "pv_call_stats(p)" = "Call counts and per-call timing",
        "pv_print_call_stats(p)" = "Print call statistics table",
        "pv_call_depth(p)" = "Time distribution by call stack depth",
        "pv_recursive(p)" = "Detect recursive functions"
      )
    ),
    source = list(
      title = "Source Context",
      functions = c(
        "pv_source_context(p, \"file.R\")" = "Show source with profiling overlay",
        "pv_file_summary(p)" = "Time breakdown by source file",
        "pv_print_file_summary(p)" = "Print file-level summary"
      )
    ),
    visualization = list(
      title = "Visualization",
      functions = c(
        "pv_flame(p)" = "Text-based flame graph",
        "pv_flame_condense(p)" = "Condensed flame graph view"
      )
    ),
    comparison = list(
      title = "Comparison",
      functions = c(
        "pv_compare(before, after)" = "Compare two profiles",
        "pv_print_compare(before, after)" = "Print profile comparison",
        "pv_compare_many(...)" = "Compare multiple profiles",
        "pv_print_compare_many(...)" = "Print multi-profile comparison"
      )
    ),
    diagnostics = list(
      title = "Diagnostics",
      functions = c(
        "pv_gc_pressure(p)" = "Detect garbage collection overhead",
        "pv_print_gc_pressure(p)" = "Print GC pressure analysis",
        "pv_suggestions(p)" = "Get optimization suggestions",
        "pv_print_suggestions(p)" = "Print actionable suggestions"
      )
    ),
    export = list(
      title = "Export",
      functions = c(
        "pv_to_list(p)" = "Export as R list",
        "pv_to_json(p)" = "Export as JSON string"
      )
    )
  )

  valid_categories <- names(categories)

  if (!is.null(category)) {
    category <- tolower(category)
    if (!category %in% valid_categories) {
      obj <- structure(
        list(
          categories = categories,
          filtered = NULL,
          invalid_category = category,
          valid_categories = valid_categories
        ),
        class = "debrief_help"
      )
      print(obj)
      return(invisible(obj))
    }
    categories <- categories[category]
  }

  obj <- structure(
    list(
      categories = categories,
      filtered = category,
      invalid_category = NULL,
      valid_categories = valid_categories
    ),
    class = "debrief_help"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_help <- function(x, ...) {
  if (!is.null(x$invalid_category)) {
    cat("Unknown category:", x$invalid_category, "\n")
    cat("Valid categories:", paste(x$valid_categories, collapse = ", "), "\n")
    return(invisible(x))
  }

  cat("## DEBRIEF FUNCTIONS\n\n")

  if (is.null(x$filtered)) {
    cat("Use pv_help(\"category\") to see functions in a specific category.\n")
    cat("Categories:", paste(x$valid_categories, collapse = ", "), "\n\n")
  }

  for (cat_info in x$categories) {
    cat("###", cat_info$title, "\n")
    funcs <- cat_info$functions
    for (i in seq_along(funcs)) {
      cat(sprintf("  %-35s %s\n", names(funcs)[i], funcs[i]))
    }
    cat("\n")
  }

  cat("### Typical Workflow\n")
  cat(
    "  1. pv_print_debrief(p)
        -> Get overview, identify hot functions\n"
  )
  cat(
    "  2. pv_focus(p, \"hot_func\")
        -> Deep dive into the hottest function\n"
  )
  cat(
    "  3. pv_hot_lines(p)
        -> Find exact lines consuming time\n"
  )
  cat(
    "  4. pv_source_context(p, \"file.R\")
        -> See code with profiling overlay\n"
  )
  cat(
    "  5. pv_suggestions(p)
        -> Get optimization recommendations\n"
  )

  invisible(x)
}

test_that("pv_help returns invisibly", {
  result <- expect_invisible(pv_help())
  expect_type(result, "list")
})

test_that("pv_help snapshot", {
  expect_snapshot(pv_help())
})

test_that("pv_help with category returns invisibly", {
  result <- expect_invisible(pv_help("time"))
  expect_type(result, "list")
})

test_that("pv_help with category snapshot", {
  expect_snapshot(pv_help("time"))
})

test_that("pv_help hotspots snapshot", {
  expect_snapshot(pv_help("hotspots"))
})

test_that("pv_help memory snapshot", {
  expect_snapshot(pv_help("memory"))
})

test_that("pv_help calls snapshot", {
  expect_snapshot(pv_help("calls"))
})

test_that("pv_help source snapshot", {
  expect_snapshot(pv_help("source"))
})

test_that("pv_help comparison snapshot", {
  expect_snapshot(pv_help("comparison"))
})

test_that("pv_help diagnostics snapshot", {
  expect_snapshot(pv_help("diagnostics"))
})

test_that("pv_help export snapshot", {
  expect_snapshot(pv_help("export"))
})

test_that("pv_help with unknown category snapshot", {
  expect_snapshot(pv_help("nonexistent"))
})

#' Hot call paths
#'
#' Returns the most common complete call stacks. This shows which execution
#' paths through the code consume the most time.
#'
#' @param x A profvis object.
#' @param n Maximum number of paths to return. If `NULL`, returns all.
#' @param include_source If `TRUE` and source references are available, include
#'   file:line information in the path labels.
#'
#' @return A data frame with columns:
#'   - `stack`: The call path (functions separated by arrows)
#'   - `samples`: Number of profiling samples with this exact path
#'   - `time_ms`: Time in milliseconds
#'   - `pct`: Percentage of total time
#'
#' @examples
#' p <- pv_example()
#' pv_hot_paths(p)
#' @export
pv_hot_paths <- function(x, n = NULL, include_source = TRUE) {
  pd <- extract_profile_data(x)
  has_source <- has_source_refs(x)

  # Build call stacks (bottom-up: caller first)
  prof_sorted <- pd$prof[order(pd$prof$time, pd$prof$depth), ]

  # Create label with optional source location
  if (has_source && include_source) {
    prof_sorted$display <- ifelse(
      is.na(prof_sorted$filename),
      prof_sorted$label,
      paste0(
        prof_sorted$label,
        " (",
        prof_sorted$filename,
        ":",
        prof_sorted$linenum,
        ")"
      )
    )
  } else {
    prof_sorted$display <- prof_sorted$label
  }

  stacks <- tapply(
    prof_sorted$display,
    prof_sorted$time,
    function(labels) paste(labels, collapse = " -> ")
  )

  stack_df <- data.frame(
    time = as.integer(names(stacks)),
    stack = as.character(stacks)
  )

  counts <- table(stack_df$stack)
  result <- data.frame(
    stack = names(counts),
    samples = as.integer(counts)
  )
  result$time_ms <- result$samples * pd$interval_ms
  result$pct <- round(100 * result$samples / pd$total_samples, 1)
  result <- result[order(-result$samples), ]
  rownames(result) <- NULL

  if (!is.null(n)) {
    result <- head(result, n)
  }

  result
}

#' Print hot paths in readable format
#'
#' @param x A profvis object.
#' @param n Number of paths to show.
#' @param include_source Include source references in output.
#'
#' @return Invisibly returns a `debrief_hot_paths` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_hot_paths(p, n = 3)
#'
#' @export
pv_print_hot_paths <- function(x, n = 10, include_source = TRUE) {
  check_profvis(x)
  check_empty_profile(x)

  paths <- pv_hot_paths(x, n = n, include_source = include_source)
  obj <- structure(list(paths = paths), class = "debrief_hot_paths")
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_hot_paths <- function(x, ...) {
  paths <- x$paths

  if (nrow(paths) == 0) {
    cat("No profiling data available.\n")
    cat_help_hint()
    return(invisible(x))
  }

  cat_header("HOT CALL PATHS")
  cat("\n")

  for (i in seq_len(nrow(paths))) {
    row <- paths[i, ]
    cat(sprintf(
      "Rank %d: %.0f ms (%.1f%%) - %d samples\n",
      i,
      row$time_ms,
      row$pct,
      row$samples
    ))

    parts <- strsplit(row$stack, " -> ")[[1]]
    cat("    ", paste(parts, collapse = "\n  -> "), "\n\n", sep = "")
  }

  # Next steps suggestions
  if (nrow(paths) > 0) {
    # Extract the leaf function from the hottest path (last in the chain)
    parts <- strsplit(paths$stack[1], " -> ")[[1]]
    # Remove source info if present: "func (file:line)" -> "func"
    leaf_func <- sub(" \\(.*\\)$", "", parts[length(parts)])
    suggestions <- "pv_flame(p)"
    if (is_user_function(leaf_func)) {
      suggestions <- c(sprintf("pv_focus(p, \"%s\")", leaf_func), suggestions)
    }
    cat_next_steps(suggestions)
  }

  invisible(x)
}

# Internal utility functions

check_profvis <- function(x, call = parent.frame()) {
  if (!inherits(x, "profvis")) {
    stop("`x` must be a profvis object.", call. = FALSE)
  }
}

check_empty_profile <- function(x) {
  prof <- extract_prof(x)
  if (nrow(prof) == 0) {
    stop(
      "Profile contains no samples.\n",
      "Your code ran too fast to capture any profiling data.\n",
      "Try wrapping your code in a loop: for (i in 1:10) { ... }\n",
      "Increase the iteration count until samples appear.",
      call. = FALSE
    )
  }
}

extract_prof <- function(x) {
  x$x$message$prof
}

extract_interval <- function(x) {
  x$x$message$interval
}

extract_files <- function(x) {
  x$x$message$files
}

extract_total_samples <- function(x) {
  max(extract_prof(x)$time)
}
# Extract rows where each function is at the top of the call stack (self-time)
# This is used for self-time calculations throughout the package
extract_top_of_stack <- function(prof) {
  max_depths <- tapply(prof$depth, prof$time, max)
  max_depth_df <- data.frame(
    time = as.integer(names(max_depths)),
    max_depth = as.integer(max_depths)
  )
  prof_merged <- merge(prof, max_depth_df, by = "time")
  prof_merged[prof_merged$depth == prof_merged$max_depth, ]
}

has_source_refs <- function(x) {
  prof <- extract_prof(x)
  any(!is.na(prof$filename))
}

build_file_contents <- function(files) {
  if (is.null(files) || length(files) == 0) {
    return(list())
  }
  contents <- list()
  for (f in files) {
    if (!is.null(f$content) && !is.null(f$filename)) {
      lines <- strsplit(f$content, "\n", fixed = TRUE)[[1]]
      contents[[f$filename]] <- lines
    }
  }
  contents
}

get_source_line <- function(filename, linenum, file_contents) {
  if (is.null(file_contents) || length(file_contents) == 0) {
    return(NULL)
  }
  lines <- file_contents[[filename]]
  if (is.null(lines) || linenum > length(lines)) {
    return(NULL)
  }
  trimws(lines[linenum])
}

get_source_lines <- function(filename, start, end, file_contents) {
  if (is.null(file_contents) || length(file_contents) == 0) {
    return(NULL)
  }
  lines <- file_contents[[filename]]
  if (is.null(lines)) {
    return(NULL)
  }
  start <- max(1L, start)
  end <- min(length(lines), end)
  if (start > end) {
    return(NULL)
  }
  lines[start:end]
}

# Formatting helpers
cat_header <- function(text) {
  cat("## ", text, "\n\n", sep = "")
}

cat_section <- function(text) {
  cat("\n### ", text, "\n", sep = "")
}

# Format time with percentage - used throughout package for consistent output
fmt_time <- function(time_ms, pct, time_width = 6, pct_width = 5) {
  sprintf(
    paste0("%", time_width, ".0f ms (%", pct_width, ".1f%%)"),
    time_ms,
    pct
  )
}

# Format memory in MB - used throughout package for consistent output
fmt_memory <- function(mem_mb, width = 8) {
  sprintf(paste0("%", width, ".2f MB"), mem_mb)
}

truncate_string <- function(s, max_len = 60) {
  if (nchar(s) > max_len) {
    paste0(substr(s, 1, max_len - 3), "...")
  } else {
    s
  }
}

# Print "Next steps" suggestions for AI agents
# suggestions should be a character vector of R expressions
cat_next_steps <- function(suggestions) {
  if (length(suggestions) == 0) {
    return(invisible())
  }
  cat("\n### Next steps\n")
  for (s in suggestions) {
    cat(s, "\n", sep = "")
  }
}

# Print a hint about pv_help() when no data is found
cat_help_hint <- function() {
  cat("Run pv_help() to see all available functions.\n")
}

# Check if a function name is a user function (not internal R machinery)
# Internal functions start with: ( like (top-level), < like <GC>, [ like [.data.frame
is_user_function <- function(func_name) {
  !grepl("^[(<\\[]", func_name)
}

# Extract common profile data in one call - reduces boilerplate
extract_profile_data <- function(x) {
  check_profvis(x)

  check_empty_profile(x)
  prof <- extract_prof(x)
  list(
    prof = prof,
    interval_ms = extract_interval(x),
    total_samples = extract_total_samples(x)
  )
}

# Build a result data frame from label counts with time calculations
build_count_result <- function(counts, interval_ms, total_samples) {
  result <- data.frame(
    label = names(counts),
    samples = as.integer(counts)
  )
  result$time_ms <- result$samples * interval_ms
  result$pct <- round(100 * result$samples / total_samples, 1)
  result <- result[order(-result$samples), ]
  rownames(result) <- NULL
  result
}

# Apply standard filters and limit to result data frames
apply_result_filters <- function(
  result,
  n = NULL,
  min_pct = 0,
  min_time_ms = 0
) {
  result <- result[result$pct >= min_pct & result$time_ms >= min_time_ms, ]
  if (!is.null(n)) {
    result <- head(result, n)
  }
  result
}

# Empty data frame factories for consistent return types
empty_label_samples_pct <- function() {
  data.frame(
    label = character(),
    samples = integer(),
    pct = numeric()
  )
}

empty_time_result <- function() {
  data.frame(
    label = character(),
    samples = integer(),
    time_ms = numeric(),
    pct = numeric()
  )
}

empty_location_result <- function() {
  data.frame(
    location = character(),
    label = character(),
    filename = character(),
    linenum = integer(),
    samples = integer(),
    time_ms = numeric(),
    pct = numeric()
  )
}

empty_memory_result <- function() {
  data.frame(
    label = character(),
    mem_mb = numeric()
  )
}

empty_memory_lines_result <- function() {
  data.frame(
    location = character(),
    label = character(),
    filename = character(),
    linenum = integer(),
    mem_mb = numeric()
  )
}

test_that("truncate_string truncates long strings", {
  long <- paste(rep("a", 70), collapse = "")
  result <- debrief:::truncate_string(long, max_len = 10)
  expect_equal(nchar(result), 10)
  expect_match(result, "\\.\\.\\.$")
})

test_that("truncate_string keeps short strings unchanged", {
  s <- "short"
  expect_equal(debrief:::truncate_string(s), s)
})

test_that("fmt_time formats correctly", {
  result <- debrief:::fmt_time(100, 25.5)
  expect_type(result, "character")
  expect_match(result, "100")
  expect_match(result, "25\\.5")
})

test_that("fmt_time respects width parameters", {
  result <- debrief:::fmt_time(50, 10.0, time_width = 8, pct_width = 6)
  expect_type(result, "character")
  expect_match(result, "50")
})

test_that("cat_next_steps returns invisible when empty", {
  expect_invisible(debrief:::cat_next_steps(character()))
})

test_that("cat_next_steps prints suggestions", {
  expect_output(debrief:::cat_next_steps(c("foo()", "bar()")), "foo")
})

test_that("empty_time_result returns empty data frame with correct columns", {
  result <- debrief:::empty_time_result()
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
  expect_named(result, c("label", "samples", "time_ms", "pct"))
})

test_that("get_source_line returns NULL for empty file_contents", {
  result <- debrief:::get_source_line("R/main.R", 5, list())
  expect_null(result)
})

test_that("get_source_line returns NULL for unknown file", {
  contents <- list("R/other.R" = c("line1", "line2"))
  result <- debrief:::get_source_line("R/main.R", 1, contents)
  expect_null(result)
})

test_that("get_source_line returns NULL when linenum exceeds file length", {
  contents <- list("R/main.R" = c("line1", "line2"))
  result <- debrief:::get_source_line("R/main.R", 100, contents)
  expect_null(result)
})

test_that("get_source_lines returns NULL for empty file_contents", {
  result <- debrief:::get_source_lines("R/main.R", 1, 5, list())
  expect_null(result)
})

test_that("get_source_lines returns NULL for unknown file", {
  contents <- list("R/other.R" = c("line1", "line2"))
  result <- debrief:::get_source_lines("R/main.R", 1, 5, contents)
  expect_null(result)
})

test_that("get_source_lines returns NULL when start > end after clamping", {
  contents <- list("R/main.R" = c("line1", "line2", "line3"))
  result <- debrief:::get_source_lines("R/main.R", 10, 2, contents)
  expect_null(result)
})

#' Focused analysis of a specific function
#'
#' Provides a comprehensive analysis of a single function including time
#' breakdown, callers, callees, and source context if available.
#'
#' @param x A profvis object.
#' @param func The function name to analyze.
#' @param context Number of source lines to show around hotspots.
#'
#' @return Invisibly returns a `debrief_focus` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_focus(p, "inner")
#' @export
pv_focus <- function(x, func, context = 5) {
  pd <- extract_profile_data(x)
  files <- extract_files(x)
  file_contents <- build_file_contents(files)

  # Check if function exists
  func_rows <- pd$prof[pd$prof$label == func, ]
  if (nrow(func_rows) == 0) {
    total_time <- pv_total_time(x)
    obj <- structure(
      list(
        func = func,
        not_found = TRUE,
        total_time = total_time
      ),
      class = "debrief_focus"
    )
    print(obj)
    return(invisible(obj))
  }

  # Calculate time stats
  func_times <- unique(func_rows$time)
  func_total_time <- length(func_times) * pd$interval_ms
  func_total_pct <- round(100 * length(func_times) / pd$total_samples, 1)

  # Calculate self-time (deepest frame for each time point)
  top_of_stack <- extract_top_of_stack(pd$prof)
  self_samples <- sum(top_of_stack$label == func)
  self_time <- self_samples * pd$interval_ms
  self_pct <- round(100 * self_samples / pd$total_samples, 1)

  # Get callers and callees
  callers <- pv_callers(x, func)
  callees <- pv_callees(x, func)

  # Get source locations for this function
  func_with_source <- func_rows[!is.na(func_rows$filename), ]
  has_source <- nrow(func_with_source) > 0

  # Filter to self-time only (needed for source locations and next steps)
  func_top_of_stack <- top_of_stack[
    top_of_stack$label == func & !is.na(top_of_stack$filename),
  ]

  obj <- structure(
    list(
      func = func,
      not_found = FALSE,
      func_times = func_times,
      func_total_time = func_total_time,
      func_total_pct = func_total_pct,
      self_time = self_time,
      self_pct = self_pct,
      callers = callers,
      callees = callees,
      has_source = has_source,
      func_with_source = func_with_source,
      func_top_of_stack = func_top_of_stack,
      file_contents = file_contents,
      context = context,
      interval_ms = pd$interval_ms,
      total_samples = pd$total_samples
    ),
    class = "debrief_focus"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_focus <- function(x, ...) {
  func <- x$func

  if (x$not_found) {
    cat(sprintf("Function '%s' not found in profiling data.\n\n", func))
    cat("Available functions (top 20 by time):\n")
    total_time <- x$total_time
    for (i in seq_len(min(20, nrow(total_time)))) {
      cat(sprintf("  %s\n", total_time$label[i]))
    }
    cat("\n")
    cat_help_hint()
    return(invisible(x))
  }

  func_times <- x$func_times
  func_total_time <- x$func_total_time
  func_total_pct <- x$func_total_pct
  self_time <- x$self_time
  self_pct <- x$self_pct
  callers <- x$callers
  callees <- x$callees
  has_source <- x$has_source
  func_with_source <- x$func_with_source
  func_top_of_stack <- x$func_top_of_stack
  file_contents <- x$file_contents
  context <- x$context
  interval_ms <- x$interval_ms
  total_samples <- x$total_samples

  # Print output
  cat_header(sprintf("FOCUS: %s", func))
  cat("\n")

  cat("### Time Analysis\n")
  cat(sprintf(
    "  Total time:   %6.0f ms (%5.1f%%)  - time on call stack\n",
    func_total_time,
    func_total_pct
  ))
  cat(sprintf(
    "  Self time:    %6.0f ms (%5.1f%%)  - time at top of stack\n",
    self_time,
    self_pct
  ))
  cat(sprintf(
    "  Child time:   %6.0f ms (%5.1f%%)  - time in callees\n",
    func_total_time - self_time,
    func_total_pct - self_pct
  ))
  cat(sprintf("  Appearances:  %6d samples\n\n", length(func_times)))

  cat("### Called By\n")
  if (nrow(callers) == 0) {
    cat("  Callers: none\n")
  } else {
    for (i in seq_len(min(5, nrow(callers)))) {
      cat(sprintf(
        "  %5d calls (%5.1f%%)  %s\n",
        callers$samples[i],
        callers$pct[i],
        callers$label[i]
      ))
    }
    if (nrow(callers) > 5) {
      cat(sprintf("  ... and %d more callers\n", nrow(callers) - 5))
    }
  }
  cat("\n")

  cat("### Calls To\n")
  if (nrow(callees) == 0) {
    cat("  Callees: none\n")
  } else {
    for (i in seq_len(min(5, nrow(callees)))) {
      cat(sprintf(
        "  %5d calls (%5.1f%%)  %s\n",
        callees$samples[i],
        callees$pct[i],
        callees$label[i]
      ))
    }
    if (nrow(callees) > 5) {
      cat(sprintf("  ... and %d more callees\n", nrow(callees) - 5))
    }
  }
  cat("\n")

  # Show source locations if available
  if (has_source) {
    cat("### Source Locations\n")

    # Get hottest lines for this function
    func_with_source$location <- paste0(
      func_with_source$filename,
      ":",
      func_with_source$linenum
    )

    if (nrow(func_top_of_stack) > 0) {
      func_top_of_stack$location <- paste0(
        func_top_of_stack$filename,
        ":",
        func_top_of_stack$linenum
      )
      line_counts <- sort(table(func_top_of_stack$location), decreasing = TRUE)

      cat("Hot lines (by self-time):\n")
      for (i in seq_len(min(5, length(line_counts)))) {
        loc <- names(line_counts)[i]
        samples <- as.integer(line_counts[i])
        time_ms <- samples * interval_ms
        pct <- round(100 * samples / total_samples, 1)

        cat(sprintf("  %5.0f ms (%4.1f%%)  %s\n", time_ms, pct, loc))

        # Show source line
        parts <- strsplit(loc, ":")[[1]]
        fn <- paste(parts[-length(parts)], collapse = ":")
        ln <- as.integer(parts[length(parts)])
        src_line <- get_source_line(fn, ln, file_contents)
        if (!is.null(src_line) && nchar(src_line) > 0) {
          cat(sprintf("                   %s\n", truncate_string(src_line)))
        }
      }
    } else {
      cat("  No self-time with source info.\n")
    }

    # Show source file context for hottest line
    if (nrow(func_top_of_stack) > 0) {
      line_counts <- sort(table(func_top_of_stack$location), decreasing = TRUE)
      hottest <- names(line_counts)[1]
      parts <- strsplit(hottest, ":")[[1]]
      fn <- paste(parts[-length(parts)], collapse = ":")
      ln <- as.integer(parts[length(parts)])

      cat(sprintf("\n### Source Context: %s\n", fn))

      start_line <- max(1L, ln - context)
      end_line <- ln + context
      source_lines <- get_source_lines(fn, start_line, end_line, file_contents)

      if (!is.null(source_lines)) {
        for (j in seq_along(source_lines)) {
          line_num <- start_line + j - 1
          marker <- if (line_num == ln) ">    " else "     "
          cat(sprintf("%s%4d: %s\n", marker, line_num, source_lines[j]))
        }
      }
    }
  } else {
    cat("### Source Locations\n")
    cat("  Source references not available.\n")
    cat("  Use devtools::load_all() to enable.\n")
  }

  # Next steps suggestions
  suggestions <- character()
  if (nrow(callees) > 0 && is_user_function(callees$label[1])) {
    suggestions <- c(
      suggestions,
      sprintf("pv_focus(p, \"%s\")", callees$label[1])
    )
  }
  if (nrow(callers) > 0) {
    suggestions <- c(suggestions, sprintf("pv_callers(p, \"%s\")", func))
    if (is_user_function(callers$label[1])) {
      suggestions <- c(
        suggestions,
        sprintf("pv_focus(p, \"%s\")", callers$label[1])
      )
    }
  }
  if (has_source && nrow(func_top_of_stack) > 0) {
    line_counts <- sort(table(func_top_of_stack$location), decreasing = TRUE)
    hottest <- names(line_counts)[1]
    parts <- strsplit(hottest, ":")[[1]]
    fn <- paste(parts[-length(parts)], collapse = ":")
    suggestions <- c(suggestions, sprintf("pv_source_context(p, \"%s\")", fn))
  }
  cat_next_steps(suggestions)

  invisible(x)
}

#' Call depth breakdown
#'
#' Shows time distribution across different call stack depths. Useful for
#' understanding how deeply nested the hot code paths are.
#'
#' @param x A profvis object.
#'
#' @return A data frame with columns:
#'   - `depth`: Call stack depth (1 = top level)
#'   - `samples`: Number of profiling samples at this depth
#'   - `time_ms`: Time in milliseconds
#'   - `pct`: Percentage of total time
#'   - `top_funcs`: Most common functions at this depth
#'
#' @examples
#' p <- pv_example()
#' pv_call_depth(p)
#' @export
pv_call_depth <- function(x) {
  pd <- extract_profile_data(x)

  # Get unique depths present in each time sample
  depths <- sort(unique(pd$prof$depth))

  result <- lapply(depths, function(d) {
    at_depth <- pd$prof[pd$prof$depth == d, ]
    unique_times <- unique(at_depth$time)
    n_samples <- length(unique_times)

    # Find top functions at this depth
    func_counts <- sort(table(at_depth$label), decreasing = TRUE)
    top_funcs <- paste(head(names(func_counts), 3), collapse = ", ")

    data.frame(
      depth = d,
      samples = n_samples,
      time_ms = n_samples * pd$interval_ms,
      pct = round(100 * n_samples / pd$total_samples, 1),
      top_funcs = top_funcs
    )
  })

  do.call(rbind, result)
}

#' Print call depth breakdown
#'
#' @param x A profvis object.
#'
#' @return Invisibly returns a `debrief_call_depth` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_call_depth(p)
#'
#' @export
pv_print_call_depth <- function(x) {
  check_profvis(x)
  check_empty_profile(x)

  depth_df <- pv_call_depth(x)
  obj <- structure(list(depth_df = depth_df), class = "debrief_call_depth")
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_call_depth <- function(x, ...) {
  depth_df <- x$depth_df

  if (nrow(depth_df) == 0) {
    cat("No profiling data available.\n")
    cat_help_hint()
    return(invisible(x))
  }

  cat_header("CALL DEPTH BREAKDOWN")
  cat("\n")
  cat("Depth  Time (ms)   Pct   Top functions\n")

  for (i in seq_len(nrow(depth_df))) {
    row <- depth_df[i, ]
    cat(sprintf(
      "%5d  %8.0f  %5.1f%%  %s\n",
      row$depth,
      row$time_ms,
      row$pct,
      truncate_string(row$top_funcs, 45)
    ))
  }

  # Next steps - suggest focusing on a function at the deepest hot depth
  if (nrow(depth_df) > 0) {
    # Find the depth with most time
    hot_depth <- depth_df$depth[which.max(depth_df$time_ms)]
    hot_row <- depth_df[depth_df$depth == hot_depth, ]
    first_func <- strsplit(hot_row$top_funcs, ", ")[[1]][1]
    if (
      !is.na(first_func) &&
        nchar(first_func) > 0 &&
        is_user_function(first_func)
    ) {
      cat_next_steps(c(
        sprintf("pv_focus(p, \"%s\")", first_func),
        "pv_flame(p)"
      ))
    }
  }

  invisible(x)
}

#' Compare two profvis profiles
#'
#' Compares two profiling runs to show what changed. Useful for measuring
#' the impact of optimizations.
#'
#' @param before A profvis object (before optimization).
#' @param after A profvis object (after optimization).
#' @param n Number of top functions to compare.
#'
#' @return A list with:
#'   - `summary`: Overall comparison summary
#'   - `by_function`: Function-by-function comparison
#'   - `improved`: Functions that got faster
#'   - `regressed`: Functions that got slower
#'
#' @examples
#' p1 <- pv_example()
#' p2 <- pv_example()
#' pv_compare(p1, p2)
#' @export
pv_compare <- function(before, after, n = 20) {
  check_profvis(before)
  check_profvis(after)
  check_empty_profile(before)
  check_empty_profile(after)

  # Get timing info
  before_interval <- extract_interval(before)
  after_interval <- extract_interval(after)

  before_samples <- extract_total_samples(before)
  after_samples <- extract_total_samples(after)

  before_time <- before_samples * before_interval
  after_time <- after_samples * after_interval

  # Summary
  time_diff <- after_time - before_time
  time_pct_change <- round(100 * (after_time - before_time) / before_time, 1)
  speedup <- round(before_time / after_time, 2)

  summary_df <- data.frame(
    metric = c("Total time (ms)", "Samples", "Speedup"),
    before = c(before_time, before_samples, 1.0),
    after = c(after_time, after_samples, speedup),
    change = c(
      sprintf("%+.0f ms (%+.1f%%)", time_diff, time_pct_change),
      sprintf("%+d", after_samples - before_samples),
      sprintf("%.2fx", speedup)
    )
  )

  # Function-by-function comparison (self-time)
  before_self <- pv_self_time(before)
  after_self <- pv_self_time(after)

  # Merge
  all_funcs <- union(before_self$label, after_self$label)
  func_comparison <- lapply(all_funcs, function(func) {
    before_row <- before_self[before_self$label == func, ]
    after_row <- after_self[after_self$label == func, ]

    before_ms <- if (nrow(before_row) > 0) before_row$time_ms[1] else 0
    after_ms <- if (nrow(after_row) > 0) after_row$time_ms[1] else 0

    diff_ms <- after_ms - before_ms
    if (before_ms > 0) {
      pct_change <- round(100 * (after_ms - before_ms) / before_ms, 1)
    } else if (after_ms > 0) {
      pct_change <- Inf
    } else {
      pct_change <- 0
    }

    data.frame(
      label = func,
      before_ms = before_ms,
      after_ms = after_ms,
      diff_ms = diff_ms,
      pct_change = pct_change
    )
  })

  func_df <- do.call(rbind, func_comparison)
  func_df <- func_df[order(-abs(func_df$diff_ms)), ]
  rownames(func_df) <- NULL

  # Split into improved/regressed
  improved <- func_df[func_df$diff_ms < -5, ] # At least 5ms improvement
  improved <- improved[order(improved$diff_ms), ]

  regressed <- func_df[func_df$diff_ms > 5, ] # At least 5ms regression
  regressed <- regressed[order(-regressed$diff_ms), ]

  list(
    summary = summary_df,
    by_function = head(func_df, n),
    improved = improved,
    regressed = regressed
  )
}

#' Print profile comparison
#'
#' @param before A profvis object (before optimization).
#' @param after A profvis object (after optimization).
#' @param n Number of functions to show in detailed comparison.
#'
#' @return Invisibly returns a `debrief_compare` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p1 <- pv_example()
#' p2 <- pv_example()
#' pv_print_compare(p1, p2)
#'
#' @export
pv_print_compare <- function(before, after, n = 15) {
  check_profvis(before)
  check_profvis(after)

  comp <- pv_compare(before, after, n = n)
  obj <- structure(list(comp = comp, n = n), class = "debrief_compare")
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_compare <- function(x, ...) {
  comp <- x$comp
  n <- x$n

  cat_header("PROFILE COMPARISON")
  cat("\n")

  # Summary
  cat_section("Overall")
  before_time <- comp$summary$before[1]
  after_time <- comp$summary$after[1]
  speedup <- comp$summary$after[3]
  diff_time <- after_time - before_time

  cat(sprintf("before_ms: %.0f\n", before_time))
  cat(sprintf("after_ms: %.0f\n", after_time))
  cat(sprintf("diff_ms: %+.0f\n", diff_time))
  cat(sprintf("speedup: %.2fx\n", speedup))

  # Top changes
  cat_section("Biggest Changes")
  cat(sprintf(
    "%-30s %10s %10s %10s %8s\n",
    "Function",
    "Before",
    "After",
    "Diff",
    "Change"
  ))

  for (i in seq_len(min(n, nrow(comp$by_function)))) {
    row <- comp$by_function[i, ]
    if (abs(row$diff_ms) < 1) {
      next
    } # Skip tiny changes

    label <- truncate_string(row$label, 30)
    change_str <- if (is.finite(row$pct_change)) {
      sprintf("%+.0f%%", row$pct_change)
    } else {
      "new"
    }

    cat(sprintf(
      "%-30s %10.0f %10.0f %+10.0f %8s\n",
      label,
      row$before_ms,
      row$after_ms,
      row$diff_ms,
      change_str
    ))
  }

  # Improved functions
  if (nrow(comp$improved) > 0) {
    cat_section("Top Improvements")
    for (i in seq_len(min(5, nrow(comp$improved)))) {
      row <- comp$improved[i, ]
      cat(sprintf(
        "  %s: %.0f -> %.0f (%+.0f ms)\n",
        truncate_string(row$label, 35),
        row$before_ms,
        row$after_ms,
        row$diff_ms
      ))
    }
  }

  # Regressed functions
  if (nrow(comp$regressed) > 0) {
    cat_section("Regressions")
    for (i in seq_len(min(5, nrow(comp$regressed)))) {
      row <- comp$regressed[i, ]
      cat(sprintf(
        "  %s: %.0f -> %.0f (%+.0f ms)\n",
        truncate_string(row$label, 35),
        row$before_ms,
        row$after_ms,
        row$diff_ms
      ))
    }
  }

  # Next steps - suggest drilling into the profile with most change
  suggestions <- character()
  if (nrow(comp$by_function) > 0) {
    top_func <- comp$by_function$label[1]
    if (is_user_function(top_func)) {
      suggestions <- c(
        suggestions,
        sprintf("pv_focus(p_before, \"%s\")", top_func),
        sprintf("pv_focus(p_after, \"%s\")", top_func)
      )
    }
  }
  if (length(suggestions) > 0) {
    cat_next_steps(suggestions)
  }

  invisible(x)
}

#' Compare multiple profvis profiles
#'
#' Compares multiple profiling runs to identify the fastest. Useful for
#' comparing different optimization approaches.
#'
#' @param ... Named profvis objects to compare, or a single named list of
#'   profvis objects.
#'
#' @return A data frame with columns:
#'   - `name`: Profile name
#'   - `time_ms`: Total time in milliseconds
#'   - `samples`: Number of samples
#'   - `vs_fastest`: How much slower than the fastest (e.g., "1.5x")
#'   - `rank`: Rank from fastest (1) to slowest
#'
#' @examples
#' p1 <- pv_example()
#' p2 <- pv_example("gc")
#' p3 <- pv_example("recursive")
#' pv_compare_many(baseline = p1, gc_heavy = p2, recursive = p3)
#'
#' # Or pass a named list
#' profiles <- list(baseline = p1, gc_heavy = p2)
#' pv_compare_many(profiles)
#' @export
pv_compare_many <- function(...) {
  args <- list(...)

  # Handle single list argument
  if (
    length(args) == 1 && is.list(args[[1]]) && !inherits(args[[1]], "profvis")
  ) {
    profiles <- args[[1]]
  } else {
    profiles <- args
  }

  if (length(profiles) < 2) {
    stop("At least 2 profiles are required for comparison.", call. = FALSE)
  }

  # Check names
  names_vec <- names(profiles)
  if (is.null(names_vec) || any(names_vec == "")) {
    stop("All profiles must be named.", call. = FALSE)
  }

  # Validate all are profvis objects
  for (nm in names_vec) {
    if (!inherits(profiles[[nm]], "profvis")) {
      stop(sprintf("'%s' must be a profvis object.", nm), call. = FALSE)
    }
    check_empty_profile(profiles[[nm]])
  }

  # Extract timing for each
  results <- lapply(names_vec, function(nm) {
    p <- profiles[[nm]]
    interval_ms <- extract_interval(p)
    samples <- extract_total_samples(p)
    time_ms <- samples * interval_ms

    data.frame(
      name = nm,
      time_ms = time_ms,
      samples = samples
    )
  })

  result <- do.call(rbind, results)
  result <- result[order(result$time_ms), ]

  # Calculate vs_fastest
  fastest_time <- min(result$time_ms)
  result$vs_fastest <- sprintf("%.2fx", result$time_ms / fastest_time)
  result$vs_fastest[result$time_ms == fastest_time] <- "fastest"

  # Add rank
  result$rank <- seq_len(nrow(result))

  rownames(result) <- NULL
  result
}

#' Print comparison of multiple profiles
#'
#' @param ... Named profvis objects to compare, or a single named list.
#'
#' @return Invisibly returns a `debrief_compare_many` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p1 <- pv_example()
#' p2 <- pv_example("gc")
#' pv_print_compare_many(baseline = p1, gc_heavy = p2)
#' @export
pv_print_compare_many <- function(...) {
  result <- pv_compare_many(...)
  obj <- structure(list(result = result), class = "debrief_compare_many")
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_compare_many <- function(x, ...) {
  result <- x$result

  cat_header("MULTI-PROFILE COMPARISON")
  cat("\n")

  cat(sprintf(
    "%4s  %-25s %10s %8s %10s\n",
    "Rank",
    "Profile",
    "Time (ms)",
    "Samples",
    "vs Fastest"
  ))

  for (i in seq_len(nrow(result))) {
    row <- result[i, ]
    marker <- if (row$rank == 1) "*" else " "
    cat(sprintf(
      "%3d%s  %-25s %10.0f %8d %10s\n",
      row$rank,
      marker,
      truncate_string(row$name, 25),
      row$time_ms,
      row$samples,
      row$vs_fastest
    ))
  }

  cat("\n* = fastest\n")

  invisible(x)
}

test_that("pv_compare returns correct structure", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()

  result <- pv_compare(p1, p2)

  expect_type(result, "list")
  expect_named(result, c("summary", "by_function", "improved", "regressed"))
  expect_s3_class(result$summary, "data.frame")
  expect_s3_class(result$by_function, "data.frame")
  expect_s3_class(result$improved, "data.frame")
  expect_s3_class(result$regressed, "data.frame")
})

test_that("pv_compare detects improvement", {
  # Create "before" with more samples
  prof_before <- data.frame(
    time = 1:10,
    depth = rep(1L, 10),
    label = rep("slow", 10),
    filename = rep(NA_character_, 10),
    linenum = rep(NA_integer_, 10),
    filenum = rep(NA_integer_, 10),
    memalloc = seq(100, 1000, by = 100),
    meminc = rep(100, 10),
    stringsAsFactors = FALSE
  )

  # Create "after" with fewer samples
  prof_after <- data.frame(
    time = 1:3,
    depth = rep(1L, 3),
    label = rep("slow", 3),
    filename = rep(NA_character_, 3),
    linenum = rep(NA_integer_, 3),
    filenum = rep(NA_integer_, 3),
    memalloc = c(100, 200, 300),
    meminc = rep(100, 3),
    stringsAsFactors = FALSE
  )

  p1 <- mock_profvis(prof = prof_before, files = list())
  p2 <- mock_profvis(prof = prof_after, files = list())

  result <- pv_compare(p1, p2)

  # After should be faster (less time)
  expect_lt(result$summary$after[1], result$summary$before[1])
})

test_that("pv_compare detects regression", {
  # Create "before" with fewer samples
  prof_before <- data.frame(
    time = 1:3,
    depth = rep(1L, 3),
    label = rep("func", 3),
    filename = rep(NA_character_, 3),
    linenum = rep(NA_integer_, 3),
    filenum = rep(NA_integer_, 3),
    memalloc = c(100, 200, 300),
    meminc = rep(100, 3),
    stringsAsFactors = FALSE
  )

  # Create "after" with more samples (slower)
  prof_after <- data.frame(
    time = 1:10,
    depth = rep(1L, 10),
    label = rep("func", 10),
    filename = rep(NA_character_, 10),
    linenum = rep(NA_integer_, 10),
    filenum = rep(NA_integer_, 10),
    memalloc = seq(100, 1000, by = 100),
    meminc = rep(100, 10),
    stringsAsFactors = FALSE
  )

  p1 <- mock_profvis(prof = prof_before, files = list())
  p2 <- mock_profvis(prof = prof_after, files = list())

  result <- pv_compare(p1, p2)

  # After should be slower (more time)
  expect_gt(result$summary$after[1], result$summary$before[1])
})

test_that("pv_compare respects n parameter", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()

  result <- pv_compare(p1, p2, n = 2)

  expect_lte(nrow(result$by_function), 2)
})

test_that("pv_print_compare snapshot", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()
  expect_snapshot(pv_print_compare(p1, p2))
})

test_that("pv_compare rejects non-profvis input", {
  p <- mock_profvis()

  expect_error(pv_compare(list(), p), "must be a profvis object")
  expect_error(pv_compare(p, "bad"), "must be a profvis object")
})

test_that("pv_compare handles identical profiles", {
  p <- mock_profvis()

  result <- pv_compare(p, p)

  # Same profile should show no change
  expect_equal(result$summary$before[1], result$summary$after[1])
})

make_simple_prof <- function(labels, interval = 10) {
  n <- length(labels)
  data.frame(
    time = seq_len(n),
    depth = rep(1L, n),
    label = labels,
    filename = rep(NA_character_, n),
    linenum = rep(NA_real_, n),
    filenum = rep(NA_real_, n),
    memalloc = seq(100, by = 100, length.out = n),
    meminc = rep(0, n),
    stringsAsFactors = FALSE
  )
}

test_that("pv_compare pct_change is Inf when function only in after", {
  prof_before <- make_simple_prof(rep("common", 5))
  prof_after <- make_simple_prof(c(
    rep("common", 3),
    "new_func",
    "new_func",
    "new_func"
  ))

  p1 <- mock_profvis(prof = prof_before, files = list())
  p2 <- mock_profvis(prof = prof_after, files = list())

  result <- pv_compare(p1, p2)
  new_row <- result$by_function[result$by_function$label == "new_func", ]

  expect_equal(new_row$pct_change, Inf)
})

test_that("pv_print_compare shows improvements and regressions", {
  prof_before <- make_simple_prof(c(rep("slow_func", 5), rep("common", 5)))
  prof_after <- make_simple_prof(c(rep("new_func", 5), rep("common", 5)))

  p1 <- mock_profvis(prof = prof_before, files = list())
  p2 <- mock_profvis(prof = prof_after, files = list())

  expect_snapshot(pv_print_compare(p1, p2))
})

test_that("pv_compare_many returns correct structure", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()

  result <- pv_compare_many(a = p1, b = p2)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("name", "time_ms", "samples", "vs_fastest", "rank"))
})

test_that("pv_compare_many marks fastest correctly", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()

  result <- pv_compare_many(a = p1, b = p2)

  expect_equal(result$rank[1], 1)
  expect_equal(result$vs_fastest[1], "fastest")
})

test_that("pv_compare_many accepts named list", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()

  result <- pv_compare_many(list(a = p1, b = p2))

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
})

test_that("pv_compare_many requires at least 2 profiles", {
  p <- mock_profvis()
  expect_snapshot(error = TRUE, pv_compare_many(a = p))
})

test_that("pv_compare_many requires named profiles", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()
  expect_snapshot(error = TRUE, pv_compare_many(p1, p2))
})

test_that("pv_compare_many rejects non-profvis input", {
  p <- mock_profvis()
  expect_snapshot(error = TRUE, pv_compare_many(a = p, b = list()))
})

test_that("pv_print_compare_many snapshot", {
  p1 <- mock_profvis()
  p2 <- mock_profvis()
  expect_snapshot(pv_print_compare_many(a = p1, b = p2))
})

#' Show source context for a specific location
#'
#' Displays source code around a specific file and line number with profiling
#' information for each line.
#'
#' @param x A profvis object.
#' @param filename The source file to examine.
#' @param linenum The line number to center on. If `NULL`, shows the hottest
#'   line in the file.
#' @param context Number of lines to show before and after.
#'
#' @return Invisibly returns a `debrief_source_context` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_source_context(p, "R/main.R", linenum = 10)
#' @export
pv_source_context <- function(x, filename, linenum = NULL, context = 10) {
  check_profvis(x)
  check_empty_profile(x)

  prof <- extract_prof(x)
  interval_ms <- extract_interval(x)
  total_samples <- extract_total_samples(x)
  files <- extract_files(x)
  file_contents <- build_file_contents(files)

  # Find matching filename (partial match allowed)
  available_files <- unique(prof$filename[!is.na(prof$filename)])
  matching <- grep(filename, available_files, value = TRUE, fixed = TRUE)

  if (length(matching) == 0) {
    obj <- structure(
      list(
        not_found = TRUE,
        available_files = available_files,
        line_data = NULL,
        source_lines = NULL
      ),
      class = "debrief_source_context"
    )
    print(obj)
    return(invisible(obj))
  }

  multiple_match <- length(matching) > 1
  filename <- matching[1]

  # Get profiling data for this file
  file_prof <- prof[!is.na(prof$filename) & prof$filename == filename, ]

  # If no linenum specified, find the hottest line
  auto_linenum <- is.null(linenum)
  if (auto_linenum) {
    # Self-time: top of stack (filter to this file)
    top_of_stack <- extract_top_of_stack(prof)
    top_of_stack <- top_of_stack[
      !is.na(top_of_stack$filename) & top_of_stack$filename == filename,
    ]

    if (nrow(top_of_stack) > 0) {
      line_counts <- table(top_of_stack$linenum)
      linenum <- as.integer(names(which.max(line_counts)))
    } else {
      linenum <- min(file_prof$linenum, na.rm = TRUE)
    }
  }

  # Get line-by-line profiling data
  line_data <- aggregate_lines(file_prof, interval_ms, total_samples)

  # Get source lines
  start_line <- max(1L, linenum - context)
  end_line <- linenum + context
  source_lines <- get_source_lines(
    filename,
    start_line,
    end_line,
    file_contents
  )

  if (is.null(source_lines)) {
    obj <- structure(
      list(
        not_found = FALSE,
        no_source = TRUE,
        line_data = line_data
      ),
      class = "debrief_source_context"
    )
    print(obj)
    return(invisible(obj))
  }

  actual_end <- min(start_line + length(source_lines) - 1, end_line)

  obj <- structure(
    list(
      not_found = FALSE,
      no_source = FALSE,
      multiple_match = multiple_match,
      auto_linenum = auto_linenum,
      filename = filename,
      linenum = linenum,
      start_line = start_line,
      actual_end = actual_end,
      source_lines = source_lines,
      line_data = line_data,
      file_prof = file_prof
    ),
    class = "debrief_source_context"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_source_context <- function(x, ...) {
  if (x$not_found) {
    cat("File not found in profiling data.\n")
    cat("Available files:\n")
    for (f in x$available_files) {
      cat("  ", f, "\n")
    }
    cat("\n")
    cat_help_hint()
    return(invisible(x))
  }

  if (isTRUE(x$no_source)) {
    cat("Source code not available for this file.\n")
    return(invisible(x))
  }

  if (x$multiple_match) {
    cat("Multiple files match. Using:", x$filename, "\n")
  }

  if (x$auto_linenum) {
    cat(sprintf("Showing context around hottest line: %d\n\n", x$linenum))
  }

  filename <- x$filename
  linenum <- x$linenum
  start_line <- x$start_line
  actual_end <- x$actual_end
  source_lines <- x$source_lines
  line_data <- x$line_data
  file_prof <- x$file_prof

  cat_header(sprintf("SOURCE: %s", filename))
  cat("\n")
  cat(sprintf(
    "Lines %d-%d (centered on %d)\n\n",
    start_line,
    actual_end,
    linenum
  ))
  cat("  Time   Mem   Line  Source\n")

  for (i in seq_along(source_lines)) {
    ln <- start_line + i - 1
    line_info <- line_data[line_data$linenum == ln, ]

    if (nrow(line_info) > 0) {
      time_str <- sprintf("%5.0f", line_info$time_ms[1])
      mem_str <- sprintf("%5.1f", line_info$mem_mb[1])
    } else {
      time_str <- "    -"
      mem_str <- "    -"
    }

    marker <- if (ln == linenum) ">" else " "
    cat(sprintf(
      "%s %s %s %4d: %s\n",
      marker,
      time_str,
      mem_str,
      ln,
      source_lines[i]
    ))
  }

  # Next steps - suggest focusing on the function at the hot line
  if (nrow(file_prof) > 0) {
    hot_line_prof <- file_prof[file_prof$linenum == linenum, ]
    if (nrow(hot_line_prof) > 0) {
      func <- hot_line_prof$label[1]
      suggestions <- "pv_hot_lines(p)"
      if (is_user_function(func)) {
        suggestions <- c(sprintf("pv_focus(p, \"%s\")", func), suggestions)
      }
      cat_next_steps(suggestions)
    }
  }

  invisible(x)
}

aggregate_lines <- function(file_prof, interval_ms, total_samples) {
  if (nrow(file_prof) == 0) {
    return(data.frame(
      linenum = integer(),
      samples = integer(),
      time_ms = numeric(),
      pct = numeric(),
      mem_mb = numeric()
    ))
  }

  # Aggregate samples by line
  line_counts <- table(file_prof$linenum)
  mem_sums <- tapply(
    pmax(0, file_prof$meminc),
    file_prof$linenum,
    sum
  )

  result <- data.frame(
    linenum = as.integer(names(line_counts)),
    samples = as.integer(line_counts)
  )
  result$time_ms <- result$samples * interval_ms
  result$pct <- round(100 * result$samples / total_samples, 1)
  result$mem_mb <- as.numeric(mem_sums[as.character(result$linenum)])
  result$mem_mb[is.na(result$mem_mb)] <- 0

  result[order(-result$samples), ]
}

#' Call statistics summary
#'
#' Shows call count, total time, self time, and time per call for each
#' function. This is especially useful for identifying functions that are
#' called many times (where per-call optimization or reducing call count
#' would help).
#'
#' @param x A profvis object.
#' @param n Maximum number of functions to return. If `NULL`, returns all.
#'
#' @return A data frame with columns:
#'   - `label`: Function name
#'   - `calls`: Estimated number of calls (based on stack appearances)
#'   - `total_ms`: Total time on call stack
#'   - `self_ms`: Time at top of stack (self time)
#'   - `child_ms`: Time in callees
#'   - `ms_per_call`: Average milliseconds per call
#'   - `pct`: Percentage of total profile time
#'
#' @examples
#' p <- pv_example()
#' pv_call_stats(p)
#' @export
pv_call_stats <- function(x, n = NULL) {
  pd <- extract_profile_data(x)

  # Get unique functions
  all_funcs <- unique(pd$prof$label)

  # Pre-calculate top of stack (self-time) once for efficiency
  top_of_stack <- extract_top_of_stack(pd$prof)
  top_of_stack_times <- split(top_of_stack$label, top_of_stack$time)

  # Calculate stats for each function
  stats <- lapply(all_funcs, function(func) {
    func_rows <- pd$prof[pd$prof$label == func, ]

    # Total time: unique time points where function appears
    total_times <- unique(func_rows$time)
    total_samples_func <- length(total_times)
    total_ms <- total_samples_func * pd$interval_ms

    # Self time: times when function is at top of stack
    self_samples <- sum(vapply(
      as.character(total_times),
      function(t) func %in% top_of_stack_times[[t]],
      logical(1)
    ))
    self_ms <- self_samples * pd$interval_ms

    # Estimate call count by counting "entries" into the function
    # A new call is when the function appears at a time point where it
    # wasn't at the previous time point, or when depth increases
    call_count <- estimate_call_count(func, pd$prof)

    data.frame(
      label = func,
      calls = call_count,
      total_ms = total_ms,
      self_ms = self_ms,
      child_ms = total_ms - self_ms,
      ms_per_call = if (call_count > 0) total_ms / call_count else 0,
      pct = round(100 * total_samples_func / pd$total_samples, 1)
    )
  })

  result <- do.call(rbind, stats)
  result <- result[order(-result$total_ms), ]
  rownames(result) <- NULL

  if (!is.null(n)) {
    result <- head(result, n)
  }

  result
}

estimate_call_count <- function(func, prof) {
  func_rows <- prof[prof$label == func, ]
  if (nrow(func_rows) == 0) {
    return(0)
  }

  times <- sort(unique(func_rows$time))
  if (length(times) == 0) {
    return(0)
  }

  # Count transitions: function not present -> present
  all_times <- sort(unique(prof$time))
  func_times_set <- times

  count <- 0
  prev_present <- FALSE
  prev_depth <- -1

  for (t in all_times) {
    curr_present <- t %in% func_times_set
    if (curr_present) {
      curr_depth <- max(func_rows$depth[func_rows$time == t])
      # New call if: wasn't present before, or depth increased (re-entry/recursion)
      if (!prev_present) {
        count <- count + 1
      } else if (curr_depth > prev_depth) {
        # Recursion: depth increased
        count <- count + (curr_depth - prev_depth)
      }
      prev_depth <- curr_depth
    } else {
      prev_depth <- -1
    }
    prev_present <- curr_present
  }

  count
}

#' Print call statistics
#'
#' @param x A profvis object.
#' @param n Number of functions to show.
#'
#' @return Invisibly returns a `debrief_call_stats` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_call_stats(p)
#'
#' @export
pv_print_call_stats <- function(x, n = 20) {
  check_profvis(x)
  check_empty_profile(x)

  stats <- pv_call_stats(x, n = n)
  obj <- structure(list(stats = stats), class = "debrief_call_stats")
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_call_stats <- function(x, ...) {
  stats <- x$stats

  if (nrow(stats) == 0) {
    cat("No profiling data available.\n")
    cat_help_hint()
    return(invisible(x))
  }

  cat_header("CALL STATISTICS")
  cat("\n")
  cat(sprintf(
    "%-35s %8s %10s %10s %10s %6s\n",
    "Function",
    "Calls",
    "Total ms",
    "Self ms",
    "ms/call",
    "Pct"
  ))

  for (i in seq_len(nrow(stats))) {
    row <- stats[i, ]
    label <- truncate_string(row$label, 35)
    cat(sprintf(
      "%-35s %8d %10.0f %10.0f %10.2f %5.1f%%\n",
      label,
      row$calls,
      row$total_ms,
      row$self_ms,
      row$ms_per_call,
      row$pct
    ))
  }

  # Next steps
  if (nrow(stats) > 0) {
    top_func <- stats$label[1]
    if (is_user_function(top_func)) {
      cat_next_steps(c(
        sprintf("pv_focus(p, \"%s\")", top_func),
        sprintf("pv_callers(p, \"%s\")", top_func)
      ))
    }
  }

  invisible(x)
}

#' Comprehensive profiling data
#'
#' Returns all profiling analysis in a single list for programmatic access.
#' This is the primary function for AI agents and scripts that need
#' comprehensive profiling data without printed output.
#'
#' @param x A profvis object from [profvis::profvis()].
#' @param n Maximum number of items to include in each category (default 10).
#'
#' @return A list containing:
#'   - `total_time_ms`: Total profiling time in milliseconds
#'   - `total_samples`: Number of profiling samples
#'   - `interval_ms`: Sampling interval in milliseconds
#'   - `has_source`: Whether source references are available
#'   - `self_time`: Data frame of functions by self-time
#'   - `total_time`: Data frame of functions by total time
#'   - `hot_lines`: Data frame of hot source lines (or NULL if no source refs)
#'   - `hot_paths`: Data frame of hot call paths
#'   - `suggestions`: Data frame of optimization suggestions
#'   - `gc_pressure`: Data frame of GC pressure analysis
#'   - `memory`: Data frame of memory allocation by function
#'   - `memory_lines`: Data frame of memory allocation by line (or NULL)
#'
#' @examples
#' p <- pv_example()
#' d <- pv_debrief(p)
#' names(d)
#' d$self_time
#' @export
pv_debrief <- function(x, n = 10) {
  check_profvis(x)
  check_empty_profile(x)

  interval_ms <- extract_interval(x)
  total_samples <- extract_total_samples(x)
  total_time_ms <- total_samples * interval_ms

  has_source <- has_source_refs(x)

  list(
    total_time_ms = total_time_ms,
    total_samples = total_samples,
    interval_ms = interval_ms,
    has_source = has_source,
    self_time = pv_self_time(x, n = n),
    total_time = pv_total_time(x, n = n),
    hot_lines = if (has_source) pv_hot_lines(x, n = n) else NULL,
    hot_paths = pv_hot_paths(x, n = n),
    suggestions = pv_suggestions(x),
    gc_pressure = pv_gc_pressure(x),
    memory = pv_memory(x, n = n),
    memory_lines = if (has_source) pv_memory_lines(x, n = n) else NULL
  )
}

#' Print profiling summary
#'
#' Prints a comprehensive text summary of profiling data suitable for
#' terminal output or AI agent consumption.
#'
#' @param x A profvis object from [profvis::profvis()].
#' @param n_functions Number of top functions to show (default 10).
#' @param n_lines Number of hot source lines to show (default 10).
#' @param n_paths Number of hot paths to show (default 5).
#' @param n_memory Number of memory hotspots to show (default 5).
#'
#' @return Invisibly returns a `debrief_debrief` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_debrief(p)
#' @export
pv_print_debrief <- function(
  x,
  n_functions = 10,
  n_lines = 10,
  n_paths = 5,
  n_memory = 5
) {
  check_profvis(x)
  check_empty_profile(x)

  interval_ms <- extract_interval(x)
  total_samples <- extract_total_samples(x)
  total_time_ms <- total_samples * interval_ms
  has_source <- has_source_refs(x)
  files <- extract_files(x)
  file_contents <- build_file_contents(files)

  # Compute all summaries
  self_time <- pv_self_time(x)
  total_time <- pv_total_time(x)
  hot_lines <- if (has_source) pv_hot_lines(x) else NULL
  hot_paths <- pv_hot_paths(x)
  memory_funcs <- pv_memory(x)
  memory_lines <- if (has_source) pv_memory_lines(x) else NULL

  obj <- structure(
    list(
      interval_ms = interval_ms,
      total_samples = total_samples,
      total_time_ms = total_time_ms,
      has_source = has_source,
      file_contents = file_contents,
      self_time = self_time,
      total_time = total_time,
      hot_lines = hot_lines,
      hot_paths = hot_paths,
      memory_funcs = memory_funcs,
      memory_lines = memory_lines,
      n_functions = n_functions,
      n_lines = n_lines,
      n_paths = n_paths,
      n_memory = n_memory,
      debrief = pv_debrief(
        x,
        n = max(n_functions, n_lines, n_paths, n_memory)
      )
    ),
    class = "debrief_debrief"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_debrief <- function(x, ...) {
  interval_ms <- x$interval_ms
  total_samples <- x$total_samples
  total_time_ms <- x$total_time_ms
  has_source <- x$has_source
  file_contents <- x$file_contents
  self_time <- x$self_time
  total_time <- x$total_time
  hot_lines <- x$hot_lines
  hot_paths <- x$hot_paths
  memory_funcs <- x$memory_funcs
  memory_lines <- x$memory_lines
  n_functions <- x$n_functions
  n_lines <- x$n_lines
  n_paths <- x$n_paths
  n_memory <- x$n_memory

  # Print output
  cat_header("PROFILING SUMMARY")
  cat("\n")

  cat(sprintf(
    "Total time: %.0f ms (%d samples @ %.0f ms interval)\n",
    total_time_ms,
    total_samples,
    interval_ms
  ))
  if (has_source) {
    cat("Source references: available\n")
  } else {
    cat("Source references: not available (use devtools::load_all())\n")
  }
  cat("\n")

  cat_section("TOP FUNCTIONS BY SELF-TIME")
  print_time_df(head(self_time, n_functions))

  cat_section("TOP FUNCTIONS BY TOTAL TIME")
  print_time_df(head(total_time, n_functions))

  if (has_source && !is.null(hot_lines) && nrow(hot_lines) > 0) {
    cat_section("HOT LINES (by self-time)")
    print_lines_df(head(hot_lines, n_lines), file_contents)
  }

  cat_section("HOT CALL PATHS")
  print_paths_df(head(hot_paths, n_paths))

  cat_section("MEMORY ALLOCATION (by function)")
  print_memory_df(head(memory_funcs, n_memory))

  if (has_source && !is.null(memory_lines) && nrow(memory_lines) > 0) {
    cat_section("MEMORY ALLOCATION (by line)")
    print_memory_lines_df(head(memory_lines, n_memory), file_contents)
  }

  # Next steps suggestions
  suggestions <- character()
  if (nrow(self_time) > 0) {
    top_func <- self_time$label[1]
    if (is_user_function(top_func)) {
      suggestions <- c(suggestions, sprintf("pv_focus(p, \"%s\")", top_func))
    }
  }
  if (has_source && !is.null(hot_lines) && nrow(hot_lines) > 0) {
    top_file <- hot_lines$filename[1]
    suggestions <- c(
      suggestions,
      sprintf("pv_source_context(p, \"%s\")", top_file)
    )
  }
  suggestions <- c(suggestions, "pv_suggestions(p)", "pv_help()")
  cat_next_steps(suggestions)

  invisible(x)
}

# Print helpers for debrief
print_time_df <- function(df) {
  if (is.null(df) || nrow(df) == 0) {
    cat("No data.\n")
    return(invisible())
  }
  for (i in seq_len(nrow(df))) {
    cat(sprintf(
      "%6.0f ms (%5.1f%%)  %s\n",
      df$time_ms[i],
      df$pct[i],
      df$label[i]
    ))
  }
}

print_lines_df <- function(df, file_contents) {
  if (is.null(df) || nrow(df) == 0) {
    cat("No source location data available.\n")
    return(invisible())
  }
  for (i in seq_len(nrow(df))) {
    cat(sprintf(
      "%6.0f ms (%5.1f%%)  %s\n",
      df$time_ms[i],
      df$pct[i],
      df$location[i]
    ))
    src_line <- get_source_line(df$filename[i], df$linenum[i], file_contents)
    if (!is.null(src_line) && nchar(src_line) > 0) {
      cat(sprintf("                   %s\n", truncate_string(src_line)))
    }
  }
}

print_paths_df <- function(df) {
  if (is.null(df) || nrow(df) == 0) {
    cat("No data.\n")
    return(invisible())
  }
  for (i in seq_len(nrow(df))) {
    cat(sprintf(
      "\n%.0f ms (%.1f%%) - %d samples:\n",
      df$time_ms[i],
      df$pct[i],
      df$samples[i]
    ))
    parts <- strsplit(df$stack[i], " -> ")[[1]]
    cat("    ", paste(parts, collapse = "\n  -> "), "\n", sep = "")
  }
}

print_memory_df <- function(df) {
  if (is.null(df) || nrow(df) == 0) {
    cat("No significant memory allocations detected.\n")
    return(invisible())
  }
  for (i in seq_len(nrow(df))) {
    cat(fmt_memory(df$mem_mb[i]), " ", df$label[i], "\n", sep = "")
  }
}

print_memory_lines_df <- function(df, file_contents) {
  if (is.null(df) || nrow(df) == 0) {
    cat("No source location data available.\n")
    return(invisible())
  }
  for (i in seq_len(nrow(df))) {
    cat(fmt_memory(df$mem_mb[i]), " ", df$location[i], "\n", sep = "")
    src_line <- get_source_line(df$filename[i], df$linenum[i], file_contents)
    if (!is.null(src_line) && nchar(src_line) > 0) {
      cat(sprintf("            %s\n", truncate_string(src_line, 58)))
    }
  }
}

#' Detect GC pressure
#'
#' Analyzes the profile to detect excessive garbage collection, which is a
#' universal indicator of memory allocation issues in R code.
#'
#' @param x A profvis object.
#' @param threshold Minimum GC percentage to report (default 10). Set lower to
#'   detect smaller GC overhead.
#'
#' @return A data frame with columns:
#'   - `severity`: "high" (>25%), "medium" (>15%), or "low" (>threshold%)
#'   - `pct`: Percentage of total time spent in GC
#'   - `time_ms`: Time spent in garbage collection
#'   - `issue`: Short description of the problem
#'   - `cause`: What typically causes this issue
#'   - `actions`: Comma-separated list of things to look for
#'
#' Returns an empty data frame (0 rows) if GC is below the threshold.
#'
#' @details
#' GC pressure above 10% typically indicates the code is allocating and
#' discarding memory faster than necessary. Common causes include:
#' - Growing vectors with `c(x, new)` instead of pre-allocation
#' - Building data frames row-by-row with `rbind()`
#' - Creating unnecessary copies of large objects
#' - String concatenation in loops
#'
#' @examples
#' p <- pv_example("gc")
#' pv_gc_pressure(p)
#'
#' # More sensitive detection
#' pv_gc_pressure(p, threshold = 5)
#'
#' # No GC pressure in default example
#' p2 <- pv_example()
#' pv_gc_pressure(p2)
#' @export
pv_gc_pressure <- function(x, threshold = 10) {
  pd <- extract_profile_data(x)

  # Detect GC pressure - the one universal anti-pattern signal
  gc_times <- unique(pd$prof$time[pd$prof$label == "<GC>"])

  if (length(gc_times) == 0) {
    return(empty_gc_pressure_df())
  }

  time_ms <- length(gc_times) * pd$interval_ms
  pct <- round(100 * length(gc_times) / pd$total_samples, 1)

  # Only report if GC exceeds threshold
  if (pct <= threshold) {
    return(empty_gc_pressure_df())
  }

  severity <- if (pct > 25) {
    "high"
  } else if (pct > 15) {
    "medium"
  } else {
    "low"
  }

  data.frame(
    severity = severity,
    pct = pct,
    time_ms = time_ms,
    issue = sprintf("High GC overhead (%.1f%%)", pct),
    cause = "Excessive memory allocation",
    actions = "growing vectors, repeated data frame ops, unnecessary copies"
  )
}

empty_gc_pressure_df <- function() {
  data.frame(
    severity = character(),
    pct = numeric(),
    time_ms = numeric(),
    issue = character(),
    cause = character(),
    actions = character()
  )
}

#' Print GC pressure analysis
#'
#' @param x A profvis object.
#' @param threshold Minimum GC percentage to report (default 10).
#'
#' @return Invisibly returns a `debrief_gc_pressure` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example("gc")
#' pv_print_gc_pressure(p)
#'
#' @export
pv_print_gc_pressure <- function(x, threshold = 10) {
  check_profvis(x)
  check_empty_profile(x)

  gc_data <- pv_gc_pressure(x, threshold = threshold)
  obj <- structure(
    list(gc_data = gc_data, threshold = threshold),
    class = "debrief_gc_pressure"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_gc_pressure <- function(x, ...) {
  gc_data <- x$gc_data
  threshold <- x$threshold

  cat_header("GC PRESSURE")
  cat("\n")

  if (nrow(gc_data) == 0) {
    cat(sprintf(
      "No significant GC pressure detected (<%d%% of time).\n",
      threshold
    ))
    cat_help_hint()
    return(invisible(x))
  }

  row <- gc_data[1, ]
  cat(sprintf("severity: %s\n", row$severity))
  cat(sprintf("pct: %.1f\n", row$pct))
  cat(sprintf("time_ms: %.0f\n", row$time_ms))
  cat(sprintf("issue: %s\n", row$issue))
  cat(sprintf("cause: %s\n", row$cause))
  cat(sprintf("actions: %s\n", row$actions))

  # Next steps
  cat_next_steps(c(
    "pv_print_memory(p, by = \"function\")",
    "pv_print_memory(p, by = \"line\")",
    "pv_suggestions(p)"
  ))

  invisible(x)
}

#' Get callers of a function
#'
#' Returns the functions that call a specified function, based on profiling
#' data. Shows who invokes the target function.
#'
#' @param x A profvis object.
#' @param func The function name to analyze.
#'
#' @return A data frame with columns:
#'   - `label`: Caller function name
#'   - `samples`: Number of times this caller appeared
#'   - `pct`: Percentage of calls from this caller
#'
#' @examples
#' p <- pv_example()
#' pv_callers(p, "inner")
#' @export
pv_callers <- function(x, func) {
  find_adjacent_functions(x, func, direction = "caller")
}

#' Get callees of a function
#'
#' Returns the functions that a specified function calls, based on profiling
#' data. Shows what the target function invokes.
#'
#' @param x A profvis object.
#' @param func The function name to analyze.
#'
#' @return A data frame with columns:
#'   - `label`: Callee function name
#'   - `samples`: Number of times this callee appeared
#'   - `pct`: Percentage of calls to this callee
#'
#' @examples
#' p <- pv_example()
#' pv_callees(p, "outer")
#' @export
pv_callees <- function(x, func) {
  find_adjacent_functions(x, func, direction = "callee")
}

# Shared implementation for callers/callees
find_adjacent_functions <- function(
  x,
  func,
  direction = c("caller", "callee")
) {
  check_profvis(x)
  check_empty_profile(x)
  direction <- match.arg(direction)

  prof <- extract_prof(x)
  target_times <- unique(prof$time[prof$label == func])

  if (length(target_times) == 0) {
    message(sprintf("Function '%s' not found in profiling data.", func))
    return(empty_label_samples_pct())
  }

  # Configuration based on direction
  depth_offset <- if (direction == "caller") -1L else 1L
  use_first_occurrence <- direction == "caller"
  default_value <- if (direction == "caller") "(top-level)" else NULL

  # Find adjacent function for each time point
  adjacent <- lapply(target_times, function(t) {
    stack <- prof[prof$time == t, ]
    stack <- stack[order(stack$depth), ]

    target_idx <- which(stack$label == func)
    if (length(target_idx) == 0) {
      return(NULL)
    }

    # Use first or last occurrence based on direction
    idx <- if (use_first_occurrence) {
      target_idx[1]
    } else {
      target_idx[length(target_idx)]
    }
    target_depth <- stack$depth[idx]
    adjacent_row <- stack[stack$depth == target_depth + depth_offset, ]

    if (nrow(adjacent_row) > 0) {
      adjacent_row$label[1]
    } else {
      default_value
    }
  })

  adjacent <- unlist(adjacent)
  if (length(adjacent) == 0) {
    return(empty_label_samples_pct())
  }

  counts <- table(adjacent)
  result <- data.frame(
    label = names(counts),
    samples = as.integer(counts)
  )

  # Percentage denominator differs: callers use sum, callees use target_times
  denom <- if (direction == "caller") {
    sum(result$samples)
  } else {
    length(target_times)
  }
  result$pct <- round(100 * result$samples / denom, 1)
  result <- result[order(-result$samples), ]
  rownames(result) <- NULL
  result
}

#' Print caller/callee analysis for a function
#'
#' Shows both callers (who calls this function) and callees (what this function
#' calls) in a single view.
#'
#' @param x A profvis object.
#' @param func The function name to analyze.
#' @param n Maximum number of callers/callees to show.
#'
#' @return Invisibly returns a `debrief_callers_callees` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_callers_callees(p, "inner")
#' @export
pv_print_callers_callees <- function(x, func, n = 10) {
  check_profvis(x)
  check_empty_profile(x)

  callers <- pv_callers(x, func)
  callees <- pv_callees(x, func)

  pd <- extract_profile_data(x)

  # Get time stats for this function
  func_times <- unique(pd$prof$time[pd$prof$label == func])
  func_total_time <- length(func_times) * pd$interval_ms
  func_pct <- round(100 * length(func_times) / pd$total_samples, 1)

  obj <- structure(
    list(
      callers = callers,
      callees = callees,
      func = func,
      func_times = func_times,
      func_total_time = func_total_time,
      func_pct = func_pct,
      n = n
    ),
    class = "debrief_callers_callees"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_callers_callees <- function(x, ...) {
  callers <- x$callers
  callees <- x$callees
  func <- x$func
  func_times <- x$func_times
  func_total_time <- x$func_total_time
  func_pct <- x$func_pct
  n <- x$n

  cat_header(sprintf("FUNCTION ANALYSIS: %s", func))
  cat("\n")
  cat(sprintf(
    "Total time: %.0f ms (%.1f%% of profile)\n",
    func_total_time,
    func_pct
  ))
  cat(sprintf("Appearances: %d samples\n\n", length(func_times)))

  cat("### Called by\n")
  if (nrow(callers) == 0) {
    cat("  Callers: none\n")
  } else {
    for (i in seq_len(min(n, nrow(callers)))) {
      cat(sprintf(
        "  %5d samples (%5.1f%%)  %s\n",
        callers$samples[i],
        callers$pct[i],
        callers$label[i]
      ))
    }
  }

  cat("\n### Calls to\n")
  if (nrow(callees) == 0) {
    cat("  Callees: none\n")
  } else {
    for (i in seq_len(min(n, nrow(callees)))) {
      cat(sprintf(
        "  %5d samples (%5.1f%%)  %s\n",
        callees$samples[i],
        callees$pct[i],
        callees$label[i]
      ))
    }
  }

  # Next steps suggestions
  suggestions <- c(sprintf("pv_focus(p, \"%s\")", func))
  if (nrow(callers) > 0 && is_user_function(callers$label[1])) {
    suggestions <- c(
      suggestions,
      sprintf("pv_focus(p, \"%s\")", callers$label[1])
    )
  }
  if (nrow(callees) > 0 && is_user_function(callees$label[1])) {
    suggestions <- c(
      suggestions,
      sprintf("pv_focus(p, \"%s\")", callees$label[1])
    )
  }
  cat_next_steps(suggestions)

  invisible(x)
}

#' Memory allocation by function
#'
#' Returns memory allocation aggregated by function name.
#'
#' @param x A profvis object.
#' @param n Maximum number of functions to return. If `NULL`, returns all.
#'
#' @return A data frame with columns:
#'   - `label`: Function name
#'   - `mem_mb`: Memory allocated in megabytes
#'
#' @examples
#' p <- pv_example()
#' pv_memory(p)
#' @export
pv_memory <- function(x, n = NULL) {
  check_profvis(x)
  check_empty_profile(x)

  prof <- extract_prof(x)

  # Sum positive memory increments by function
  prof_pos <- prof[prof$meminc > 0, ]
  if (nrow(prof_pos) == 0) {
    return(empty_memory_result())
  }

  mem_sums <- tapply(prof_pos$meminc, prof_pos$label, sum)
  result <- data.frame(
    label = names(mem_sums),
    mem_mb = as.numeric(mem_sums)
  )
  result <- result[order(-result$mem_mb), ]
  rownames(result) <- NULL

  if (!is.null(n)) {
    result <- head(result, n)
  }

  result
}

#' Memory allocation by source line
#'
#' Returns memory allocation aggregated by source location. Requires source
#' references (use `devtools::load_all()` for best results).
#'
#' @param x A profvis object.
#' @param n Maximum number of lines to return. If `NULL`, returns all.
#'
#' @return A data frame with columns:
#'   - `location`: File path and line number
#'   - `label`: Function name at this location
#'   - `filename`: Source file path
#'   - `linenum`: Line number
#'   - `mem_mb`: Memory allocated in megabytes
#'
#' @examples
#' p <- pv_example()
#' pv_memory_lines(p)
#' @export
pv_memory_lines <- function(x, n = NULL) {
  check_profvis(x)
  check_empty_profile(x)

  prof <- extract_prof(x)

  # Sum positive memory increments by source location
  prof_pos <- prof[prof$meminc > 0 & !is.na(prof$filename), ]
  if (nrow(prof_pos) == 0) {
    return(empty_memory_lines_result())
  }

  prof_pos$location <- paste0(prof_pos$filename, ":", prof_pos$linenum)

  mem_sums <- tapply(prof_pos$meminc, prof_pos$location, sum)
  result <- data.frame(
    location = names(mem_sums),
    mem_mb = as.numeric(mem_sums)
  )

  # Get label, filename, linenum for each location
  first_occurrence <- prof_pos[!duplicated(prof_pos$location), ]
  result <- merge(
    result,
    first_occurrence[, c("location", "label", "filename", "linenum")],
    by = "location"
  )

  result <- result[order(-result$mem_mb), ]
  rownames(result) <- NULL

  if (!is.null(n)) {
    result <- head(result, n)
  }

  result
}

#' Print memory allocation summary
#'
#' @param x A profvis object.
#' @param n Number of top allocators to show.
#' @param by Either "function" or "line".
#'
#' @return Invisibly returns a `debrief_memory` object. Use
#'   `capture.output()` to capture the formatted text output.
#'
#' @examples
#' p <- pv_example()
#' pv_print_memory(p, by = "function")
#' pv_print_memory(p, by = "line")
#'
#' @export
pv_print_memory <- function(x, n = 10, by = c("function", "line")) {
  check_profvis(x)
  check_empty_profile(x)
  by <- match.arg(by)

  if (by == "function") {
    mem_df <- pv_memory(x, n = n)
    file_contents <- NULL
  } else {
    mem_df <- pv_memory_lines(x, n = n)
    files <- extract_files(x)
    file_contents <- build_file_contents(files)
  }

  obj <- structure(
    list(mem_df = mem_df, by = by, file_contents = file_contents),
    class = "debrief_memory"
  )
  print(obj)
  invisible(obj)
}

#' @exportS3Method
print.debrief_memory <- function(x, ...) {
  mem_df <- x$mem_df
  by <- x$by
  file_contents <- x$file_contents

  if (by == "function") {
    if (nrow(mem_df) == 0) {
      cat("No significant memory allocations detected.\n")
      cat_help_hint()
      return(invisible(x))
    }

    cat_header("MEMORY ALLOCATION BY FUNCTION")
    cat("\n")

    for (i in seq_len(nrow(mem_df))) {
      cat(fmt_memory(mem_df$mem_mb[i]), " ", mem_df$label[i], "\n", sep = "")
    }
  } else {
    if (nrow(mem_df) == 0) {
      cat("No source location data available.\n")
      cat("Use devtools::load_all() to enable source references.\n")
      cat_help_hint()
      return(invisible(x))
    }

    cat_header("MEMORY ALLOCATION BY LINE")
    cat("\n")

    for (i in seq_len(nrow(mem_df))) {
      row <- mem_df[i, ]
      cat(fmt_memory(row$mem_mb), " ", row$location, "\n", sep = "")
      src_line <- get_source_line(row$filename, row$linenum, file_contents)
      if (!is.null(src_line) && nchar(src_line) > 0) {
        cat(sprintf("            %s\n", truncate_string(src_line, 58)))
      }
    }
  }

  # Next steps suggestions
  if (nrow(mem_df) > 0) {
    top_func <- mem_df$label[1]
    suggestions <- character()
    if (is_user_function(top_func)) {
      suggestions <- c(suggestions, sprintf("pv_focus(p, \"%s\")", top_func))
    }
    if (by == "function") {
      suggestions <- c(suggestions, "pv_gc_pressure(p)")
    } else {
      suggestions <- c(
        suggestions,
        sprintf("pv_source_context(p, \"%s\")", mem_df$filename[1])
      )
    }
    cat_next_steps(suggestions)
  }

  invisible(x)
}

test_that("pv_memory returns correct structure", {
  p <- mock_profvis()
  result <- pv_memory(p)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("label", "mem_mb"))
  expect_type(result$label, "character")
  expect_type(result$mem_mb, "double")
})

test_that("pv_memory respects n parameter", {
  p <- mock_profvis()

  result_all <- pv_memory(p)
  result_2 <- pv_memory(p, n = 2)

  expect_lte(nrow(result_2), 2)
})

test_that("pv_memory only counts positive allocations", {
  p <- mock_profvis()
  result <- pv_memory(p)

  # All memory values should be positive
  expect_true(all(result$mem_mb >= 0))
})

test_that("pv_memory_lines returns correct structure", {
  p <- mock_profvis()
  result <- pv_memory_lines(p)

  expect_s3_class(result, "data.frame")
  expect_named(result, c("location", "mem_mb", "label", "filename", "linenum"))
})

test_that("pv_memory_lines returns empty when no source refs", {
  p <- mock_profvis_no_source()
  result <- pv_memory_lines(p)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_memory_lines respects n parameter", {
  p <- mock_profvis()

  result_all <- pv_memory_lines(p)
  result_2 <- pv_memory_lines(p, n = 2)

  expect_lte(nrow(result_2), 2)
})

test_that("pv_print_memory by function snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_memory(p, by = "function"))
})

test_that("pv_print_memory by line snapshot", {
  p <- mock_profvis()
  expect_snapshot(pv_print_memory(p, by = "line"))
})

test_that("pv_print_memory handles no source refs for line mode", {
  p <- mock_profvis_no_source()
  expect_snapshot(pv_print_memory(p, by = "line"))
})

test_that("memory functions reject non-profvis input", {
  expect_error(pv_memory(list()), "must be a profvis object")
  expect_error(pv_memory_lines("bad"), "must be a profvis object")
  expect_error(pv_print_memory(42), "must be a profvis object")
})

test_that("pv_memory handles profile with no allocations", {
  prof <- data.frame(
    time = c(1L, 2L, 3L),
    depth = c(1L, 1L, 1L),
    label = c("a", "b", "c"),
    filename = rep(NA_character_, 3),
    linenum = rep(NA_integer_, 3),
    filenum = rep(NA_integer_, 3),
    memalloc = c(100, 100, 100),
    meminc = c(0, 0, 0),
    stringsAsFactors = FALSE
  )

  p <- mock_profvis(prof = prof, files = list())
  result <- pv_memory(p)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("pv_print_memory shows no allocations message for by function", {
  prof <- data.frame(
    time = c(1L, 2L),
    depth = c(1L, 1L),
    label = c("func", "func"),
    filename = rep(NA_character_, 2),
    linenum = rep(NA_real_, 2),
    filenum = rep(NA_real_, 2),
    memalloc = c(100, 100),
    meminc = c(0, 0),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_print_memory(p, by = "function"))
})

test_that("pv_print_debrief handles profile with no memory allocations", {
  prof <- data.frame(
    time = c(1L, 1L, 2L, 2L),
    depth = c(1L, 2L, 1L, 2L),
    label = c("outer", "inner", "outer", "inner"),
    filename = rep(NA_character_, 4),
    linenum = rep(NA_real_, 4),
    filenum = rep(NA_real_, 4),
    memalloc = c(100, 100, 100, 100),
    meminc = c(0, 0, 0, 0),
    stringsAsFactors = FALSE
  )
  p <- mock_profvis(prof = prof, files = list())
  expect_snapshot(pv_print_debrief(p))
})

# BRIEF: debrief

> Dynamically generated BRIEF.md summaries for your Python projects, to make them accessible to new contributors or coding agents.

## 1. Overview (README)

```markdown
# debrief: Project Summarizer

`debrief` generates a comprehensive `BRIEF.md` file designed to provide new contributors or coding agents with a high-density overview of a Python project.

## Installation

'''bash
pip install debrief
'''

## Usage

'''bash
debrief [path] [--output BRIEF.md]
'''

### Arguments

| Argument               | Description                                | Default         |
| :--------------------- | :----------------------------------------- | :-------------- |
... (truncated) [Read more](file:///D:/_DISK_/_Documentos_/Mis_repositorios/debrief/README.md)
```

## 2. Dependencies

```text
_No dependencies found._
```

## 3. Directory Structure

```text
debrief/
├── debrief
    ├── __init__.py
    ├── analysis.py
    ├── constants.py
    ├── ignore.py
    ├── linting.py
    ├── main.py
    ├── resolve.py
    └── tree.py
├── .gitignore
├── .python-version
├── README.md
├── pyproject.toml
└── uv.lock
```

## 4. Import Tree

```text
- debrief\__init__.py

- debrief\main.py
  - debrief\analysis.py
    - debrief\constants.py
    - debrief\ignore.py
      - debrief\constants.py (...)
    - debrief\resolve.py
  - debrief\ignore.py (...)
  - debrief\linting.py
    - debrief\constants.py (...)
    - debrief\resolve.py (...)
  - debrief\resolve.py (...)
  - debrief\tree.py
    - debrief\ignore.py (...)
```

## 5. Module Definitions

### debrief\analysis.py

```text
- class Analyzer
  - def get_arg_str(self, arg)
  - def scan(self)
  - def analyze_file(self, path, rel_path)
  - def get_import_tree(self)
```

### debrief\ignore.py

```text
- def load_gitignore(root_path)
- def is_ignored(path, root_path, patterns)
```

### debrief\linting.py

```text
- class ProjectLinter
  - def check_metadata(self)
  - def track_doc(self, node)
  - def print_summary(self)
```

### debrief\main.py

```text
- def parse_arguments()
- def main()
```

### debrief\resolve.py

```text
- def log(level: str, msg: str) -> None
- def resolve_readme(root: str) -> Optional[str]
- def resolve_pyproject(root: str) -> Optional[str]
- def resolve_requirements(root: str) -> Optional[str]
- def check_dependencies(root: str) -> None
- def get_project_description(root: str) -> Optional[str]
- def get_project_dependencies(root: str) -> List[str]
- def truncate_line(line: str, max_chars: int) -> str
- def format_fenced_block(content: str, lang: str) -> str
```

### debrief\tree.py

```text
- def generate_tree_at_depth(root_path, max_depth, patterns, max_siblings)
- def get_adaptive_tree(root_path, max_lines, patterns, max_siblings)
```

---
title: "Case study: Using debrief for iterative R performance optimization"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Case study: Using debrief for iterative R performance optimization}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

## About this case study

This case study was conducted by Claude (an AI assistant) to optimize tidypredict's random forest parsing. Two important notes about the debrief package usage:

1. **No prior knowledge**: Claude had no prior knowledge of the debrief package before this session. The package was not part of Claude's training data.

2. **No documentation lookup**: Claude did not read debrief's documentation, README, or help files. The only information provided was the user's prompt showing the basic usage pattern (`pv_print_debrief(p)`). All interpretation of the output was done from first principles based on the section names and content.

3. **Test-driven verification**: After each optimization, Claude ran the package's existing test suite (`devtools::test()`) to verify the changes didn't break functionality. All 38 relevant tests (ranger: 18, randomForest: 13, partykit: 7) passed after the optimizations.

This demonstrates that debrief's output is self-explanatory enough to guide optimization work without requiring prior familiarity with the package. The section names (TOP FUNCTIONS BY SELF-TIME, HOT LINES, MEMORY ALLOCATION) clearly communicate what each part shows, and the "Next steps" suggestions in the output hinted at additional functions available.

## Overview

This case study demonstrates how the **debrief** package can guide iterative performance optimization in R. We use tidypredict's random forest parsing as a real-world example, achieving a **331x speedup** through systematic profiling and targeted fixes.

## Problem statement

[tidymodels/orbital#83](https://github.com/tidymodels/orbital/issues/83) reported that parsing random forest models in tidypredict takes 45+ seconds for a single tree. We rolled back to the problematic version (commit `0dde432`) to investigate.

### User prompt

The user provided the following instructions to guide the profiling session:

````md
profile the tidypredict_fit call in the following code.

library(tidymodels)
library(orbital)

rec_spec <- recipe(Sale_Price ~ ., data = ames) |>
  step_normalize(all_numeric_predictors()) |>
  step_dummy(all_nominal_predictors())

mod_spec <- rand_forest("regression", trees = 1)
wf_spec <- workflow(rec_spec, mod_spec)
wf_fit <- fit(wf_spec, ames)
rf_tree <- extract_fit_engine(wf_fit)

tmp <- tidypredict_fit(rf_tree)

I want you to use the profvis and debrief package to iteratively make improvements to the function

library(profvis)
library(debrief)

# Profile some code
p <- profvis({
  # your code here
})

pv_print_debrief(p)
````

This prompt established the workflow: profile with profvis, analyze with `pv_print_debrief()`, optimize, and repeat.

## Iteration 1: Initial profiling

### Running debrief

```{r}
p <- profvis({
  tmp <- tidypredict_fit(rf_tree)
})

pv_print_debrief(p)
```

### Debrief output (key sections)

```
## PROFILING SUMMARY

Total time: 47340 ms (4734 samples @ 10 ms interval)

### TOP FUNCTIONS BY SELF-TIME
 15180 ms ( 32.1%)  [.data.frame
  5280 ms ( 11.2%)  <GC>
  4220 ms (  8.9%)  sys.call
  3950 ms (  8.3%)  [[.data.frame

### HOT LINES (by self-time)
 12710 ms ( 26.8%)  /R/model-ranger.R:6
                   row <- tree[tree$nodeID == j, ]
  2880 ms (  6.1%)  /R/model-ranger.R:7
                   lc <- row["leftChild"][[1]] == find
  2580 ms (  5.4%)  /R/model-ranger.R:8
                   lr <- row["rightChild"][[1]] == find

### MEMORY ALLOCATION (by line)
15655.00 MB /R/model-ranger.R:6
            row <- tree[tree$nodeID == j, ]
 3450.47 MB /R/model-ranger.R:7
            lc <- row["leftChild"][[1]] == find
```

### How I interpreted this output

1. **TOP FUNCTIONS BY SELF-TIME** showed `[.data.frame` at 32.1% - this is the data.frame subsetting operator, suggesting repeated data.frame operations are the culprit.

2. **HOT LINES** pinpointed the exact line: `row <- tree[tree$nodeID == j, ]` consuming 26.8% of total time. This single line was the primary bottleneck.

3. **MEMORY ALLOCATION** revealed this same line was allocating 15.6 GB - explaining the high GC time (11.2%). Each subsetting operation creates a new data.frame copy.

4. **The pattern**: Line 6 is inside a loop (indicated by the variable `j`). Filtering a data.frame with `tree[tree$nodeID == j, ]` is O(n) per call. In a loop, this becomes O(n^2).

### Optimization applied

Pre-extract columns as vectors and use direct indexing:

```{r}
# Before (O(n) per lookup):
row <- tree[tree$nodeID == j, ]
lc <- row["leftChild"][[1]] == find

# After (O(1) per lookup):
idx <- j + 1L
lc <- left_child[idx] == find
```

### Result

| Metric  | Before    | After     |
|:--------|----------:|----------:|
| Time    | 47,340 ms | 280 ms    |
| Memory  | ~18 GB    | ~20 MB    |
| Speedup | -         | **169x**  |

## Iteration 2: Finding the next bottleneck

### Debrief output after first optimization

```
## PROFILING SUMMARY

Total time: 290 ms (29 samples @ 10 ms interval)

### TOP FUNCTIONS BY SELF-TIME
    60 ms ( 20.7%)  is.na
    20 ms (  6.9%)  .f
    20 ms (  6.9%)  if (lc || lr) {
    20 ms (  6.9%)  lc <- left_child[idx] == find

### HOT LINES (by self-time)
    60 ms ( 20.7%)  /R/model-ranger.R:16
                   if (is.na(lc)) {
    20 ms (  6.9%)  /R/model-ranger.R:14
                   lc <- left_child[idx] == find
    20 ms (  6.9%)  /R/model-ranger.R:22
                   if (lc || lr) {
```

### How I interpreted this output

1. **TOP FUNCTIONS BY SELF-TIME** now showed `is.na` as the top consumer at 20.7%. The bottleneck shifted from data.frame operations to NA checking.

2. **HOT LINES** showed the `is.na` checks on lines 16 and 19 were being called repeatedly in the loop.

3. **The insight**: The code was searching for parent nodes by iterating from `node_id` down to 0, checking each potential parent. This linear search could be replaced with a direct lookup.

### Optimization applied

Pre-compute parent relationships once, then traverse directly:

```{r}
# Before: Linear search for parent
for (j in node_id:0) {
  idx <- j + 1L
  lc <- left_child[idx] == find
  lr <- right_child[idx] == find
  if (is.na(lc)) lc <- FALSE
  if (is.na(lr)) lr <- FALSE
  if (lc || lr) { ... }
}

# After: Direct traversal using parent lookup
while (!is.na(parent[current + 1L])) {
  current <- parent[current + 1L]
  path <- c(path, current)
}
```

### Result

| Metric  | Before | After          |
|:--------|-------:|:---------------|
| Time    | 290 ms | 160 ms         |
| Speedup | -      | **45% faster** |

## Iteration 3: Diminishing returns

### Debrief output (using 10 trees for better sampling)

```
## PROFILING SUMMARY

Total time: 1650 ms (165 samples @ 10 ms interval)

### TOP FUNCTIONS BY SELF-TIME
   280 ms ( 17.0%)  <GC>
   200 ms ( 12.1%)  enexpr
   110 ms (  6.7%)  reduce_impl

### HOT LINES (by self-time)
    60 ms (  3.6%)  /R/model-rf.R:80
                   if (.x$op == "less") i <- expr(!!sym(.x$col) < !!.x$val)
    30 ms (  1.8%)  /R/model-rf.R:78
                   if (.x$op == "more") i <- expr(!!sym(.x$col) > !!.x$val)

### TOP FUNCTIONS BY TOTAL TIME
  1020 ms ( 61.8%)  path_formulas
   500 ms ( 30.3%)  expr
```

### How I interpreted this output

1. **TOP FUNCTIONS BY TOTAL TIME** showed `path_formulas` at 61.8% - the bottleneck shifted to formula building.

2. **HOT LINES** showed multiple if-statements checking `.x$op` repeatedly (lines 78-81).

3. **TOP FUNCTIONS BY SELF-TIME** showed `enexpr` (expression building) and `reduce_impl` as top consumers - these are doing useful work.

4. **The insight**: The if-chain could be replaced with `switch()`, but we're now optimizing code that's doing inherent work. Further gains will be marginal.

### Optimization applied

Replace if-chain with switch statement:

```{r}
# Before:
if (.x$op == "more") i <- expr(!!sym(.x$col) > !!.x$val)
if (.x$op == "more-equal") i <- expr(!!sym(.x$col) >= !!.x$val)
if (.x$op == "less") i <- expr(!!sym(.x$col) < !!.x$val)
if (.x$op == "less-equal") i <- expr(!!sym(.x$col) <= !!.x$val)

# After:
switch(
  .x$op,
  "more" = expr(!!col_sym > !!val),
  "more-equal" = expr(!!col_sym >= !!val),
  "less" = expr(!!col_sym < !!val),
  "less-equal" = expr(!!col_sym <= !!val)
)
```

### Result

Minor improvement (~4%). The remaining time is in expression building which is inherent to the task.

## Why I only used `pv_print_debrief()`

The debrief package offers several functions, and `pv_print_debrief()` suggests others at the end of its output:

```
### Next steps
pv_source_context(p, "/R/model-ranger.R")
pv_suggestions(p)
pv_focus(p, ".f")
pv_help()
```

### Functions I didn't use

| Function              | Purpose                                    | Why I skipped it                                                                                                         |
|:----------------------|:-------------------------------------------|:-------------------------------------------------------------------------------------------------------------------------|
| `pv_source_context()` | Shows more code context around hot lines   | I was already reading full source files directly. Once HOT LINES said "line 6 in model-ranger.R", I just read that file. |
| `pv_suggestions()`    | Provides automated optimization suggestions| The bottlenecks were obvious enough (32% in `[.data.frame`, 15.6 GB on one line) that I could diagnose the issue myself. |
| `pv_focus()`          | Filters analysis to a specific function    | The initial output was already clear about where problems were. No need to drill down.                                   |

### When `pv_print_debrief()` alone is sufficient

The single function was enough because:

1. **HOT LINES gave exact locations** - File path and line number meant I could go straight to the problem code
2. **MEMORY ALLOCATION correlated with time** - The same lines appeared in both sections, confirming the diagnosis
3. **Patterns were recognizable** - Data.frame subsetting in a loop is a known anti-pattern

### When to use the other functions

I would reach for additional functions if:

- **Diffuse profile** - Time spread across many functions without a clear culprit
- **Complex call paths** - Need to understand how a bottleneck is reached from different entry points
- **Unfamiliar codebase** - `pv_source_context()` would help see surrounding code without switching tools
- **Stuck on diagnosis** - `pv_suggestions()` might catch patterns I missed

## Final results

| Stage                | Time       | Speedup  | Key debrief insight    |
|:---------------------|-----------:|---------:|:-----------------------|
| Original             | 47,340 ms  | 1x       | `[.data.frame` at 32%  |
| After vectorization  | 280 ms     | 169x     | `is.na` now at 20%     |
| After parent lookup  | 160 ms     | 296x     | `path_formulas` at 60% |
| After switch         | **143 ms** | **331x** | Diminishing returns    |

# pv_print_gc_pressure snapshot with high GC

    Code
      pv_print_gc_pressure(p)
    Output
      ## GC PRESSURE
      
      
      severity: high
      pct: 40.0
      time_ms: 40
      issue: High GC overhead (40.0%)
      cause: Excessive memory allocation
      actions: growing vectors, repeated data frame ops, unnecessary copies
      
      ### Next steps
      pv_print_memory(p, by = "function")
      pv_print_memory(p, by = "line")
      pv_suggestions(p)

# pv_print_gc_pressure snapshot with no GC

    Code
      pv_print_gc_pressure(p)
    Output
      ## GC PRESSURE
      
      
      No significant GC pressure detected (<10% of time).
      Run pv_help() to see all available functions.

# pv_print_suggestions snapshot with GC pressure

    Code
      pv_print_suggestions(p)
    Output
      ## OPTIMIZATION SUGGESTIONS
      
      
      ### Priority 1
      
      category: hot line
      location: R/work.R:5
      action: Optimize hot line (60.0%)
      pattern: work
      potential_impact: 60 ms (60.0%)
      
      ### Priority 2
      
      category: memory
      location: memory allocation hotspots
      action: Reduce memory allocation
      pattern: c(x, new), rbind(), growing vectors
      replacement: pre-allocate to final size
      potential_impact: Up to 20 ms (20%)
      
      category: hot function
      location: work
      action: Profile in isolation (60.0% self-time)
      pattern: work
      potential_impact: 60 ms (60.0%)
      
      
      ### Next steps
      pv_hot_lines(p)
      pv_gc_pressure(p)

# pv_print_suggestions handles profile with no suggestions

    Code
      pv_print_suggestions(p)
    Output
      ## OPTIMIZATION SUGGESTIONS
      
      
      ### Priority 2
      
      category: hot function
      location: x
      action: Profile in isolation (100.0% self-time)
      pattern: x
      potential_impact: 10 ms (100.0%)
      
      
      ### Next steps
      pv_focus(p, "x")
      pv_gc_pressure(p)

# pv_print_suggestions handles profile with truly no suggestions

    Code
      pv_print_suggestions(p)
    Output
      ## OPTIMIZATION SUGGESTIONS
      
      
      No suggestions.
      Run pv_help() to see all available functions.

# pv_source_context snapshot

    Code
      pv_source_context(p, "R/main.R")
    Output
      Showing context around hottest line: 10
      
      ## SOURCE: R/main.R
      
      
      Lines 1-15 (centered on 10)
      
        Time   Mem   Line  Source
            -     -    1: # Main file
            -     -    2: outer <- function() {
            -     -    3:   x <- 1
            -     -    4:   y <- 2
            -     -    5:   inner()
            -     -    6: }
            -     -    7: 
            -     -    8: inner <- function() {
            -     -    9:   result <- deep()
      >    50   0.0   10:   result
            -     -   11: }
            -     -   12: 
            -     -   13: 
            -     -   14: 
           30 150.0   15:   z <- heavy_computation()
      
      ### Next steps
      pv_focus(p, "outer")
      pv_hot_lines(p)

# pv_source_context handles non-existent file

    Code
      pv_source_context(p, "nonexistent.R")
    Output
      File not found in profiling data.
      Available files:
         R/main.R 
         R/utils.R 
         R/helper.R 
      
      Run pv_help() to see all available functions.

# pv_source_context auto-selects hottest line when linenum is NULL

    Code
      pv_source_context(p, "R/main.R", linenum = NULL)
    Output
      Showing context around hottest line: 10
      
      ## SOURCE: R/main.R
      
      
      Lines 1-15 (centered on 10)
      
        Time   Mem   Line  Source
            -     -    1: # Main file
            -     -    2: outer <- function() {
            -     -    3:   x <- 1
            -     -    4:   y <- 2
            -     -    5:   inner()
            -     -    6: }
            -     -    7: 
            -     -    8: inner <- function() {
            -     -    9:   result <- deep()
      >    50   0.0   10:   result
            -     -   11: }
            -     -   12: 
            -     -   13: 
            -     -   14: 
           30 150.0   15:   z <- heavy_computation()
      
      ### Next steps
      pv_focus(p, "outer")
      pv_hot_lines(p)

# pv_source_context respects linenum parameter

    Code
      pv_source_context(p, "R/main.R", linenum = 5)
    Output
      ## SOURCE: R/main.R
      
      
      Lines 1-15 (centered on 5)
      
        Time   Mem   Line  Source
            -     -    1: # Main file
            -     -    2: outer <- function() {
            -     -    3:   x <- 1
            -     -    4:   y <- 2
      >     -     -    5:   inner()
            -     -    6: }
            -     -    7: 
            -     -    8: inner <- function() {
            -     -    9:   result <- deep()
           50   0.0   10:   result
            -     -   11: }
            -     -   12: 
            -     -   13: 
            -     -   14: 
           30 150.0   15:   z <- heavy_computation()

# pv_print_file_summary snapshot

    Code
      pv_print_file_summary(p)
    Output
      ## FILE SUMMARY
      
      
          50 ms (100.0%)  R/main.R
          20 ms ( 40.0%)  R/utils.R
          10 ms ( 20.0%)  R/helper.R
      
      ### Next steps
      pv_source_context(p, "R/main.R")
      pv_hot_lines(p)

# pv_print_file_summary handles no source refs

    Code
      pv_print_file_summary(p)
    Output
      No source location data available.
      Use devtools::load_all() to enable source references.
      Run pv_help() to see all available functions.

# pv_source_context warns when multiple files match

    Code
      pv_source_context(p, "R/")
    Output
      Multiple files match. Using: R/main.R 
      Showing context around hottest line: 10
      
      ## SOURCE: R/main.R
      
      
      Lines 1-15 (centered on 10)
      
        Time   Mem   Line  Source
            -     -    1: # Main file
            -     -    2: outer <- function() {
            -     -    3:   x <- 1
            -     -    4:   y <- 2
            -     -    5:   inner()
            -     -    6: }
            -     -    7: 
            -     -    8: inner <- function() {
            -     -    9:   result <- deep()
      >    50   0.0   10:   result
            -     -   11: }
            -     -   12: 
            -     -   13: 
            -     -   14: 
           30 150.0   15:   z <- heavy_computation()
      
      ### Next steps
      pv_focus(p, "outer")
      pv_hot_lines(p)

# pv_source_context uses min linenum when no top_of_stack for file

    Code
      pv_source_context(p, "R/main.R", linenum = NULL)
    Output
      Showing context around hottest line: 10
      
      ## SOURCE: R/main.R
      
      
      Lines 1-15 (centered on 10)
      
        Time   Mem   Line  Source
            -     -    1: line1
            -     -    2: line2
            -     -    3: line3
            -     -    4: line4
            -     -    5: line5
            -     -    6: line6
            -     -    7: line7
            -     -    8: line8
            -     -    9: line9
      >    10   0.0   10: line10
            -     -   11: line11
            -     -   12: line12
            -     -   13: line13
            -     -   14: line14
            -     -   15: line15
      
      ### Next steps
      pv_focus(p, "outer")
      pv_hot_lines(p)

# pv_source_context shows source not available when files empty

    Code
      pv_source_context(p, "R/main.R", linenum = 5)
    Output
      Source code not available for this file.

# pv_print_debrief snapshot

    Code
      pv_print_debrief(p)
    Output
      ## PROFILING SUMMARY
      
      
      Total time: 50 ms (5 samples @ 10 ms interval)
      Source references: available
      
      
      ### TOP FUNCTIONS BY SELF-TIME
          20 ms ( 40.0%)  deep
          10 ms ( 20.0%)  helper
          10 ms ( 20.0%)  inner
          10 ms ( 20.0%)  outer
      
      ### TOP FUNCTIONS BY TOTAL TIME
          50 ms (100.0%)  outer
          30 ms ( 60.0%)  inner
          20 ms ( 40.0%)  deep
          10 ms ( 20.0%)  helper
      
      ### HOT LINES (by self-time)
          20 ms ( 40.0%)  R/utils.R:5
                         x <- rnorm(1000)
          10 ms ( 20.0%)  R/helper.R:20
                         do_work()
          10 ms ( 20.0%)  R/main.R:10
                         result
          10 ms ( 20.0%)  R/main.R:15
                         z <- heavy_computation()
      
      ### HOT CALL PATHS
      
      20 ms (40.0%) - 2 samples:
          outer (R/main.R:10)
        -> inner (R/main.R:15)
        -> deep (R/utils.R:5)
      
      10 ms (20.0%) - 1 samples:
          outer (R/main.R:10)
      
      10 ms (20.0%) - 1 samples:
          outer (R/main.R:10)
        -> helper (R/helper.R:20)
      
      10 ms (20.0%) - 1 samples:
          outer (R/main.R:10)
        -> inner (R/main.R:15)
      
      ### MEMORY ALLOCATION (by function)
        150.00 MB inner
        100.00 MB deep
         50.00 MB helper
      
      ### MEMORY ALLOCATION (by line)
        150.00 MB R/main.R:15
                  z <- heavy_computation()
        100.00 MB R/utils.R:5
                  x <- rnorm(1000)
         50.00 MB R/helper.R:20
                  do_work()
      
      ### Next steps
      pv_focus(p, "deep")
      pv_source_context(p, "R/utils.R")
      pv_suggestions(p)
      pv_help()

# pv_print_debrief handles no source refs

    Code
      pv_print_debrief(p)
    Output
      ## PROFILING SUMMARY
      
      
      Total time: 30 ms (3 samples @ 10 ms interval)
      Source references: not available (use devtools::load_all())
      
      
      ### TOP FUNCTIONS BY SELF-TIME
          10 ms ( 33.3%)  bar
          10 ms ( 33.3%)  baz
          10 ms ( 33.3%)  foo
      
      ### TOP FUNCTIONS BY TOTAL TIME
          30 ms (100.0%)  foo
          10 ms ( 33.3%)  bar
          10 ms ( 33.3%)  baz
      
      ### HOT CALL PATHS
      
      10 ms (33.3%) - 1 samples:
          foo
      
      10 ms (33.3%) - 1 samples:
          foo
        -> bar
      
      10 ms (33.3%) - 1 samples:
          foo
        -> baz
      
      ### MEMORY ALLOCATION (by function)
         50.00 MB bar
         50.00 MB baz
      
      ### Next steps
      pv_focus(p, "bar")
      pv_suggestions(p)
      pv_help()

# pv_print_recursive snapshot for recursive profile

    Code
      pv_print_recursive(p)
    Output
      ## RECURSIVE FUNCTIONS
      
      
      Function                       MaxDepth AvgDepth   Total ms      Pct
      recurse                               5      4.0         30   100.0%
      
      ### Next steps
      pv_focus(p, "recurse")
      pv_suggestions(p)

# pv_print_recursive handles non-recursive profile

    Code
      pv_print_recursive(p)
    Output
      No recursive functions detected in the profile.
      Run pv_help() to see all available functions.

# pv_print_memory by function snapshot

    Code
      pv_print_memory(p, by = "function")
    Output
      ## MEMORY ALLOCATION BY FUNCTION
      
      
        150.00 MB inner
        100.00 MB deep
         50.00 MB helper
      
      ### Next steps
      pv_focus(p, "inner")
      pv_gc_pressure(p)

# pv_print_memory by line snapshot

    Code
      pv_print_memory(p, by = "line")
    Output
      ## MEMORY ALLOCATION BY LINE
      
      
        150.00 MB R/main.R:15
                  z <- heavy_computation()
        100.00 MB R/utils.R:5
                  x <- rnorm(1000)
         50.00 MB R/helper.R:20
                  do_work()
      
      ### Next steps
      pv_focus(p, "inner")
      pv_source_context(p, "R/main.R")

# pv_print_memory handles no source refs for line mode

    Code
      pv_print_memory(p, by = "line")
    Output
      No source location data available.
      Use devtools::load_all() to enable source references.
      Run pv_help() to see all available functions.

# pv_print_memory shows no allocations message for by function

    Code
      pv_print_memory(p, by = "function")
    Output
      No significant memory allocations detected.
      Run pv_help() to see all available functions.

# pv_print_debrief handles profile with no memory allocations

    Code
      pv_print_debrief(p)
    Output
      ## PROFILING SUMMARY
      
      
      Total time: 20 ms (2 samples @ 10 ms interval)
      Source references: not available (use devtools::load_all())
      
      
      ### TOP FUNCTIONS BY SELF-TIME
          20 ms (100.0%)  inner
      
      ### TOP FUNCTIONS BY TOTAL TIME
          20 ms (100.0%)  inner
          20 ms (100.0%)  outer
      
      ### HOT CALL PATHS
      
      20 ms (100.0%) - 2 samples:
          outer
        -> inner
      
      ### MEMORY ALLOCATION (by function)
      No significant memory allocations detected.
      
      ### Next steps
      pv_focus(p, "inner")
      pv_suggestions(p)
      pv_help()

# pv_print_call_depth snapshot

    Code
      pv_print_call_depth(p)
    Output
      ## CALL DEPTH BREAKDOWN
      
      
      Depth  Time (ms)   Pct   Top functions
          1        50  100.0%  outer
          2        40   80.0%  inner, helper
          3        20   40.0%  deep
      
      ### Next steps
      pv_focus(p, "outer")
      pv_flame(p)

# pv_print_callers_callees snapshot

    Code
      pv_print_callers_callees(p, "inner")
    Output
      ## FUNCTION ANALYSIS: inner
      
      
      Total time: 30 ms (60.0% of profile)
      Appearances: 3 samples
      
      ### Called by
            3 samples (100.0%)  outer
      
      ### Calls to
            2 samples ( 66.7%)  deep
      
      ### Next steps
      pv_focus(p, "inner")
      pv_focus(p, "outer")
      pv_focus(p, "deep")

# pv_print_call_stats snapshot

    Code
      pv_print_call_stats(p)
    Output
      ## CALL STATISTICS
      
      
      Function                               Calls   Total ms    Self ms    ms/call    Pct
      outer                                      1         50         10      50.00 100.0%
      inner                                      2         30         10      15.00  60.0%
      deep                                       2         20         20      10.00  40.0%
      helper                                     1         10         10      10.00  20.0%
      
      ### Next steps
      pv_focus(p, "outer")
      pv_callers(p, "outer")

# pv_print_callers_callees shows none when callees empty

    Code
      pv_print_callers_callees(p, "bar")
    Output
      ## FUNCTION ANALYSIS: bar
      
      
      Total time: 10 ms (33.3% of profile)
      Appearances: 1 samples
      
      ### Called by
            1 samples (100.0%)  foo
      
      ### Calls to
        Callees: none
      
      ### Next steps
      pv_focus(p, "bar")
      pv_focus(p, "foo")

# pv_help snapshot

    Code
      pv_help()
    Output
      ## DEBRIEF FUNCTIONS
      
      Use pv_help("category") to see functions in a specific category.
      Categories: overview, time, hotspots, memory, calls, source, visualization, comparison, diagnostics, export 
      
      ### Overview 
        pv_print_debrief(p)                 Print comprehensive profiling summary
        pv_debrief(p)                       Get all profiling data as a list
        pv_help()                           Show this help
        pv_example()                        Load example profvis data for testing
      
      ### Time Analysis 
        pv_self_time(p)                     Functions by self-time (time at top of stack)
        pv_total_time(p)                    Functions by total time (including callees)
      
      ### Hot Spots 
        pv_hot_lines(p)                     Hottest source lines by self-time
        pv_print_hot_lines(p)               Print hot lines with source context
        pv_hot_paths(p)                     Most common call stack paths
        pv_print_hot_paths(p)               Print hot paths in readable format
        pv_worst_line(p)                    Get the single hottest line with context
      
      ### Memory Analysis 
        pv_memory(p)                        Memory allocation by function
        pv_memory_lines(p)                  Memory allocation by source line
        pv_print_memory(p)                  Print memory allocation summary
      
      ### Call Analysis 
        pv_focus(p, "func")                 Deep dive into a specific function
        pv_callers(p, "func")               Who calls this function?
        pv_callees(p, "func")               What does this function call?
        pv_print_callers_callees(p, "func") Print caller/callee analysis
        pv_call_stats(p)                    Call counts and per-call timing
        pv_print_call_stats(p)              Print call statistics table
        pv_call_depth(p)                    Time distribution by call stack depth
        pv_recursive(p)                     Detect recursive functions
      
      ### Source Context 
        pv_source_context(p, "file.R")      Show source with profiling overlay
        pv_file_summary(p)                  Time breakdown by source file
        pv_print_file_summary(p)            Print file-level summary
      
      ### Visualization 
        pv_flame(p)                         Text-based flame graph
        pv_flame_condense(p)                Condensed flame graph view
      
      ### Comparison 
        pv_compare(before, after)           Compare two profiles
        pv_print_compare(before, after)     Print profile comparison
        pv_compare_many(...)                Compare multiple profiles
        pv_print_compare_many(...)          Print multi-profile comparison
      
      ### Diagnostics 
        pv_gc_pressure(p)                   Detect garbage collection overhead
        pv_print_gc_pressure(p)             Print GC pressure analysis
        pv_suggestions(p)                   Get optimization suggestions
        pv_print_suggestions(p)             Print actionable suggestions
      
      ### Export 
        pv_to_list(p)                       Export as R list
        pv_to_json(p)                       Export as JSON string
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help with category snapshot

    Code
      pv_help("time")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Time Analysis 
        pv_self_time(p)                     Functions by self-time (time at top of stack)
        pv_total_time(p)                    Functions by total time (including callees)
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help hotspots snapshot

    Code
      pv_help("hotspots")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Hot Spots 
        pv_hot_lines(p)                     Hottest source lines by self-time
        pv_print_hot_lines(p)               Print hot lines with source context
        pv_hot_paths(p)                     Most common call stack paths
        pv_print_hot_paths(p)               Print hot paths in readable format
        pv_worst_line(p)                    Get the single hottest line with context
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help memory snapshot

    Code
      pv_help("memory")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Memory Analysis 
        pv_memory(p)                        Memory allocation by function
        pv_memory_lines(p)                  Memory allocation by source line
        pv_print_memory(p)                  Print memory allocation summary
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help calls snapshot

    Code
      pv_help("calls")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Call Analysis 
        pv_focus(p, "func")                 Deep dive into a specific function
        pv_callers(p, "func")               Who calls this function?
        pv_callees(p, "func")               What does this function call?
        pv_print_callers_callees(p, "func") Print caller/callee analysis
        pv_call_stats(p)                    Call counts and per-call timing
        pv_print_call_stats(p)              Print call statistics table
        pv_call_depth(p)                    Time distribution by call stack depth
        pv_recursive(p)                     Detect recursive functions
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help source snapshot

    Code
      pv_help("source")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Source Context 
        pv_source_context(p, "file.R")      Show source with profiling overlay
        pv_file_summary(p)                  Time breakdown by source file
        pv_print_file_summary(p)            Print file-level summary
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help comparison snapshot

    Code
      pv_help("comparison")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Comparison 
        pv_compare(before, after)           Compare two profiles
        pv_print_compare(before, after)     Print profile comparison
        pv_compare_many(...)                Compare multiple profiles
        pv_print_compare_many(...)          Print multi-profile comparison
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help diagnostics snapshot

    Code
      pv_help("diagnostics")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Diagnostics 
        pv_gc_pressure(p)                   Detect garbage collection overhead
        pv_print_gc_pressure(p)             Print GC pressure analysis
        pv_suggestions(p)                   Get optimization suggestions
        pv_print_suggestions(p)             Print actionable suggestions
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help export snapshot

    Code
      pv_help("export")
    Output
      ## DEBRIEF FUNCTIONS
      
      ### Export 
        pv_to_list(p)                       Export as R list
        pv_to_json(p)                       Export as JSON string
      
      ### Typical Workflow
        1. pv_print_debrief(p)
              -> Get overview, identify hot functions
        2. pv_focus(p, "hot_func")
              -> Deep dive into the hottest function
        3. pv_hot_lines(p)
              -> Find exact lines consuming time
        4. pv_source_context(p, "file.R")
              -> See code with profiling overlay
        5. pv_suggestions(p)
              -> Get optimization recommendations

# pv_help with unknown category snapshot

    Code
      pv_help("nonexistent")
    Output
      Unknown category: nonexistent 
      Valid categories: overview, time, hotspots, memory, calls, source, visualization, comparison, diagnostics, export 

# pv_print_compare snapshot

    Code
      pv_print_compare(p1, p2)
    Output
      ## PROFILE COMPARISON
      
      
      
      ### Overall
      before_ms: 50
      after_ms: 50
      diff_ms: +0
      speedup: 1.00x
      
      ### Biggest Changes
      Function                           Before      After       Diff   Change
      
      ### Next steps
      pv_focus(p_before, "deep")
      pv_focus(p_after, "deep")

# pv_print_compare shows improvements and regressions

    Code
      pv_print_compare(p1, p2)
    Output
      ## PROFILE COMPARISON
      
      
      
      ### Overall
      before_ms: 100
      after_ms: 100
      diff_ms: +0
      speedup: 1.00x
      
      ### Biggest Changes
      Function                           Before      After       Diff   Change
      slow_func                              50          0        -50    -100%
      new_func                                0         50        +50      new
      
      ### Top Improvements
        slow_func: 50 -> 0 (-50 ms)
      
      ### Regressions
        new_func: 0 -> 50 (+50 ms)
      
      ### Next steps
      pv_focus(p_before, "slow_func")
      pv_focus(p_after, "slow_func")

# pv_compare_many requires at least 2 profiles

    Code
      pv_compare_many(a = p)
    Condition
      Error:
      ! At least 2 profiles are required for comparison.

# pv_compare_many requires named profiles

    Code
      pv_compare_many(p1, p2)
    Condition
      Error:
      ! All profiles must be named.

# pv_compare_many rejects non-profvis input

    Code
      pv_compare_many(a = p, b = list())
    Condition
      Error:
      ! 'b' must be a profvis object.

# pv_print_compare_many snapshot

    Code
      pv_print_compare_many(a = p1, b = p2)
    Output
      ## MULTI-PROFILE COMPARISON
      
      
      Rank  Profile                    Time (ms)  Samples vs Fastest
        1*  a                                 50        5    fastest
        2   b                                 50        5    fastest
      
      * = fastest

# pv_print_hot_lines snapshot

    Code
      pv_print_hot_lines(p, n = 3)
    Output
      ## HOT SOURCE LINES
      
      
      Rank 1: R/utils.R:5 (20 ms, 40.0%)
      Function: deep
      
              2: deep <- function() {
              3:   Sys.sleep(0.01)
              4:   42
      >       5:   x <- rnorm(1000)
              6: }
      
      Rank 2: R/helper.R:20 (10 ms, 20.0%)
      Function: helper
      
             17: 
             18: 
             19: 
      >      20:   do_work()
      
      Rank 3: R/main.R:10 (10 ms, 20.0%)
      Function: outer
      
              7: 
              8: inner <- function() {
              9:   result <- deep()
      >      10:   result
             11: }
             12: 
             13: 
      
      
      ### Next steps
      pv_focus(p, "deep")
      pv_source_context(p, "R/utils.R")

# pv_print_hot_lines handles no source refs

    Code
      pv_print_hot_lines(p)
    Output
      No source location data available.
      Use devtools::load_all() to enable source references.
      Run pv_help() to see all available functions.

# pv_print_hot_paths snapshot

    Code
      pv_print_hot_paths(p, n = 3)
    Output
      ## HOT CALL PATHS
      
      
      Rank 1: 20 ms (40.0%) - 2 samples
          outer (R/main.R:10)
        -> inner (R/main.R:15)
        -> deep (R/utils.R:5)
      
      Rank 2: 10 ms (20.0%) - 1 samples
          outer (R/main.R:10)
      
      Rank 3: 10 ms (20.0%) - 1 samples
          outer (R/main.R:10)
        -> helper (R/helper.R:20)
      
      
      ### Next steps
      pv_focus(p, "deep")
      pv_flame(p)

# pv_print_hot_lines shows source not available when files empty

    Code
      pv_print_hot_lines(p)
    Output
      ## HOT SOURCE LINES
      
      
      Rank 1: R/main.R:5 (30 ms, 100.0%)
      Function: func
      
        (source not available)
      
      
      ### Next steps
      pv_focus(p, "func")
      pv_source_context(p, "R/main.R")

# pv_focus snapshot for existing function

    Code
      pv_focus(p, "inner")
    Output
      ## FOCUS: inner
      
      
      ### Time Analysis
        Total time:       30 ms ( 60.0%)  - time on call stack
        Self time:        10 ms ( 20.0%)  - time at top of stack
        Child time:       20 ms ( 40.0%)  - time in callees
        Appearances:       3 samples
      
      ### Called By
            3 calls (100.0%)  outer
      
      ### Calls To
            2 calls ( 66.7%)  deep
      
      ### Source Locations
      Hot lines (by self-time):
           10 ms (20.0%)  R/main.R:15
                         z <- heavy_computation()
      
      ### Source Context: R/main.R
             10:   result
             11: }
             12: 
             13: 
             14: 
      >      15:   z <- heavy_computation()
      
      ### Next steps
      pv_focus(p, "deep")
      pv_callers(p, "inner")
      pv_focus(p, "outer")
      pv_source_context(p, "R/main.R")

# pv_focus handles non-existent function

    Code
      pv_focus(p, "nonexistent")
    Output
      Function 'nonexistent' not found in profiling data.
      
      Available functions (top 20 by time):
        outer
        inner
        deep
        helper
      
      Run pv_help() to see all available functions.

# pv_focus handles no source refs

    Code
      pv_focus(p, "foo")
    Output
      ## FOCUS: foo
      
      
      ### Time Analysis
        Total time:       30 ms (100.0%)  - time on call stack
        Self time:        10 ms ( 33.3%)  - time at top of stack
        Child time:       20 ms ( 66.7%)  - time in callees
        Appearances:       3 samples
      
      ### Called By
            3 calls (100.0%)  (top-level)
      
      ### Calls To
            1 calls ( 33.3%)  bar
            1 calls ( 33.3%)  baz
      
      ### Source Locations
        Source references not available.
        Use devtools::load_all() to enable.
      
      ### Next steps
      pv_focus(p, "bar")
      pv_callers(p, "foo")

# pv_focus handles more than 5 callers

    Code
      pv_focus(p, "target")
    Output
      ## FOCUS: target
      
      
      ### Time Analysis
        Total time:       60 ms (100.0%)  - time on call stack
        Self time:        60 ms (100.0%)  - time at top of stack
        Child time:        0 ms (  0.0%)  - time in callees
        Appearances:       6 samples
      
      ### Called By
            1 calls ( 16.7%)  caller1
            1 calls ( 16.7%)  caller2
            1 calls ( 16.7%)  caller3
            1 calls ( 16.7%)  caller4
            1 calls ( 16.7%)  caller5
        ... and 1 more callers
      
      ### Calls To
        Callees: none
      
      ### Source Locations
        Source references not available.
        Use devtools::load_all() to enable.
      
      ### Next steps
      pv_callers(p, "target")
      pv_focus(p, "caller1")

# pv_focus shows callees none for leaf function

    Code
      pv_focus(p, "bar")
    Output
      ## FOCUS: bar
      
      
      ### Time Analysis
        Total time:       10 ms ( 33.3%)  - time on call stack
        Self time:        10 ms ( 33.3%)  - time at top of stack
        Child time:        0 ms (  0.0%)  - time in callees
        Appearances:       1 samples
      
      ### Called By
            1 calls (100.0%)  foo
      
      ### Calls To
        Callees: none
      
      ### Source Locations
        Source references not available.
        Use devtools::load_all() to enable.
      
      ### Next steps
      pv_callers(p, "bar")
      pv_focus(p, "foo")

# pv_focus handles more than 5 callees

    Code
      pv_focus(p, "target")
    Output
      ## FOCUS: target
      
      
      ### Time Analysis
        Total time:       60 ms (100.0%)  - time on call stack
        Self time:         0 ms (  0.0%)  - time at top of stack
        Child time:       60 ms (100.0%)  - time in callees
        Appearances:       6 samples
      
      ### Called By
            6 calls (100.0%)  (top-level)
      
      ### Calls To
            1 calls ( 16.7%)  callee1
            1 calls ( 16.7%)  callee2
            1 calls ( 16.7%)  callee3
            1 calls ( 16.7%)  callee4
            1 calls ( 16.7%)  callee5
        ... and 1 more callees
      
      ### Source Locations
        Source references not available.
        Use devtools::load_all() to enable.
      
      ### Next steps
      pv_focus(p, "callee1")
      pv_callers(p, "target")

# pv_focus handles function with source refs but no self-time with source

    Code
      pv_focus(p, "target")
    Output
      ## FOCUS: target
      
      
      ### Time Analysis
        Total time:       10 ms (100.0%)  - time on call stack
        Self time:         0 ms (  0.0%)  - time at top of stack
        Child time:       10 ms (100.0%)  - time in callees
        Appearances:       1 samples
      
      ### Called By
            1 calls (100.0%)  (top-level)
      
      ### Calls To
            1 calls (100.0%)  deep
      
      ### Source Locations
        No self-time with source info.
      
      ### Next steps
      pv_focus(p, "deep")
      pv_callers(p, "target")

# empty profile errors with helpful message

    Code
      pv_self_time(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_total_time(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_debrief(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_print_debrief(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_worst_line(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_hot_lines(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_hot_paths(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_flame(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_memory(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_callers(empty_prof, "foo")
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_callees(empty_prof, "foo")
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_focus(empty_prof, "foo")
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_suggestions(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_gc_pressure(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_recursive(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_call_stats(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_call_depth(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_file_summary(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_source_context(empty_prof, "foo.R")
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_to_json(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_to_list(empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

# compare functions error on empty profile

    Code
      pv_compare(empty_prof, valid_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_compare(valid_prof, empty_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

---

    Code
      pv_compare_many(a = empty_prof, b = valid_prof)
    Condition
      Error:
      ! Profile contains no samples.
      Your code ran too fast to capture any profiling data.
      Try wrapping your code in a loop: for (i in 1:10) { ... }
      Increase the iteration count until samples appear.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/file-summary.R
\name{pv_file_summary}
\alias{pv_file_summary}
\title{File-level time summary}
\usage{
pv_file_summary(x)
}
\arguments{
\item{x}{A profvis object.}
}
\value{
A data frame with columns:
\itemize{
\item \code{filename}: Source file path
\item \code{samples}: Number of profiling samples
\item \code{time_ms}: Time in milliseconds
\item \code{pct}: Percentage of total time
}
}
\description{
Aggregates profiling time by source file. Requires source references
(use \code{devtools::load_all()} for best results).
}
\examples{
p <- pv_example()
pv_file_summary(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flame.R
\name{pv_flame_condense}
\alias{pv_flame_condense}
\title{Condensed flame graph}
\usage{
pv_flame_condense(x, n = 10, width = 50)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Number of hot paths to show.}

\item{width}{Width of bars.}
}
\value{
Invisibly returns a \code{debrief_flame_condense} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Shows a simplified, condensed view of the flame graph focusing on the
hottest paths.
}
\examples{
p <- pv_example()
pv_flame_condense(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/source-context.R
\name{pv_source_context}
\alias{pv_source_context}
\title{Show source context for a specific location}
\usage{
pv_source_context(x, filename, linenum = NULL, context = 10)
}
\arguments{
\item{x}{A profvis object.}

\item{filename}{The source file to examine.}

\item{linenum}{The line number to center on. If \code{NULL}, shows the hottest
line in the file.}

\item{context}{Number of lines to show before and after.}
}
\value{
Invisibly returns a \code{debrief_source_context} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Displays source code around a specific file and line number with profiling
information for each line.
}
\examples{
p <- pv_example()
pv_source_context(p, "R/main.R", linenum = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/call-depth.R
\name{pv_call_depth}
\alias{pv_call_depth}
\title{Call depth breakdown}
\usage{
pv_call_depth(x)
}
\arguments{
\item{x}{A profvis object.}
}
\value{
A data frame with columns:
\itemize{
\item \code{depth}: Call stack depth (1 = top level)
\item \code{samples}: Number of profiling samples at this depth
\item \code{time_ms}: Time in milliseconds
\item \code{pct}: Percentage of total time
\item \code{top_funcs}: Most common functions at this depth
}
}
\description{
Shows time distribution across different call stack depths. Useful for
understanding how deeply nested the hot code paths are.
}
\examples{
p <- pv_example()
pv_call_depth(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/help.R
\name{pv_help}
\alias{pv_help}
\title{List available debrief functions}
\usage{
pv_help(category = NULL)
}
\arguments{
\item{category}{Optional category to filter by. If \code{NULL}, shows all
categories. Options: "overview", "time", "hotspots", "memory", "calls",
"source", "visualization", "comparison", "diagnostics", "export".}
}
\value{
Invisibly returns a \code{debrief_help} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Prints a summary of all available functions in the debrief package,
organized by category. Useful for discovering the API.
}
\examples{
pv_help()
pv_help("time")
pv_help("hotspots")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hot-lines.R
\name{pv_hot_lines}
\alias{pv_hot_lines}
\title{Hot source lines by self-time}
\usage{
pv_hot_lines(x, n = NULL, min_pct = 0, min_time_ms = 0)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Maximum number of lines to return. If \code{NULL}, returns all that pass
the filters.}

\item{min_pct}{Minimum percentage of total time to include (default 0).}

\item{min_time_ms}{Minimum time in milliseconds to include (default 0).}
}
\value{
A data frame with columns:
\itemize{
\item \code{location}: File path and line number (e.g., "R/foo.R:42")
\item \code{label}: Function name at this location
\item \code{filename}: Source file path
\item \code{linenum}: Line number
\item \code{samples}: Number of profiling samples
\item \code{time_ms}: Time in milliseconds
\item \code{pct}: Percentage of total time
}
}
\description{
Returns the source lines where the most CPU time is spent. Requires source
references (use \code{devtools::load_all()} for best results).
}
\examples{
p <- pv_example()
pv_hot_lines(p)

# Only lines with >= 10\% of time
pv_hot_lines(p, min_pct = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hot-paths.R
\name{pv_hot_paths}
\alias{pv_hot_paths}
\title{Hot call paths}
\usage{
pv_hot_paths(x, n = NULL, include_source = TRUE)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Maximum number of paths to return. If \code{NULL}, returns all.}

\item{include_source}{If \code{TRUE} and source references are available, include
file:line information in the path labels.}
}
\value{
A data frame with columns:
\itemize{
\item \code{stack}: The call path (functions separated by arrows)
\item \code{samples}: Number of profiling samples with this exact path
\item \code{time_ms}: Time in milliseconds
\item \code{pct}: Percentage of total time
}
}
\description{
Returns the most common complete call stacks. This shows which execution
paths through the code consume the most time.
}
\examples{
p <- pv_example()
pv_hot_paths(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/call-depth.R
\name{pv_print_call_depth}
\alias{pv_print_call_depth}
\title{Print call depth breakdown}
\usage{
pv_print_call_depth(x)
}
\arguments{
\item{x}{A profvis object.}
}
\value{
Invisibly returns a \code{debrief_call_depth} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Print call depth breakdown
}
\examples{
p <- pv_example()
pv_print_call_depth(p)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export.R
\name{pv_to_json}
\alias{pv_to_json}
\title{Export profiling results as JSON}
\usage{
pv_to_json(
  x,
  file = NULL,
  pretty = TRUE,
  include = c("summary", "self_time", "total_time", "hot_lines", "memory", "gc_pressure",
    "suggestions", "recursive"),
  system_info = FALSE
)
}
\arguments{
\item{x}{A profvis object.}

\item{file}{Optional file path to write JSON to. If \code{NULL}, returns the
JSON string.}

\item{pretty}{If \code{TRUE}, formats JSON with indentation for readability.}

\item{include}{Character vector specifying which analyses to include.
Options: "summary", "self_time", "total_time", "hot_lines", "memory",
"callers", "gc_pressure", "suggestions", "recursive". Default includes all.}

\item{system_info}{If \code{TRUE}, includes R version and platform info in
metadata. Useful for reproducibility.}
}
\value{
If \code{file} is \code{NULL}, returns a JSON string. Otherwise writes to file
and returns the file path invisibly.
}
\description{
Exports profiling analysis results in JSON format for consumption by
AI agents, automated tools, or external applications.
}
\examples{
p <- pv_example()
json <- pv_to_json(p)
cat(json)

# Include only specific analyses
json <- pv_to_json(p, include = c("self_time", "hot_lines"))

# Include system info for reproducibility
json <- pv_to_json(p, system_info = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flame.R
\name{pv_flame}
\alias{pv_flame}
\title{Text-based flame graph}
\usage{
pv_flame(x, width = 70, min_pct = 2, max_depth = 15)
}
\arguments{
\item{x}{A profvis object.}

\item{width}{Width of the flame graph in characters.}

\item{min_pct}{Minimum percentage to show (filters small slices).}

\item{max_depth}{Maximum depth to display.}
}
\value{
Invisibly returns a \code{debrief_flame} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Generates an ASCII representation of a flame graph showing the hierarchical
breakdown of time spent in the call tree.
}
\examples{
p <- pv_example()
pv_flame(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recursive.R
\name{pv_recursive}
\alias{pv_recursive}
\title{Detect recursive functions}
\usage{
pv_recursive(x)
}
\arguments{
\item{x}{A profvis object.}
}
\value{
A data frame with columns:
\itemize{
\item \code{label}: Function name
\item \code{max_depth}: Maximum recursion depth observed
\item \code{avg_depth}: Average recursion depth when recursive
\item \code{recursive_samples}: Number of samples where function appears multiple times
\item \code{total_samples}: Total samples where function appears
\item \code{pct_recursive}: Percentage of appearances that are recursive
\item \code{total_ms}: Total time on call stack
\item \code{pct_time}: Percentage of total profile time
}
}
\description{
Identifies functions that call themselves (directly recursive) or appear
multiple times in the same call stack. Recursive functions in hot paths
are often optimization targets.
}
\examples{
p <- pv_example("recursive")
pv_recursive(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/antipatterns.R
\name{pv_gc_pressure}
\alias{pv_gc_pressure}
\title{Detect GC pressure}
\usage{
pv_gc_pressure(x, threshold = 10)
}
\arguments{
\item{x}{A profvis object.}

\item{threshold}{Minimum GC percentage to report (default 10). Set lower to
detect smaller GC overhead.}
}
\value{
A data frame with columns:
\itemize{
\item \code{severity}: "high" (>25\%), "medium" (>15\%), or "low" (>threshold\%)
\item \code{pct}: Percentage of total time spent in GC
\item \code{time_ms}: Time spent in garbage collection
\item \code{issue}: Short description of the problem
\item \code{cause}: What typically causes this issue
\item \code{actions}: Comma-separated list of things to look for
}

Returns an empty data frame (0 rows) if GC is below the threshold.
}
\description{
Analyzes the profile to detect excessive garbage collection, which is a
universal indicator of memory allocation issues in R code.
}
\details{
GC pressure above 10\% typically indicates the code is allocating and
discarding memory faster than necessary. Common causes include:
\itemize{
\item Growing vectors with \code{c(x, new)} instead of pre-allocation
\item Building data frames row-by-row with \code{rbind()}
\item Creating unnecessary copies of large objects
\item String concatenation in loops
}
}
\examples{
p <- pv_example("gc")
pv_gc_pressure(p)

# More sensitive detection
pv_gc_pressure(p, threshold = 5)

# No GC pressure in default example
p2 <- pv_example()
pv_gc_pressure(p2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/call-stats.R
\name{pv_print_call_stats}
\alias{pv_print_call_stats}
\title{Print call statistics}
\usage{
pv_print_call_stats(x, n = 20)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Number of functions to show.}
}
\value{
Invisibly returns a \code{debrief_call_stats} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Print call statistics
}
\examples{
p <- pv_example()
pv_print_call_stats(p)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time.R
\name{pv_total_time}
\alias{pv_total_time}
\title{Total time summary by function}
\usage{
pv_total_time(x, n = NULL, min_pct = 0, min_time_ms = 0)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Maximum number of functions to return. If \code{NULL}, returns all that
pass the filters.}

\item{min_pct}{Minimum percentage of total time to include (default 0).}

\item{min_time_ms}{Minimum time in milliseconds to include (default 0).}
}
\value{
A data frame with columns:
\itemize{
\item \code{label}: Function name
\item \code{samples}: Number of profiling samples where function appeared
\item \code{time_ms}: Time in milliseconds
\item \code{pct}: Percentage of total time
}
}
\description{
Returns the time spent in each function including all its callees. This
shows which functions are on the call stack when time is being spent.
}
\examples{
p <- pv_example()
pv_total_time(p)

# Only functions with >= 50\% total time
pv_total_time(p, min_pct = 50)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callers-callees.R
\name{pv_callees}
\alias{pv_callees}
\title{Get callees of a function}
\usage{
pv_callees(x, func)
}
\arguments{
\item{x}{A profvis object.}

\item{func}{The function name to analyze.}
}
\value{
A data frame with columns:
\itemize{
\item \code{label}: Callee function name
\item \code{samples}: Number of times this callee appeared
\item \code{pct}: Percentage of calls to this callee
}
}
\description{
Returns the functions that a specified function calls, based on profiling
data. Shows what the target function invokes.
}
\examples{
p <- pv_example()
pv_callees(p, "outer")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/debrief.R
\name{pv_print_debrief}
\alias{pv_print_debrief}
\title{Print profiling summary}
\usage{
pv_print_debrief(x, n_functions = 10, n_lines = 10, n_paths = 5, n_memory = 5)
}
\arguments{
\item{x}{A profvis object from \code{\link[profvis:profvis]{profvis::profvis()}}.}

\item{n_functions}{Number of top functions to show (default 10).}

\item{n_lines}{Number of hot source lines to show (default 10).}

\item{n_paths}{Number of hot paths to show (default 5).}

\item{n_memory}{Number of memory hotspots to show (default 5).}
}
\value{
Invisibly returns a \code{debrief_debrief} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Prints a comprehensive text summary of profiling data suitable for
terminal output or AI agent consumption.
}
\examples{
p <- pv_example()
pv_print_debrief(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\name{pv_memory_lines}
\alias{pv_memory_lines}
\title{Memory allocation by source line}
\usage{
pv_memory_lines(x, n = NULL)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Maximum number of lines to return. If \code{NULL}, returns all.}
}
\value{
A data frame with columns:
\itemize{
\item \code{location}: File path and line number
\item \code{label}: Function name at this location
\item \code{filename}: Source file path
\item \code{linenum}: Line number
\item \code{mem_mb}: Memory allocated in megabytes
}
}
\description{
Returns memory allocation aggregated by source location. Requires source
references (use \code{devtools::load_all()} for best results).
}
\examples{
p <- pv_example()
pv_memory_lines(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/antipatterns.R
\name{pv_print_gc_pressure}
\alias{pv_print_gc_pressure}
\title{Print GC pressure analysis}
\usage{
pv_print_gc_pressure(x, threshold = 10)
}
\arguments{
\item{x}{A profvis object.}

\item{threshold}{Minimum GC percentage to report (default 10).}
}
\value{
Invisibly returns a \code{debrief_gc_pressure} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Print GC pressure analysis
}
\examples{
p <- pv_example("gc")
pv_print_gc_pressure(p)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hot-lines.R
\name{pv_print_hot_lines}
\alias{pv_print_hot_lines}
\title{Print hot lines with source context}
\usage{
pv_print_hot_lines(x, n = 5, context = 3)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Number of hot lines to show.}

\item{context}{Number of lines to show before and after each hotspot.}
}
\value{
Invisibly returns a \code{debrief_hot_lines} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Prints the hot source lines along with surrounding code context.
}
\examples{
p <- pv_example()
pv_print_hot_lines(p, n = 5, context = 3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/debrief.R
\name{pv_debrief}
\alias{pv_debrief}
\title{Comprehensive profiling data}
\usage{
pv_debrief(x, n = 10)
}
\arguments{
\item{x}{A profvis object from \code{\link[profvis:profvis]{profvis::profvis()}}.}

\item{n}{Maximum number of items to include in each category (default 10).}
}
\value{
A list containing:
\itemize{
\item \code{total_time_ms}: Total profiling time in milliseconds
\item \code{total_samples}: Number of profiling samples
\item \code{interval_ms}: Sampling interval in milliseconds
\item \code{has_source}: Whether source references are available
\item \code{self_time}: Data frame of functions by self-time
\item \code{total_time}: Data frame of functions by total time
\item \code{hot_lines}: Data frame of hot source lines (or NULL if no source refs)
\item \code{hot_paths}: Data frame of hot call paths
\item \code{suggestions}: Data frame of optimization suggestions
\item \code{gc_pressure}: Data frame of GC pressure analysis
\item \code{memory}: Data frame of memory allocation by function
\item \code{memory_lines}: Data frame of memory allocation by line (or NULL)
}
}
\description{
Returns all profiling analysis in a single list for programmatic access.
This is the primary function for AI agents and scripts that need
comprehensive profiling data without printed output.
}
\examples{
p <- pv_example()
d <- pv_debrief(p)
names(d)
d$self_time
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/debrief-package.R
\docType{package}
\name{debrief-package}
\alias{debrief}
\alias{debrief-package}
\title{debrief: Text-Based Summaries for 'profvis' Profiling Data}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides text-based summaries and analysis tools for 'profvis' profiling output. Designed for terminal workflows and artificial intelligence (AI) agent consumption, offering views including hotspot analysis, call trees, source context, caller/callee relationships, and memory allocation breakdowns.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lib/debrief}
  \item \url{https://r-lib.github.io/debrief/}
  \item Report bugs at \url{https://github.com/r-lib/debrief/issues}
}

}
\author{
\strong{Maintainer}: Emil Hvitfeldt \email{emil.hvitfeldt@posit.co}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{pv_compare}
\alias{pv_compare}
\title{Compare two profvis profiles}
\usage{
pv_compare(before, after, n = 20)
}
\arguments{
\item{before}{A profvis object (before optimization).}

\item{after}{A profvis object (after optimization).}

\item{n}{Number of top functions to compare.}
}
\value{
A list with:
\itemize{
\item \code{summary}: Overall comparison summary
\item \code{by_function}: Function-by-function comparison
\item \code{improved}: Functions that got faster
\item \code{regressed}: Functions that got slower
}
}
\description{
Compares two profiling runs to show what changed. Useful for measuring
the impact of optimizations.
}
\examples{
p1 <- pv_example()
p2 <- pv_example()
pv_compare(p1, p2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/suggestions.R
\name{pv_suggestions}
\alias{pv_suggestions}
\title{Generate optimization suggestions}
\usage{
pv_suggestions(x)
}
\arguments{
\item{x}{A profvis object.}
}
\value{
A data frame with columns:
\itemize{
\item \code{priority}: 1 (highest) to 5 (lowest)
\item \code{category}: Type of optimization (e.g., "data structure", "algorithm")
\item \code{location}: Where to apply the optimization
\item \code{action}: What to do
\item \code{pattern}: Code pattern to look for (or NA)
\item \code{replacement}: Suggested replacement (or NA)
\item \code{potential_impact}: Estimated time that could be saved
}
}
\description{
Analyzes the profile and generates specific, actionable optimization
suggestions based on detected patterns and hotspots.
}
\examples{
p <- pv_example("gc")
pv_suggestions(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{pv_print_compare_many}
\alias{pv_print_compare_many}
\title{Print comparison of multiple profiles}
\usage{
pv_print_compare_many(...)
}
\arguments{
\item{...}{Named profvis objects to compare, or a single named list.}
}
\value{
Invisibly returns a \code{debrief_compare_many} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Print comparison of multiple profiles
}
\examples{
p1 <- pv_example()
p2 <- pv_example("gc")
pv_print_compare_many(baseline = p1, gc_heavy = p2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/focus.R
\name{pv_focus}
\alias{pv_focus}
\title{Focused analysis of a specific function}
\usage{
pv_focus(x, func, context = 5)
}
\arguments{
\item{x}{A profvis object.}

\item{func}{The function name to analyze.}

\item{context}{Number of source lines to show around hotspots.}
}
\value{
Invisibly returns a \code{debrief_focus} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Provides a comprehensive analysis of a single function including time
breakdown, callers, callees, and source context if available.
}
\examples{
p <- pv_example()
pv_focus(p, "inner")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/call-stats.R
\name{pv_call_stats}
\alias{pv_call_stats}
\title{Call statistics summary}
\usage{
pv_call_stats(x, n = NULL)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Maximum number of functions to return. If \code{NULL}, returns all.}
}
\value{
A data frame with columns:
\itemize{
\item \code{label}: Function name
\item \code{calls}: Estimated number of calls (based on stack appearances)
\item \code{total_ms}: Total time on call stack
\item \code{self_ms}: Time at top of stack (self time)
\item \code{child_ms}: Time in callees
\item \code{ms_per_call}: Average milliseconds per call
\item \code{pct}: Percentage of total profile time
}
}
\description{
Shows call count, total time, self time, and time per call for each
function. This is especially useful for identifying functions that are
called many times (where per-call optimization or reducing call count
would help).
}
\examples{
p <- pv_example()
pv_call_stats(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{pv_compare_many}
\alias{pv_compare_many}
\title{Compare multiple profvis profiles}
\usage{
pv_compare_many(...)
}
\arguments{
\item{...}{Named profvis objects to compare, or a single named list of
profvis objects.}
}
\value{
A data frame with columns:
\itemize{
\item \code{name}: Profile name
\item \code{time_ms}: Total time in milliseconds
\item \code{samples}: Number of samples
\item \code{vs_fastest}: How much slower than the fastest (e.g., "1.5x")
\item \code{rank}: Rank from fastest (1) to slowest
}
}
\description{
Compares multiple profiling runs to identify the fastest. Useful for
comparing different optimization approaches.
}
\examples{
p1 <- pv_example()
p2 <- pv_example("gc")
p3 <- pv_example("recursive")
pv_compare_many(baseline = p1, gc_heavy = p2, recursive = p3)

# Or pass a named list
profiles <- list(baseline = p1, gc_heavy = p2)
pv_compare_many(profiles)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/callers-callees.R
\name{pv_print_callers_callees}
\alias{pv_print_callers_callees}
\title{Print caller/callee analysis for a function}
\usage{
pv_print_callers_callees(x, func, n = 10)
}
\arguments{
\item{x}{A profvis object.}

\item{func}{The function name to analyze.}

\item{n}{Maximum number of callers/callees to show.}
}
\value{
Invisibly returns a \code{debrief_callers_callees} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Shows both callers (who calls this function) and callees (what this function
calls) in a single view.
}
\examples{
p <- pv_example()
pv_print_callers_callees(p, "inner")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hot-lines.R
\name{pv_worst_line}
\alias{pv_worst_line}
\title{Get the single hottest line}
\usage{
pv_worst_line(x, context = 5)
}
\arguments{
\item{x}{A profvis object.}

\item{context}{Number of source lines to include before and after.}
}
\value{
A list with:
\itemize{
\item \code{location}: File path and line number (e.g., "R/foo.R:42")
\item \code{label}: Function name
\item \code{filename}: Source file path
\item \code{linenum}: Line number
\item \code{time_ms}: Time in milliseconds
\item \code{pct}: Percentage of total time
\item \code{code}: The source line
\item \code{context}: Vector of surrounding source lines
\item \code{callers}: Data frame of functions that call this location
}

Returns \code{NULL} if no source references are available.
}
\description{
Returns the hottest source line with full context. Useful for quickly
identifying the #1 optimization target.
}
\examples{
p <- pv_example()
pv_worst_line(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example-data.R
\name{pv_example}
\alias{pv_example}
\title{Example profvis data}
\usage{
pv_example(type = c("default", "no_source", "recursive", "gc"))
}
\arguments{
\item{type}{Type of example data to create:
\itemize{
\item \code{"default"}: A real profile captured from example code with source refs
\item \code{"no_source"}: A synthetic profile without source references
\item \code{"recursive"}: A synthetic profile with recursive function calls
\item \code{"gc"}: A synthetic profile with garbage collection pressure
}}
}
\value{
A profvis object that can be used with all debrief functions.
}
\description{
Creates an example profvis object for use in examples and testing.
This avoids the need to run actual profiling code in examples.
}
\examples{
# Get default example data
p <- pv_example()
pv_self_time(p)

# Get example with recursive calls
p_recursive <- pv_example("recursive")
pv_recursive(p_recursive)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export.R
\name{pv_to_list}
\alias{pv_to_list}
\title{Export profiling results as a list}
\usage{
pv_to_list(
  x,
  include = c("summary", "self_time", "total_time", "hot_lines", "memory", "gc_pressure",
    "suggestions", "recursive"),
  system_info = FALSE
)
}
\arguments{
\item{x}{A profvis object.}

\item{include}{Character vector specifying which analyses to include.
Same options as \code{\link[=pv_to_json]{pv_to_json()}}.}

\item{system_info}{If \code{TRUE}, includes R version and platform info in
metadata.}
}
\value{
A named list containing the requested analyses.
}
\description{
Returns all profiling analysis results as a nested R list, useful for
programmatic access to results without JSON serialization.
}
\examples{
p <- pv_example()
results <- pv_to_list(p)
names(results)
results$self_time

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare.R
\name{pv_print_compare}
\alias{pv_print_compare}
\title{Print profile comparison}
\usage{
pv_print_compare(before, after, n = 15)
}
\arguments{
\item{before}{A profvis object (before optimization).}

\item{after}{A profvis object (after optimization).}

\item{n}{Number of functions to show in detailed comparison.}
}
\value{
Invisibly returns a \code{debrief_compare} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Print profile comparison
}
\examples{
p1 <- pv_example()
p2 <- pv_example()
pv_print_compare(p1, p2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hot-paths.R
\name{pv_print_hot_paths}
\alias{pv_print_hot_paths}
\title{Print hot paths in readable format}
\usage{
pv_print_hot_paths(x, n = 10, include_source = TRUE)
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Number of paths to show.}

\item{include_source}{Include source references in output.}
}
\value{
Invisibly returns a \code{debrief_hot_paths} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Print hot paths in readable format
}
\examples{
p <- pv_example()
pv_print_hot_paths(p, n = 3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/memory.R
\name{pv_print_memory}
\alias{pv_print_memory}
\title{Print memory allocation summary}
\usage{
pv_print_memory(x, n = 10, by = c("function", "line"))
}
\arguments{
\item{x}{A profvis object.}

\item{n}{Number of top allocators to show.}

\item{by}{Either "function" or "line".}
}
\value{
Invisibly returns a \code{debrief_memory} object. Use
\code{capture.output()} to capture the formatted text output.
}
\description{
Print memory allocation summary
}
\examples{
p <- pv_example()
pv_print_memory(p, by = "function")
pv_print_memory(p, by = "line")

}
